#include "autoclassto.h"
#include "../../memrep/clustermediator.h"

AutoClassTO::AutoClassTO(wv::CLUSTER_ID cid, wv::TRANSFORM transform, MatCPtr waves, const std::string& wfm_file) :
    ClusMethodTOBase(cid, transform, waves, waves->size1, wfm_file)
{
    _db2File = wfm_file + "." + std::to_string(cid) + ".db2-bin";
    _hd2File = wfm_file + "." + std::to_string(cid) + ".hd2";
    _modelFile = wfm_file + "." + std::to_string(cid) + ".model";
    _sParamsFile = wfm_file + "." + std::to_string(cid) + ".s-params";
    _rParamsFile = wfm_file + "." + std::to_string(cid) + ".r-params";
    _caseDataFile = wfm_file + "." +std::to_string(cid) + ".case-data-1";
}

void AutoClassTO::Start()
{
    StartTimer();

    //Generate the .db2-bin, .hd2, .model, .r-params and .s-params files :(
    MakeDb2();
    MakeHd2();
    MakeModel();
    MakeSParams();

    StringList proc_args;
    proc_args << "-search";
    proc_args << _db2File;
    proc_args << _hd2File;
    proc_args << _modelFile;
    proc_args << _sParamsFile;
    QProcess* search_proc = new QProcess(); //Must create here, so that it is in this thread
    Cluster(search_proc, proc_args);
}

void AutoClassTO::LoadClassification(int exit_code)
{
    try
    {
        //Run autoclass with the '-reports' option, which outputs the classification
        MakeRParams();
        StringList proc_args;
        proc_args << "-reports";
        proc_args << InputFullPath() + "." + std::to_string(ID()) + ".results-bin";
        proc_args << InputFullPath() + "." + std::to_string(ID()) + ".search";
        proc_args << InputFullPath() + "." + std::to_string(ID()) + ".r-params";

        _resultsProc = new QProcess();
#if defined(Q_WS_X11) || defined(Q_OS_LINUX)
                std::string proc_name = QApplication::applicationDirPath().toStdString()
                        + "/ClusProgs/" + Name() + "/" + Name();
#elif defined Q_WS_MAC || defined Q_OS_MACX
                std::string proc_name = QApplication::applicationDirPath().toStdString()
                        + "/../../../ClusProgs/" + Name() + "/" + Name();
#elif defined Q_OS_WIN
                std::string proc_name = QApplication::applicationDirPath().toStdString()
                        + "/../ClusProgs/" + Name() + "/release/" + Name();
#endif
        _resultsProc->start(proc_name.c_str(), ut::ToQStringList(proc_args));
        connect(_resultsProc,SIGNAL(readyReadStandardOutput()), this, SLOT(ReadStandardOutputR()));
        connect(_resultsProc,SIGNAL(readyReadStandardError()), this, SLOT(ReadStandardErrorR()));
        connect(_resultsProc,SIGNAL(stateChanged(QProcess::ProcessState)), this, SLOT(StateChangeR(QProcess::ProcessState)));
        connect(_resultsProc,SIGNAL(error(QProcess::ProcessError)), this, SLOT(ErrorR(QProcess::ProcessError)));
        _resultsProc->waitForFinished(-1);
        QApplication::processEvents();

        //Now read the .case-data file
        wv::float_t pval = boost::lexical_cast<wv::float_t>( Params()->Par("Classification p-value") );
        CaseDataFileIn case_data_file_in(_caseDataFile, Waves()->size1, pval);
        case_data_file_in.Load();
        if (case_data_file_in.IsNull())
            return;
        const wv_vector_int* vcluster = case_data_file_in.CData()->VClusters().at(0).second;
        case_data_file_in.Close();

        //Write the classification right back into a clu file
        std::vector<ConstClustering> clusterings;
        clusterings.push_back( ConstClustering(0.0, vcluster) );
        std::string mclu_file_name = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(), Transform(), ID(), "mclu", Electrode());
        MCluFileOut mclu_file(mclu_file_name, 0, clusterings);
        mclu_file.Save();
        mclu_file.Close(); //Make sure to close before reading this in with LoadClassification!

        //Then read the .clu file in the normal way
        ClusMethodTOBase::LoadClassification(exit_code);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

void AutoClassTO::ReadStandardOutputR()
{
    std::string out = _resultsProc->readAllStandardOutput().data();
    Log(Name() + ", -results," + " stdout: " + out);
}
void AutoClassTO::ReadStandardErrorR()
{
    std::string err = _resultsProc->readAllStandardError().data();
    Log(Name() + ", -results," + " stderr: " + err);
}
void AutoClassTO::StateChangeR(QProcess::ProcessState st)
{
    Log( Name() + ", -results," + " new state: " + std::to_string(st) );
}
void AutoClassTO::ErrorR(QProcess::ProcessError)
{
    Log( Name() + ", -results," + " error: " + _resultsProc->errorString().toStdString() );
}

void AutoClassTO::MakeDb2()
{
    int samples_begin = boost::lexical_cast<int>( Params()->Par("samples_begin") ),
            samples_end = boost::lexical_cast<int>( Params()->Par("samples_end") );
    if (samples_end-samples_begin == 0)
    {
        samples_begin = 0;
        samples_end = Waves()->size2;
    }
    Db2FileOut db2_out(_db2File, Waves(), samples_begin, samples_end);
    db2_out.Save();
}
void AutoClassTO::MakeHd2()
{
    int samples_begin = boost::lexical_cast<int>( Params()->Par("samples_begin") ),
            samples_end = boost::lexical_cast<int>( Params()->Par("samples_end") ),
            samples_range;
    if (samples_end-samples_begin == 0)
        samples_range = Waves()->size2;
    else
        samples_range = samples_end - samples_begin;

    ///Now calculate error, based on the range of sample values.
    wv_vector* err_vec = wv_vector_alloc(samples_range);
    wv::float_t scale = boost::lexical_cast<wv::float_t>( Params()->Par("rel_error") );
    if (Transform() == wv::USER)
        wv_vector_set_all(err_vec, scale);
    else
    {
        wv::float_t range = wv_matrix_max(Waves().get()) - wv_matrix_min(Waves().get());
        wv_vector_set_all(err_vec, scale*range);
    }

    Hd2FileOut hd2_out(_hd2File, Transform(), samples_range, err_vec);
    hd2_out.Save();
    wv_vector_free(err_vec);
}
void AutoClassTO::MakeModel()
{
    int samples_begin = boost::lexical_cast<int>( Params()->Par("samples_begin") ),
            samples_end = boost::lexical_cast<int>( Params()->Par("samples_end") ),
            samples_range;
    if (samples_end-samples_begin == 0)
        samples_range = Waves()->size2;
    else
        samples_range = samples_end - samples_begin;
    ModelFileOut model_file_out(_modelFile, Transform()==wv::USER, samples_range);
    model_file_out.Save();
}
void AutoClassTO::MakeSParams()
{
    ACParamsOut acs_out(_sParamsFile, ACParamsOut::SEARCH, Params(), Waves()->size1*Waves()->size2);
    acs_out.Save();
}
void AutoClassTO::MakeRParams()
{
    ACParamsOut acr_out(_rParamsFile, ACParamsOut::RESULTS, Params(), Waves()->size1*Waves()->size2);
    acr_out.Save();
}
