#include "manualto.h"
#include "../../memrep/clustermediator.h"

ManualTO::ManualTO(wv::CLUSTER_ID cid, WaveData* wave_data,
                   const std::string& wfm_file)
    : ClusMethodTOBase(cid, wv::UNKNOWN_TRANSFORM, wave_data, wave_data->NumRecords(), wfm_file)
{
    _cluSpec = ClusterMediator::Instance().GetCluSpecPtr( ID() );
    _toggledWaves = static_cast<const ClusterSpec*>(_cluSpec.get())->ToggledWaves();
}

void ManualTO::Start()
{
    try
    {
    StartTimer();

    const std::vector<const CenterSpec*>& centers = _cluSpec->Centers();
    const std::vector<CSpec*>& circles = _cluSpec->Circles();
    const std::vector<PLSpec*>& polylines = _cluSpec->Polylines();

    WaveData* wave_data = WData();
    MatCPtr ttime = wave_data->Waveforms(wv::TIMESTAMP);
    const wv_matrix* timestamps_mat = ttime.get();
    wv_vector_const_view tview = wv_matrix_const_column(timestamps_mat, 0);
    const wv_vector* timestamps = &tview.vector;
    int num_centers = centers.size(),
            num_circles = circles.size(),
            num_polylines = polylines.size();

    //Initialize cluster vector, all waveforms rejected by default *unless* there are no accept
    //circles, in which case the default is accepted.
    //  But if there are any records that are being used from the previous analysis, then the
    //subsequent clustering starts with them.
    int num_records = wave_data->NumRecords();
    wv_vector_int* vcluster = wv_vector_int_alloc(num_records);
    const boost::dynamic_bitset<>& prior_accepted = ClusterMediator::Instance().GetCluSpecPtr( ID() )->PriorAccepted();
    wv::CLUSTER_ID unaccepted = (num_centers==0) ? wv::REJECT_CLUSTER : wv::ZERO_CLUSTER;
        //So the idea is this: If there are centers, then any waveforms not explicitly rejected by a polylin
        //are divided up among them (ZERO_CLUSTER).  If there aren't, then all waveforms are rejected by
        //default unless there no circles, in which case they are accepted.
#ifdef QT_DEBUG
    qDebug() << "Number of prior accepted waveforms: " << prior_accepted.count() << " of " << prior_accepted.size();
#endif
    if (prior_accepted.count() == 0)
    {
        wv::CLUSTER_ID default_id = (num_circles==0 && num_centers==0) ? wv::ACCEPT_CLUSTER : unaccepted;
        wv_vector_int_set_all(vcluster, default_id);
    }
    else
        for (int i=0; i<num_records; ++i)
        {
            wv::CLUSTER_ID cid = prior_accepted.test(i) ? wv::ACCEPT_CLUSTER : unaccepted;
            wv_vector_int_set(vcluster, i, cid);
        }
    //If there are centers, all waveforms are divided up among them.  ZERO_CLUSTER is used so that
    //waveforms rejected by the polylines can be distinguished from waveforms that are not accepted

    //Go through the circles.  The *union* of all elements in the accepted clusters are accepted.
    for (int c=0; c<num_circles; ++c)
    {
        const CSpec* c_spec = circles.at(c);
        const Circle* circle = c_spec->GetDrawLine();

        //Setup the data for the parallelized Contains calculation
        ContainsMgr::ContainsSpec contains_spec(circle);
        contains_spec.sample_x = c_spec->sample_x;
        contains_spec.sample_y = c_spec->sample_y;
        contains_spec.time_begin = c_spec->time_begin;
        contains_spec.time_end = c_spec->time_end;
        contains_spec.time_restricted = (c_spec->time_begin!=0 || c_spec->time_end!=-1);
        contains_spec.timestamps = timestamps;
        MatCPtr tmat_x = wave_data->Waveforms(c_spec->transform_x),
               tmat_y = wave_data->Waveforms(c_spec->transform_y);
        contains_spec.transform_x = tmat_x.get();
        contains_spec.transform_y = tmat_y.get();
            //User transforms not allowed here. **And need to do this two-step process with MatCPtr and get()
            //so that get() doesn't operate on a temporary--leads to disaster

        //Now run do the calculation
        ContainsMgr mgr(contains_spec, vcluster);
        mgr.StartThreads();
        mgr.Wait();
    }

    //Now go through the polylines, setting every excluded element to wv::REJECT_CLUSTER.
    //Polylines are used only to 'trim' the set of accepted waveforms; only wv::ACCEPT_CLUSTER
    //waveforms get rejected.
    for (int p=0; p<num_polylines; ++p)
    {
        const PLSpec* pl_spec = polylines.at(p);
        const Polyline* polyline = pl_spec->GetDrawLine();
        int sample_x = pl_spec->sample_x,
                sample_y = pl_spec->sample_y;
        MatCPtr tmat_x = wave_data->Waveforms(pl_spec->transform_x),
               tmat_y = wave_data->Waveforms(pl_spec->transform_y);
        const wv_matrix* transform_x = tmat_x.get(),
               * transform_y = tmat_y.get();
            //User transforms not allowed here!
        Polyline::REJECT_DIR reject_dir = polyline->RejectDir();

        bool time_restricted = (pl_spec->time_begin!=0 || pl_spec->time_end!=-1);
        wv::float_t time_begin = pl_spec->time_begin,
                time_end = pl_spec->time_end;
        for (int i=0; i<num_records; ++i)
        {
            if (wv_vector_int_get(vcluster, i)!=wv::ACCEPT_CLUSTER) continue;
            if (time_restricted)
            {
                wv::float_t rec_time = wv_vector_get(timestamps,i);
                if (rec_time<time_begin || rec_time>time_end)
                    continue;
            }

            wv::float_t xval = wv_matrix_get(transform_x, i, sample_x),
                   polyline_val = polyline->Value(xval);
            if (polyline_val == std::numeric_limits<wv::float_t>::quiet_NaN()) continue; //E.g. if the x value is out of range
            wv::float_t yval = wv_matrix_get(transform_y, i, sample_y);

            if ( (reject_dir==Polyline::REJECT_UP && yval>polyline_val)
                 || (reject_dir==Polyline::REJECT_DOWN && yval<polyline_val) )
                wv_vector_int_set(vcluster, i, wv::REJECT_CLUSTER);
        }
//        emit SetPBarValue(100.0 * p / (wv::float_t)num_polylines);
    }

    //Now toggle all of the waves that have been manually toggled
    for (int i=0; i<num_records; ++i)
        if (_toggledWaves.test(i))
        {
            if (wv_vector_int_get(vcluster, i) == wv::ACCEPT_CLUSTER)
                wv_vector_int_set(vcluster, i, wv::REJECT_CLUSTER);
            else
                wv_vector_int_set(vcluster, i, wv::ACCEPT_CLUSTER);
        }

    //Divide up non-accepted waveforms based on proximity to the centers.  Only waveforms that
    //belong to ZERO_CLUSTER--not those which have been explicitly accepted or rejected--are
    //divided up.
    if (num_centers == 1)
    {
        for (int i=0; i<num_centers; ++i)
            if (wv_vector_int_get(vcluster, i)==wv::ZERO_CLUSTER)
                wv_vector_int_set(vcluster, i, wv::DEFAULT_CLUSTER);
    }
    else if (num_centers > 1)
    {
        //Normalize waveform ranges;
        wv::TRANSFORM transform_x = centers.at(0)->transform_x,
            transform_y = centers.at(0)->transform_y;
        int sample_x = centers.at(0)->sample_x,
            sample_y = centers.at(0)->sample_y;
        MatCPtr mat_x = WData()->Waveforms(transform_x),
            mat_y = WData()->Waveforms(transform_y);

        wv_vector* xnorm = wv_vector_alloc(num_records),
                * ynorm = wv_vector_alloc(num_records);
        wv_vector_const_view xdata = wv_matrix_const_column(mat_x.get(), sample_x),
                ydata = wv_matrix_const_column(mat_y.get(), sample_y);
        wv_vector_memcpy(xnorm, &xdata.vector);
        wv_vector_memcpy(ynorm, &ydata.vector);
        wv::float_t xmin, xmax, ymin, ymax;
        wv_vector_minmax(xnorm, &xmin, &xmax);
        wv_vector_minmax(ynorm, &ymin, &ymax);
        xmax -= xmin;
        ymax -= ymin;
        wv_vector_add_constant(xnorm, -xmin);
        wv_vector_scale(xnorm, 1.0/xmax);
        wv_vector_add_constant(ynorm, -ymin);
        wv_vector_scale(ynorm, 1.0/ymax);

        //Ultimately, dists will hold the distances for each record x cell, and dy
        //serves essentially as a temporary matrix for computation
        wv_matrix* dists = wv_matrix_alloc(num_records, num_centers),
                * dys = wv_matrix_alloc(num_records, num_centers); //Plural of dy

        //Compute the matrix of distances
        for (int c=0; c<num_centers; ++c)
        {
            //Initialize the matrix
            wv_vector_view dist_col = wv_matrix_column(dists, c),
                    dy_col = wv_matrix_column(dys, c);
            wv_vector* dist = &dist_col.vector,
                    * dy = &dy_col.vector;
            wv_vector_memcpy(dist, xnorm);
            wv_vector_memcpy(dy, ynorm);

            //Centers, declare them for legibility
            wv::float_t cx = (centers.at(c)->val_x - xmin) / xmax,
                    cy = (centers.at(c)->val_y - ymin) / ymax;

            //Calculate x-c, y-c, square, sum to get dist^2 (which is what will be
            //minimized)
            wv_vector_add_constant(dist, -cx);
            wv_vector_add_constant(dy, -cy);
            wv_vector_mul(dist, dist);
            wv_vector_mul(dy, dy);
            wv_vector_add(dist, dy);
        }

        //Go through each row and find the minimum; that's the nearest center
        for (int i=0; i<num_records; ++i)
        {
            if ( wv_vector_int_get(vcluster, i)!=wv::ZERO_CLUSTER ) continue;
            wv_vector_view row_view = wv_matrix_row(dists, i);
            int idx = wv_vector_min_index(&row_view.vector);
            wv_vector_int_set(vcluster, i, wv::DEFAULT_CLUSTER + idx);
        }

        wv_matrix_free(dists);
        wv_matrix_free(dys);
    }

    //Write the cluster file.  For the sake of consistency code/simplicity, don't manually construct
    //the classification, even though it uses the available info.
    std::vector<ConstClustering> clusterings;
    clusterings.push_back( ConstClustering(0.0, vcluster) );
    std::string mclu_file_name = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(),
                                                  Transform(), ID(), "mclu", Electrode());
    MCluFileOut mclu_file_out(mclu_file_name, 0, clusterings);
    mclu_file_out.Save();
    mclu_file_out.Close();
    wv_vector_int_free(vcluster);

    //Get the classification and return results, namely the classification
    LoadClassification(QProcess::NormalExit);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

void ManualTO::SetWindow(int, int)
{
    ClusMethodTOBase::SetWindow(0, WData()->NumRecords());  //Manual clustering
    //ignores the selection window in the hist. widget
}
