#include "spc_to.h"

SPC_TO::SPC_TO(wv::CLUSTER_ID cid, wv::TRANSFORM transform, MatCPtr waves, const std::string& wfm_file)
    : ClusMethodTOBase(cid, transform, waves, waves->size1, wfm_file)
{
//    MakeProgressBar(Name());
}

void SPC_TO::Start()
{
    StartTimer();

    //Write the parameter file for SPC
    std::string pf_name =  InputFullPath() + "." + std::to_string( ID() ) + ".run";
    QFile param_file(pf_name.c_str());
    param_file.open(QFile::WriteOnly);

    //User choice for "Temperature" dictates how to calculate it
    std::string temp = Params()->Par("Temperature");
    bool is_auto_temp = temp == "auto";

        //Parameters not accessible to the user nor knowable in advance
    std::string winb = "window_begin: " + std::to_string( WindowBegin() ) + "\n",
            wine = "window_end: " + std::to_string( WindowEnd() ) + "\n",
            nump = "NumberOfPoints: " + std::to_string( Waves()->size1 ) + "\n",
            df = "DataFile: " + InputFullPath() + "\n",
            of = "OutFile: " + InputFullPath() + "." + std::to_string( ID() ) + "\n";
    param_file.write( winb.c_str() );
    param_file.write( wine.c_str() );
    param_file.write( nump.c_str() );
    param_file.write( df.c_str() );
    param_file.write( of.c_str() );
    if (!is_auto_temp)
    {
        std::string mint = "MinTemp: " + temp + "\n",
                maxt = "MaxTemp: " + temp + "\n",
                step = "TempStep: 1.0\n";
        param_file.write( mint.c_str() );
        param_file.write( maxt.c_str() );
        param_file.write( step.c_str() );
    }

        //Parameters from the parameter menu
    int num_pars = Params()->NumPars();
    for (int i=0; i<num_pars; ++i)
    {
        std::string key = Params()->Key(i),
                value = Params()->Value(i),
                param;
        if (value=="|" || value=="~")
            param = key + value + "\n";
        else
        {
            if (!is_auto_temp &&
                    (key=="MinTemp" || key=="MaxTemp" || key=="TempStep"))
                continue;
            param = key + ": " + value + "\n";
        }
        param_file.write( param.c_str() );
    }
    param_file.close();

    //Now cluster, giving SPC the name of the file containing the parameters
    StringList proc_args;
    proc_args << pf_name;
    QProcess* clus_proc = new QProcess(); //Must create here, so that it is in this thread
    Cluster(clus_proc, proc_args);
}
/*

"In our implementation, we varied the temperature from 0 to 0.2 in increments of 0.01 and looked for the highest temperature at which a cluster containing more than 60 points appeared (not being present at lower
temperatures). Since our simulations were 60 sec long, this means that we
considered clusters corresponding to neurons with a mean ﬁring rate of at
least 1 Hz. The threshold of 1 Hz gave us optimal results for all our simulations, but it should be decreased if one considers neurons with lower
ﬁring rates. Alternatively, one can consider a fraction of the total number
of spikes. If no cluster with a minimum of 60 points was found, we kept
the minimum temperature value. Using this criterion, we can automatically
select the optimal temperature for cluster assignments, and therefore the
whole clustering procedure becomes unsupervised."
*/
void SPC_TO::LoadClassification(int exit_code)
{
    try
    {
        //Read the SPC output file into memory
        SPCFileIn spc_file(InputFullPath() + "." + std::to_string(ID()));
        int num_records = spc_file.Load();
        if (num_records == 0) return;
        wv::float_t pct = boost::lexical_cast<wv::float_t>( Params()->Par("MinClusterSizePct") );
        int min_clus_size = pct * (wv::float_t)num_records / 100.0 + 0.5;
        int max_poss_clusters = boost::lexical_cast<int>( Params()->Par("MaxPossibleClusters") ),
                clus_size_thresh = (wv::float_t)num_records / (wv::float_t)max_poss_clusters + 0.5;

        //Get initial clustering after deriving the optimal temperature
        const wv_matrix_int* cluster_counts = spc_file.ClusterCounts();
        int best_temp_index = TemperatureIndex(cluster_counts, min_clus_size, clus_size_thresh);
        wv::float_t best_temp = spc_file.Temperature(best_temp_index);
        Log( "SPC: Best temperature is " + std::to_string(best_temp)
                         + " for minimum cluster size of " + std::to_string(min_clus_size) + ".");

        //Now, for each temperature, weed out the classes that are too
        //small and find the max class, returning a new cluster vector.  Otherwise, put
        //everything into one big cluster.
        std::vector<ConstClustering> clusterings;
        int num_clusterings = cluster_counts->size1;
        for (int i=0; i<num_clusterings; ++i)
        {
            const wv_vector_int* vclus = spc_file.VCluster(i);
            wv_vector_int* new_vclus = wv_vector_int_alloc(num_records);
            RefineClusters(new_vclus, vclus);

            wv::float_t level = spc_file.Temperature(i);
            ConstClustering clus_pair(level, new_vclus);
            clusterings.push_back(clus_pair);
        }

        //Write the SPC output to a .clu file
        std::string mclu_file_name = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(), Transform(), ID(), "mclu", Electrode());
        MCluFileOut mclu_file(mclu_file_name, best_temp_index, clusterings);
        mclu_file.Save();
        mclu_file.Close(); //Make sure to close before reading this in with LoadClassification!

        for (int i=0; i<num_clusterings; ++i)
            wv_vector_int_free( const_cast<wv_vector_int*>(clusterings.at(i).second) );
            //These were created non-const, but clusterings is <double, const...> so they can't be
            //taken back out for deletion without casting them back to non-const

        //Then read the .clu file in the normal way
        ClusMethodTOBase::LoadClassification(exit_code);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

//This follows Quian-Quiroga's find_temp.m function, except that in addition, at least one
//cluster has to be sufficiently large
int SPC_TO::TemperatureIndex(const wv_matrix_int* cluster_counts, int min_clus_size, int clus_size_thresh)
{
    int temperature = 0,
            num_dtemps = cluster_counts->size1;
    int max_poss_clusters = boost::lexical_cast<int>( Params()->Par("MaxPossibleClusters") );

    for (int i=1; i<num_dtemps; ++i)
    {
        //See if a new cluster has emerged at this temperature
        bool found_new = false;
        for (int j=0; j<max_poss_clusters; ++j)
        {
            int ct_before = wv_matrix_int_get(cluster_counts, i-1, j),
                    ct_now = wv_matrix_int_get(cluster_counts, i, j);
            found_new |= (ct_now - ct_before > min_clus_size);
        }

        //If there is a new cluster, see if there is at least one cluster which exceeds the
        //size threshold
        if (found_new)
        {
            bool is_large_cluster = false;
            for (int j=0; j<max_poss_clusters; ++j)
            {
                int ct_now = wv_matrix_int_get(cluster_counts, i, j);
                is_large_cluster |= ct_now > clus_size_thresh;
            }
            found_new &= is_large_cluster;
        }
        if (found_new)
            temperature = i;
    }
    return temperature;
}
int SPC_TO::RefineClusters(wv_vector_int* new_vclus, const wv_vector_int* vclus)
{
    int num_records = vclus->size;
    wv::float_t pct = boost::lexical_cast<wv::float_t>( Params()->Par("MinClusterSizePct") );
    int clus_size_thresh = num_records * pct / 100.0 + 0.5;

    //Get the number of records in each cluster
    std::map<int, int> clus_cts;
    auto clus_cts_end = clus_cts.end();
    for (int i=0; i<num_records; ++i)
    {
        int c = wv_vector_int_get(vclus, i);
        if (clus_cts.find(c) == clus_cts_end)
            clus_cts[c] = 1;
        else
            ++clus_cts[c];
    }

    //Reassign clusters with too few elements to the rejection cluster
    int max_class = INT_MIN;
    for (int i=0; i<num_records; ++i)
    {
        int oldc = wv_vector_int_get(vclus, i);
        int newc = (clus_cts[oldc] >= clus_size_thresh) ? oldc : wv::REJECT_CLUSTER;
        if (newc>max_class) max_class = newc;
        wv_vector_int_set(new_vclus, i, newc);
    }
    //### This is pretty good but the last thing to do here is to assign waveforms from tiny clusters
    //to the nearest large cluster, *unless* the height(?) differs by >2SD(?), in which case they are
    //put into the rejection cluster.

    return max_class;
}
