#include "splineworkerto.h"

SplineWorkerTO::SplineWorkerTO(int thd_num, SplineMgr::SplineSpec spec) : WorkerTO(thd_num), _spec(spec)
{
}

void SplineWorkerTO::Start()
{
    //Unpack SplineSpec, for convenience/speed
    const wv_matrix* waves = _spec.waves.get();
    int num_samples = waves->size2;
    int i_start = _spec.i_start,
            i_end = _spec.i_end,
            left_buf = _spec.left_buf,
            right_buf = _spec.right_buf,
            spline_order = _spec.spline_order,
            num_breakpts = _spec.num_breakpts;
    wv::float_t grain = _spec.grain;

    //These vectors hold interpolated values for fine-grained calculation of height, width
    wv_vector* lmax_vec = wv_vector_alloc(grain),
            * rmax_vec = wv_vector_alloc(grain),
            * min_vec = wv_vector_alloc(grain);

    //Allocate for b-splines
    gsl_bspline_workspace* bspline_ws = gsl_bspline_alloc(spline_order, num_breakpts);
    size_t ncoeffs = gsl_bspline_ncoeffs(bspline_ws);
    wv_vector* B = wv_vector_alloc(ncoeffs);
    wv_matrix* X = wv_matrix_alloc(num_samples, ncoeffs);

    //Allocate for workspace
    gsl_multifit_linear_workspace* multifit_ws = gsl_multifit_linear_alloc(num_samples, ncoeffs);
    wv_matrix* cov = wv_matrix_alloc(ncoeffs, ncoeffs);
    wv_vector* coeffs = wv_vector_alloc(ncoeffs);
    wv::float_t chisq;

    wv_matrix* spline = _spec.spline,
            * height = _spec.height,
            * width = _spec.width,
            * slope = _spec.slope;

    //Get spline for each record
    for (int i=i_start; i<i_end; ++i)
    {
        wv_vector_const_view row = wv_matrix_const_row(waves, i);

        gsl_bspline_knots_uniform(0, num_samples, bspline_ws);

        //Now set up the coefficients matrix
        for (int j=0; j<num_samples; ++j)
        {
            gsl_bspline_eval(j, B, bspline_ws);

            for (size_t k=0; k<ncoeffs; ++k)
            {
                wv::float_t Bk = wv_vector_get(B, k);
                wv_matrix_set(X, j, k, Bk);
            }
        }

        //Get the minimum
        wv_vector_const_view row_center = wv_vector_const_subvector(&row.vector, left_buf, num_samples - (left_buf+right_buf));
        size_t imin = wv_vector_min_index(&row_center.vector);
        imin += left_buf;

        //Get the maximum which lies to the left of the minimum
        wv_vector_const_view left = wv_vector_const_subvector(&row.vector, 1, imin);
        size_t left_max = wv_vector_max_index(&left.vector) + 1;

        //Get the maximum which lies to the right of the minimum
        wv_vector_const_view right = wv_vector_const_subvector(&row.vector, imin+1, num_samples-imin-1);
        size_t right_max = wv_vector_max_index(&right.vector) + imin + 1;

        //The linear best fit.  There is also a version with weighted coefficients, but that doesn't seem
        //applicable here.
        gsl_multifit_linear(X, &row.vector, coeffs, cov, &chisq, multifit_ws);
        wv::float_t err; //Not used

        //Now get the actual interpolated values and put them into a new vector
        for (int j=0; j<num_samples; ++j)
        {
            wv::float_t new_val;
            gsl_bspline_eval(j, B, bspline_ws);
            gsl_multifit_linear_est(B, coeffs, cov, &new_val, &err);
            wv_matrix_set(spline, i, j, new_val);
        }

        //Now get height and width.  Width is calculated simplistically by finding peak-to-peak distance.
        //Finding the halfway point between the min and
        //each peak, and taking the distance between them is probably better
        //  NB slope-D is found in the simplistic way, without interpolating.
        wv::float_t step = 2.0/(wv::float_t)(grain-1); //Equivalent to sampling within +/- 1 sample about the max/min
        wv::float_t lmax_start = left_max - 1,
                rmax_start = right_max - 1,
                min_start = imin - 1;
        wv::float_t lmax_j, rmax_j, min_j; //Since these are sent by address it's a waste to redeclare them
        std::string s;
        for (int j=0; j<grain; ++j)
        {
            wv::float_t step_j = (wv::float_t)j * step;

            gsl_bspline_eval(lmax_start+step_j, B, bspline_ws);
            gsl_multifit_linear_est(B, coeffs, cov, &lmax_j, &err);
            wv_vector_set(lmax_vec, j, lmax_j);

            gsl_bspline_eval(rmax_start+step_j, B, bspline_ws);
            gsl_multifit_linear_est(B, coeffs, cov, &rmax_j, &err);
            wv_vector_set(rmax_vec, j, rmax_j);
            char cs[128];
            sprintf(cs, "%d, %f, %f\n", j, rmax_start+step_j, rmax_j);
            s += cs;

            gsl_bspline_eval(min_start+step_j, B, bspline_ws);
            gsl_multifit_linear_est(B, coeffs, cov, &min_j, &err);
            wv_vector_set(min_vec, j, min_j);
        }

        //Get indices and mins/maxes
        wv::float_t lmax_interp_idx = wv_vector_max_index(lmax_vec),

                rmax_interp_idx = wv_vector_max_index(rmax_vec),
                rmax_interp = wv_vector_get(rmax_vec, rmax_interp_idx),

                min_interp_idx = wv_vector_min_index(min_vec),
                min_interp = wv_vector_get(min_vec, min_interp_idx);

        //Calculate and set the actual values
        wv::float_t height_interp = rmax_interp - min_interp,
                rmax_abs_idx = rmax_start+rmax_interp_idx*step,
                width_interp = lmax_start*lmax_interp_idx*step + rmax_abs_idx,
                slope_interp = height_interp/(rmax_abs_idx - min_start+min_interp_idx*step);

        wv_matrix_set(height, i, 0, height_interp);
        wv_matrix_set(width, i, 0, width_interp);
        wv_matrix_set(slope, i, 0, slope_interp);
    }

    WorkFinished();
}
