#include "apmfile.h"

APMFile::APMFile(const std::string& full_path) : WaveformsInBase(full_path, true)
{
}

//This follows exactly as possible the file structure specifications in the APM Manual,
//pg. 53
int APMFile::Load()
{
    FILE* fp = File();

    int num_apm_records = 0,
            num_records = 0,
            num_samples = -1; //Don't know until we've read through.
    //First read-through gets the number of records and samples.
    std::vector<short> v_electrodes;
    v_electrodes.reserve(1024*1024);
    uint adc[1024];
    float sec_per_sample = 1.0, time_calib = 0.0;
    while ( !feof(fp) )
    {
        uint message_code;
        ushort unit_number, channel_number;
        uint num_words;

        fread(&message_code, sizeof(uint), 1, fp);
        fread(&unit_number, sizeof(ushort), 1, fp);
        fread(&channel_number, sizeof(ushort), 1, fp);
        fread(&num_words, sizeof(uint), 1, fp);

        ++num_apm_records;
        if (message_code==0 || message_code==1)
        {
            fread(adc, sizeof(uint), num_words, fp);
            AddElectrode(channel_number);
            v_electrodes.push_back(channel_number);
            ++num_records;
        }
        else if (message_code == 14)
            fread(&time_calib, sizeof(float), 1, fp);
        else if (message_code == 15)
            fread(&sec_per_sample, sizeof(float), 1, fp);
        else
            fread(adc, sizeof(uint), num_words, fp);

        if (num_records==1 && num_samples<0)
            num_samples = num_words - 1;  //The first word is the timestamp
    }
    SetNumSamples(num_samples);
    SetUsPerSample(1.0 / sec_per_sample);

    rewind(fp);

    MakeWaveData(v_electrodes, true);

    //Iterate through the entire list of trials and include them in the appropriate WaveData
    //object
    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    int num_electrodes = NumElectrodes();
    std::map<short, int> elc_ct;
    for (int i=0; i<num_electrodes; ++i) elc_ct[ v_electrodes.at(i) ] = 0;
    wv_vector* wave = wv_vector_alloc(num_samples);
    float voltage_scale = 1.0;
    for (int i=0; i<num_apm_records; ++i)
    {
        uint message_code;
        ushort unit_number, channel_number;
        uint num_words;

        fread(&message_code, sizeof(uint), 1, fp);
        fread(&unit_number, sizeof(ushort), 1, fp);
        fread(&channel_number, sizeof(ushort), 1, fp); //==electrode
        fread(&num_words, sizeof(uint), 1, fp);

        if (message_code == 15)
        {
            fread(&voltage_scale, sizeof(float), 1, fp);
            voltage_scale *= 1000000.0;
        }
        else if (message_code==0 || message_code==1)
        {
            WaveData* wave_data = (*mwdata)[channel_number];
            int ct = elc_ct[channel_number];

            //Timestamp
            uint timestamp;
            fread(&timestamp, sizeof(uint), 1, fp);
            wave_data->WriteTimestamp(timestamp);

            //Samples
            for (int j=0; j<num_samples; ++j)
            {
                short val, dummy;
                fread(&val, sizeof(short), 1, fp);
                wv_vector_set(wave, j, (wv::float_t)val * voltage_scale);
                fread(&dummy, sizeof(short), 1, fp);
            }
            wv_vector_view wave_view = wv_vector_view_array(wave->data, wave->size);
            wave_data->WriteWaveform(wave_view);

            //Online clustering
            wave_data->SetOnlineClus(ct, message_code==1 ? unit_number : wv::REJECT_CLUSTER);

            elc_ct[channel_number]++;
        }
        else
            fread(adc, sizeof(short), num_words*2, fp);
    }
    wv_vector_free(wave);
    FinishedLoading();

    return -1; // ###
}
