#include "asciifilein.h"

AsciiFileIn::AsciiFileIn(const std::string& full_path) : WaveformsInBase(full_path, false)
{
}

/*The format expected here has # of records, # of rows in the first line, then the spike data
  , one record per row, with each element separated by whitespace.  Eventually there will be
  an optional 'et' parameter to the first line, indicating the presence of electrode number
  timestamp, or both.*/
int AsciiFileIn::Load()
{
    FILE* fp = File();
    QFile qfp;
    qfp.open(fp, QIODevice::ReadOnly);

    //Header
    char ws[] = { '\t', ' ' };
    std::string line = qfp.readLine().data();
    line.erase(std::remove(line.begin(), line.end(), '\n'), line.end());
    StringVec specs;
    boost::algorithm::split(specs, line, boost::algorithm::is_any_of(ws),
                            boost::algorithm::token_compress_on);
    int num_records = boost::lexical_cast<int>(specs.at(0)),
            num_samples = boost::lexical_cast<int>(specs.at(1));
    SetNumSamples(num_samples);
    bool has_timestamps(false), has_electrodes(false);
    if (specs.size() > 2)
    {
        std::string s = specs.at(2);
        if (ut::Contains(s, 'e')) has_electrodes = true;
        if (ut::Contains(s, 't')) has_timestamps = true;
    }

    //Get the data
    wv_matrix* waves = wv_matrix_alloc(num_records, num_samples);
    wv_vector* timestamps = wv_vector_alloc(num_records);
    std::vector<short> v_electrodes;
    v_electrodes.reserve(num_records);
    for (int i=0; i<num_records; ++i)
    {
        std::string line = qfp.readLine().data();
        line.erase(std::remove(line.begin(), line.end(), '\n'), line.end());
        StringVec rec;
        boost::algorithm::split(rec, line,  boost::algorithm::is_any_of(ws),
                                boost::algorithm::token_compress_on);
        int ct=0; //For electrode, timestamp
        if (has_electrodes)
        {
            unsigned char val = boost::lexical_cast<int>( rec.at(ct) );
            v_electrodes.push_back(val);
            AddElectrode(val);
            ++ct;
        }
        else
        {
            AddElectrode(0);
            v_electrodes.push_back(0);
        }
        if (has_timestamps)
        {
            wv::float_t val = (wv::float_t)boost::lexical_cast<int>( rec.at(ct) );
            wv_vector_set(timestamps, i, val);
            ++ct;
        }
        else
            wv_vector_set(timestamps, i, i);
        for (int j=0; j<num_samples; ++j)
        {
            wv::float_t val = boost::lexical_cast<wv::float_t>( rec.at(j+ct) );
            wv_matrix_set(waves, i, j, val);
        }
    }

    MakeWaveData(v_electrodes, false);

    //Iterate through the entire list of trials and include them in the appropriate WaveData
    //object
    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    int num_electrodes = NumElectrodes();
    int* wave_ct = new int[num_electrodes];
    memset(wave_ct, 0, num_electrodes*sizeof(int));
    for (int i=0; i<num_records; ++i)
    {
        short elc = v_electrodes.at(i);
        WaveData* wave_data = (*mwdata)[elc];
        wave_data->WriteWaveform( wv_matrix_row(waves, i) );
        wave_data->WriteTimestamp(wv_vector_get(timestamps, i));
        wave_ct[elc]++;
    }

    //Free all temporary objects
    wv_matrix_free(waves);
    wv_vector_free(timestamps);

    FinishedLoading();

    return -1; // ###
}
