#include "mexfile.h"

const int MexFile::HEADER_SIZE = 512;
const int MexFile::MAX_WAVE_LEN = 512;
MexFile::MexFile(const std::string& full_path) : WaveformsInBase(full_path, true),
    _variables("TEFL")
{
}

int MexFile::Load()
{
    int num_records = ReadHeader();
//            num_records_used = 0;
    if (num_records>0)
    {
        SetData( new std::map<short, WaveData*>() );
        ReadWaveforms();
    }
    return -1; //###
}
int MexFile::ReadHeader()
{
    char* header = (char*)malloc(HEADER_SIZE*sizeof(char));
    int bytes_read = DS().readRawData(header, HEADER_SIZE);
    if (bytes_read!=HEADER_SIZE)
        throw BadValue("MexFile::ReadHeader", "Wrong number of bytes read from file for header.");
    _header = header;
    Log(_header);

    StringVec header_items = ut::Split<StringVec>(_header, "\n");

    std::string data_type = header_items.at(0);
    if (data_type != "WAVEFORM FILE")
        throw BadValue("MexFile::ReadHeader", "File data type " + data_type + " not recognized.");
    _numRecordBytes = boost::lexical_cast<int>( ut::Split<StringVec>(header_items.at(1), " ").back() );
    SetNumSamples( boost::lexical_cast<int>( ut::Split<StringVec>(header_items.at(2), " ").back() ) );
    _adZero = boost::lexical_cast<int>( ut::Split<StringVec>(header_items.at(3), " ").back() );
    _adMin = boost::lexical_cast<int>( ut::Split<StringVec>(header_items.at(4), " ").back() );
    _adMax = boost::lexical_cast<int>( ut::Split<StringVec>(header_items.at(5), " ").back() );

    _isTESFL = false;
    if (header_items.size()>6)
    {
        std::string variables = header_items.at(6);
        StringList varlist = ut::Split<StringList>(variables, " ");
        if ( ut::TakeFirst(varlist) == "Variables")
        {
            _isTESFL = std::find(varlist.begin(), varlist.end(), "spike")!=varlist.end();
            if (_isTESFL) _variables = "TESFL";
        }
        else
            Log("WARNING: Unexpected header information, " + variables + "; data file probably corrupted");
    }
    else
        Log("WARNING: No variable information provided, using " + _variables);

    _numRecordsTotal = (QFileInfo(Path().c_str()).size() - HEADER_SIZE) / _numRecordBytes;

    Log( std::to_string(_numRecordsTotal) + " records found." );
    return _numRecordsTotal;
}
void MexFile::ReadWaveforms()
{
    FILE* file = File();
    fseek(file, HEADER_SIZE, SEEK_SET);
        //The file pointer gets corrupted or something after the use of the QDataStream
        //without the reset it's in the 16000s, even though bytes_read above is 512 like
        //it should be

    int num_samples = NumSamples();
    short wave[MAX_WAVE_LEN];

    //First we get the electrodes by record
    std::vector<short> v_electrodes;
    std::vector<MexRecord> records;
    for (int i=0; i<_numRecordsTotal; ++i)
    {
        uint timestamp;
        short electrode, rex_flag, online_clus, wave_length;

        fread(&timestamp, sizeof(uint), 1, file);
        fread(&electrode, sizeof(short), 1, file);
        if (_isTESFL) fread(&online_clus, sizeof(short), 1, file);
        fread(&rex_flag, sizeof(short), 1, file);
        fread(&wave_length, sizeof(short), 1, file); //Unnecessary
        fread(wave, sizeof(short), wave_length, file);

        MexRecord record =
        {
            timestamp,
            electrode,
            online_clus,
            rex_flag,
            wave_length,
            new short[wave_length]
        };
        memcpy(record.wave, wave, wave_length*sizeof(short));

        AddElectrode(electrode);
        v_electrodes.push_back(electrode);
        records.push_back(record);
    }

    MakeWaveData(v_electrodes, _isTESFL);
    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );

    std::map<short, int> elc_ct;
    for (int i=0; i<NumElectrodes(); ++i) elc_ct[ v_electrodes.at(i) ] = 0;
    for (auto it : records)
    {
        WaveData* wave_data = (*mwdata)[ it.electrode ];
        wave_data->WriteTimestamp(it.timestamp);
        wave_data->WriteWaveform(it.wave, num_samples);
        delete it.wave;

        if (_isTESFL) wave_data->SetOnlineClus(elc_ct[it.electrode]++,
                                               (int)it.online_clus+1);
    }
    FinishedLoading();
}
