#include "nevfile.h"

const int NevFile::NS_PID_COMMENT = 0xFFFF;
const int NevFile::NS_PID_CONFLOG = 0xFFFB;

NevFile::NevFile(const std::string& full_path) : WaveformsInBase(full_path, true)
{
}

int NevFile::Load()
{
    FILE* fp = File();

    //Section 1: Header
    ns_NEURALEV header;
    fread(&header, sizeof(ns_NEURALEV), 1, fp);
    Log(Name() + " header information:\n"
        + "FileTypeId: " + std::string(header.FileTypeId) + "\n"
        + "FileSpec: " + std::to_string(header.FileSpec[0]) + " " + std::to_string(header.FileSpec[1]) + "\n"
        + "AdditionalFlags: " + std::to_string(header.AdditionalFlags) + "\n"
        + "BytesInHeaders: " + std::to_string(header.BytesInHeaders) + "\n"
        + "BytesInDataPackets: " + std::to_string(header.BytesInDataPackets) + "\n"
        + "TimeResolutionOfTimeStamps: "
            + std::to_string(header.TimeResolutionOfTimeStamps) + "\n"
        + "TimeResolutionOfSamples: "
            + std::to_string(header.TimeResolutionOfSamples) + "\n"
//        + "TimeOrigin: " + std::to_string(header.TimeOrigin) + "\n"
        + "ApplicationToCreateFile: " + std::string(header.ApplicationToCreateFile) + "\n"
        + "CommentField: " + std::string(header.CommentField) + "\n"
        + "NumberofExtendedHeaders: " + std::to_string(header.NumberofExtendedHeaders) + "\n"
        );

    if (header.AdditionalFlags % 2 != 1)
        throw WIOException("", Name(), "Sorry, variable-length waveform events currently not supported.");

    //Section 2: Extended headers
    int num_extended_headers = header.NumberofExtendedHeaders;
    for (int i=0; i<num_extended_headers; ++i)
    {
        ns_EXTENDED_HEADER extended_header;
        fread(&extended_header, sizeof(ns_EXTENDED_HEADER), 1, fp);

        _extHeaders.push_back(extended_header);
    }

    //Section 3: Data Packets
    int packet_width = header.BytesInDataPackets,
            skip = packet_width - 6, //When we're skipping over data
            num_nev_records = 0,
            num_samples = (packet_width - 8)/2; //Default is 16-bit samples
    SetNumSamples(num_samples);
    SetUsPerSample( 1000000.0 / header.TimeResolutionOfTimeStamps );

        //Set up progress bar for initial scan
    QFile qf( Path().c_str() );
    int file_sz = qf.size();
    _progressBar->setMaximum(file_sz/packet_width);
    _progressBar->show();
    _progressBar->setWindowTitle("Scanning file");

        //First run through the data and get unit # for each record
    std::vector<short> v_electrodes;
    int data_skip = skip - 1; //-1 because we've read the unit number
    while (!feof(fp))
    {
        uint timestamp;
        ushort packet_id; //channel

        int sz = fread(&timestamp, sizeof(uint), 1, fp); if (sz!=1) break;
        sz = fread(&packet_id, sizeof(ushort), 1, fp); if (sz!=1) break;
            //Going by feof doesn't seem to be working

        ++num_nev_records;
        if (packet_id>0 && packet_id<2048)
        {
            uchar dummy; //cell, not used now
            fread(&dummy, sizeof(uchar), 1, fp);
            AddElectrode(packet_id);
            v_electrodes.push_back(packet_id);

            fseek(fp, data_skip, SEEK_CUR);
        }
        else
            fseek(fp, skip, SEEK_CUR);

        if (num_nev_records%100000==0)
        {
            _progressBar->setValue(num_nev_records);
            _progressBar->update();
            QApplication::processEvents();
        }
    }

    MakeWaveData(v_electrodes, true); //Sets up the raw, timestamp files

    fseek(fp, header.BytesInHeaders, SEEK_SET); //Return file pointer to beginning of file

    //Set up progress bar for the file read-in
    QTime tic;
    tic.start();
    _progressBar->setMaximum(num_nev_records);
    _progressBar->show();
    _progressBar->setWindowTitle("Creating data structures");
    QRect geom = _progressBar->geometry();
    geom.setWidth(300);// _progressBar->sizeHint().width() );
    _progressBar->setGeometry(geom);

    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    std::map<short, int> elc_ct;
    for (int i=0; i<NumElectrodes(); ++i) elc_ct[ v_electrodes.at(i) ] = 0;
    wv_vector* wave = wv_vector_alloc(num_samples);
    for (int i=0; i<num_nev_records; ++i)
    {
        uint timestamp;
        ushort packet_id; //electrode

        fread(&timestamp, sizeof(uint), 1, fp);
        fread(&packet_id, sizeof(ushort), 1, fp);

        if (packet_id>0 && packet_id<2048)
        {
            uchar unit_number, dummy; //cell
            fread(&unit_number, sizeof(uchar), 1, fp);
            fread(&dummy, sizeof(uchar), 1, fp);

            WaveData* wave_data = (*mwdata)[packet_id];
            int ct = elc_ct[packet_id];

            wave_data->WriteTimestamp(timestamp);

            for (int j=0; j<num_samples; ++j)
            {
                short val;
                fread(&val, sizeof(short), 1, fp);
                wv_vector_set(wave, j, (wv::float_t)val);
            }
            wv_vector_view wave_view = wv_vector_view_array(wave->data, wave->size);
            wave_data->WriteWaveform(wave_view);
            wave_data->SetOnlineClus(ct, unit_number);

            elc_ct[packet_id]++;
        }
        else if (packet_id == NS_PID_COMMENT)
        {
            uchar char_set, reserved;
            fread(&char_set, sizeof(uchar), 1, fp);
            fread(&reserved, sizeof(uchar), 1, fp);
            uint color;
            fread(&color, sizeof(uint), 1, fp);
            char comment[256];
            fread(comment, sizeof(char), packet_width-12, fp);

            Log("COMMENT: Timestamp: " + std::to_string(timestamp)
                + ", comment: " + std::string(comment));
        }
        else if (packet_id == NS_PID_CONFLOG)
        {
            ushort change_type;
            fread(&change_type, sizeof(ushort), 1, fp);
            char change[256];
            fread(change, sizeof(char), packet_width-8, fp);

            Log("COMMENT: Timestamp: " + std::to_string(timestamp)
                + ", change type: " + std::to_string(change_type)
                + ", change: " + std::string(change));
        }
        else
            fseek(fp, skip, SEEK_CUR);

        if (i%100000==0)
        {
            _progressBar->setValue(i);
            _progressBar->update();
            QApplication::processEvents();
        }
    }
    wv_vector_free(wave);
    FinishedLoading();
    _progressBar->close();

    return -1; // ###
}
