#include "plxfile.h"

const int plx::PL_SingleWFType = 1,
        plx::PL_ExtEventType = 4,
        plx::PL_ADDataType = 5;

PlxFile::PlxFile(const std::string& full_path) : WaveformsInBase(full_path, true)
{
}

int PlxFile::Load()
{
    FILE* fp = File();

    //Get file header
    fread(&_fileHeader, sizeof(plx::PL_FileHeader), 1, fp);

    //DSP channel headers
    int num_dsp_channels = _fileHeader.NumDSPChannels;
    for (int i=0; i<num_dsp_channels; ++i)
    {
        plx::PL_ChanHeader chan_header;
        fread(&chan_header, sizeof(plx::PL_ChanHeader), 1, fp);
        _chanHeaders.push_back(chan_header);
    }

    //Event headers
    int num_event_channels = _fileHeader.NumEventChannels;
    for (int i=0; i<num_event_channels; ++i)
    {
        plx::PL_EventHeader event_header;
        fread(&event_header, sizeof(plx::PL_EventHeader), 1, fp);
        _eventHeaders.push_back(event_header);
    }

    //Slow channel headers
    int num_slow_channels = _fileHeader.NumSlowChannels;
    for (int i=0; i<num_slow_channels; ++i)
    {
        plx::PL_SlowChannelHeader slow_channel_header;
        fread(&slow_channel_header, sizeof(plx::PL_SlowChannelHeader), 1, fp);
        _slowChannelHeaders.push_back(slow_channel_header);
    }

    int num_samples = _fileHeader.NumPointsWave;
    std::vector<short> v_electrodes;
    std::vector<short*> records;
    std::vector<wv::float_t> timestamps;
    std::vector<int> online_vclus;
    while (!feof(fp))
    {
        plx::PL_DataBlockHeader data_block_header;
        fread(&data_block_header, sizeof(plx::PL_DataBlockHeader), 1, fp);

        if (data_block_header.Type == plx::PL_ExtEventType) continue;
        if (data_block_header.NumberOfWaveforms != 1) continue;

        //Online clustering, 0 for unsorted
        int clus = data_block_header.Unit;

        //Get timestamp, stored with 40 bit precision in the file
        unsigned int ts_lower = data_block_header.TimeStamp;
        unsigned short ts_upper = data_block_header.UpperByteOf5ByteTimestamp;
        long long int ts = ((static_cast<long long int>(ts_upper)<<32) + static_cast<long long int>(ts_lower));
        wv::float_t timestamp = 1000.0 * (double)ts / (double)_fileHeader.ADFrequency;

        //Get the samples, ignore continuous data
        int num_samples_now = data_block_header.NumberOfWordsInWaveform;
        short* record = new short[num_samples_now];
        fread(record, sizeof(short), num_samples_now, fp);
        if (data_block_header.Type == plx::PL_ADDataType)
        {
            delete[] record;
            continue;
        }

        short elc = data_block_header.Channel;
        AddElectrode(elc);
        v_electrodes.push_back((short)elc);

        timestamps.push_back(timestamp);
        records.push_back(record);
        online_vclus.push_back(clus);
    }
    int num_records_all = records.size();

    SetUsPerSample( 1000000.0 / _fileHeader.WaveformFreq );
    SetNumSamples(num_samples);

    //Make the WaveData objects
    MakeWaveData(v_electrodes, true);

    //Assign each record, timestamp, and online cluster to its respective wave_data object
    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    std::map<short, int> elc_ct;
    for (int i=0; i<NumElectrodes(); ++i) elc_ct[ v_electrodes.at(i) ] = 0;
    for (int i=0; i<num_records_all; ++i)
    {
        short elc = v_electrodes.at(i);
        WaveData* wave_data = (*mwdata)[elc];
        int ct = elc_ct[elc];

        wave_data->WriteTimestamp(timestamps.at(i));

        short* record = records.at(i);
        wave_data->WriteWaveform(record, num_samples);
        delete[] record;

        wave_data->SetOnlineClus(ct, online_vclus.at(i));

        elc_ct[elc]++;
    }
    FinishedLoading();

    return -1; // ###
}

