#include "parfileout.h"
#include "../../memrep/clustermediator.h"

ParFileOut::ParFileOut(const std::string& file_name, const std::string& data_file_name, wv::CLUSTER_ID cid)
    : OutputFileBase(file_name, false, NULL, boost::dynamic_bitset<>()), _dataFileName(data_file_name),
      _cluSpec( ClusterMediator::Instance().GetCluSpecPtr(cid) ),
      _parameters( _cluSpec->ClusterParameters() )

    //Given the way is called, data_file_name can be extracted from file_name, but having them separate
    //is more flexible
{
    ClusterMediator& cmed = ClusterMediator::Instance();
    _data = cmed.GetWaveData(cid);
}

void ParFileOut::Save() const
{
    std::string active_transform = wv::TransformName( _cluSpec->Transform() ),
            cluster_method = wv::ClusterName( _cluSpec->ClusterMethod() );
    FILE* fp = File();

    //Basic features of the analysis
    char version[256],
            data_file[256],
            active_tr[256],
             clus_method[256];
    strcpy(version, wv::VersionString().c_str());
    strcpy(data_file, std::string(FileBase::UserWSDataDir() + "/" + _dataFileName).c_str());
    strcpy(active_tr, active_transform.c_str());
    strcpy(clus_method, cluster_method.c_str());
    fprintf(fp, "%s\n", version);
    fprintf(fp, "%s\n", data_file);
    fprintf(fp, "%s\n", active_tr);
    fprintf(fp, "%s\n", clus_method);

    //If it's a user transform, the individual components of it
    if (active_transform == "user")
    {
        char user_tr[256];
        wv::TRANSFORM_ID tid = _cluSpec->TransformId();
        int num_user_specs = _data->UserSize(tid);
        for (int i=0; i<num_user_specs; ++i)
        {
            std::string user_transform = wv::TransformName( _data->UserTransform(tid, i) );
            strcpy(user_tr, user_transform.c_str());
            int sample = _data->UserSample(tid, i);
            fprintf(fp, "\n%s\t%d\n", user_tr, sample);
        }
    }

    //The parameters to the cluster method
    int numpars = _parameters->NumPars();
    fprintf(fp, "Parameters:\n");
    for (int i=0; i<numpars; ++i)
    {
        char key[256], value[256];
        strcpy(key, _parameters->Key(i).c_str());
        strcpy(value, _parameters->Value(i).c_str());
        fprintf(fp, "%s:\t%s\n", key, value);
    }

    //If it's manual clustering, the coordinates of the circles and lines
    if (cluster_method == "Manual") //Subclass?
    {
        fprintf(fp, "Polylines:\n");
        const std::vector<PLSpec*>& polylines = _cluSpec->Polylines();
        int num_polylines = polylines.size();
        for (int i=0; i<num_polylines; ++i)
        {
            const PLSpec* pl_spec = polylines.at(i);

            std::string transform_x = wv::TransformName( pl_spec->transform_x ),
                    transform_y = wv::TransformName( pl_spec->transform_y );
            int sample_x = pl_spec->sample_x,
                    sample_y = pl_spec->sample_y;

            fprintf(fp, "%d:\n", i+1);

            char x_tr[256], y_tr[256];
            strcpy(x_tr, transform_x.c_str());
            strcpy(y_tr, transform_y.c_str());
            fprintf(fp, "X Axis: %s\t%d\n", x_tr, sample_x);
            fprintf(fp, "Y Axis: %s\t%d\n", y_tr, sample_y);

            int time_begin = pl_spec->time_begin,
                    time_end = pl_spec->time_end;
            fprintf(fp, "Time:\t%d\t%d\n", time_begin, time_end);

            const Polyline* polyline = pl_spec->GetDrawLine();
            fprintf(fp, "Direction:\t%d\n", (int)polyline->RejectDir());
            int length = polyline->Length();
            for (int j=0; j<length; ++j)
                fprintf(fp, "\t%f\t%f\n", (float)polyline->X(j), (float)polyline->Y(j));
        }

        fprintf(fp, "Circles:\n");
        const std::vector<CSpec*>& circles = _cluSpec->Circles();
        int num_circles = circles.size();
        for (int i=0; i<num_circles; ++i)
        {
            const CSpec* c_spec = circles.at(i);

            std::string transform_x = wv::TransformName( c_spec->transform_x ),
                    transform_y = wv::TransformName( c_spec->transform_y );
            int sample_x = c_spec->sample_x,
                    sample_y = c_spec->sample_y;

            fprintf(fp, "%d:\n", i+1);

            char x_tr[256], y_tr[256];
            strcpy(x_tr, transform_x.c_str());
            strcpy(y_tr, transform_y.c_str());
            fprintf(fp, "X Axis: %s\t%d\n", x_tr, sample_x);
            fprintf(fp, "Y Axis: %s\t%d\n", y_tr, sample_y);

            int time_begin = c_spec->time_begin,
                    time_end = c_spec->time_end;
            fprintf(fp, "Time: %d\%d\n", time_begin, time_end);

            const Circle* circle = c_spec->GetDrawLine();
            if (const BiEllipse* be = dynamic_cast<const BiEllipse*>(circle))
            {
                wv::float_t sd = be->SD();
                fprintf(fp, "Ellipse SD: %f\n", (float)sd);
            }
            int length = circle->Length();
            for (int j=0; j<length; ++j)
                fprintf(fp, "\t%f\t%f\n", (float)circle->X(j), (float)circle->Y(j));
        }

        fprintf(fp, "Centers:\n");
        const std::vector<const CenterSpec*>& centers = _cluSpec->Centers();
        int num_centers = centers.size();
        for (int i=0; i<num_centers; ++i)
        {
            const CenterSpec* center = centers.at(i);

            std::string transform_x = wv::TransformName( center->transform_x ),
                    transform_y = wv::TransformName( center->transform_y ),
                    sample_x = std::to_string( center->sample_x ),
                    sample_y = std::to_string( center->sample_y ),
                    val_x = std::to_string( center->sample_x ),
                    val_y = std::to_string( center->sample_y );
            std::string center_str = transform_x + " " + sample_x + " " + val_x + " "
                    + transform_y + " " + sample_y + " " + val_y;

            fprintf(fp, "%d:\n", i+1);

            char center_ch[256];
            strcpy(center_ch, center_str.c_str());
            fprintf(fp, "%s\n", center_ch);
        }

        fprintf(fp, "Toggled:\n");
        const boost::dynamic_bitset<>& is_toggled =
                static_cast<const ClusterSpec*>(_cluSpec.get())->ToggledWaves();
            //Otherwise the non-const, private version of the function will be called
        int num_records = is_toggled.size();
        for (int i=0; i<num_records; ++i)
            if (is_toggled.test(i))
                fprintf(fp, "%d\n", i);
    }
    fprintf(fp, "@@@:\n");
}
