#include "userdatafileout.h"
#include "../../memrep/clustermediator.h"

UserDataFileOut::UserDataFileOut(const std::string& file_name, wv::CLUSTER_ID cid)
    : OutputFileBase(file_name, false, NULL, ClusterMediator::Instance().Accepted(cid)), _cid(cid),
      _isAccepted(ClusterMediator::Instance().Accepted(cid))
{
    ClusterMediator& cmed = ClusterMediator::Instance();
    _class = cmed.GetClassification(cid).get();
    _level = cmed.ClusLevel(cid);
    _transform = cmed.GetCluSpecPtr(cid)->Transform();
    _tid = cmed.GetCluSpecPtr(cid)->TransformId();
    _waveData = cmed.GetWaveData(cid);
}

void UserDataFileOut::Save() const
{
    FILE* fp = File();
    if (fp)
    {
        ClusterMediator& cmed = ClusterMediator::Instance();
        const boost::dynamic_bitset<>& is_scrubbed = cmed.GetWaveData(_cid)->Scrubbed();
            //This has the size of the *original* number of records.
        int num_all_records = is_scrubbed.size();
        const wv_vector_int* all_timestamps = _waveData->AllTimestamps();

        Clustering clustering = _class->VClusters().at(_level);
        const wv_vector_int* vcluster = clustering.second;
        const std::vector<int>& clusters = _class->Clusters(_level);
        MatCPtr timestamps = _waveData->Waveforms(wv::TIMESTAMP),
                height = _waveData->Waveforms(wv::HEIGHT),
                width = _waveData->Waveforms(wv::WIDTH),
                slope = _waveData->Waveforms(wv::SLOPE),
                slopeD = _waveData->Waveforms(wv::SLOPE_D),
#ifndef USE_GSL
                energy = _waveData->Waveforms(wv::ENERGY),
#endif
                waves = _waveData->Waveforms(_transform, _tid);
        int num_samples = waves->size2;

        //Print number of records and samples in the data set
        fprintf(fp, "%d\t%d\n", num_all_records, num_samples);

        //Now each cluster
        int num_clusters = _class->NumClusters(_level);
        for (int i=0; i<num_clusters; ++i)
            fprintf(fp, "%d\t", clusters.at(i));
        fprintf(fp, "\n");

        //Name of the transform
        std::string tf = wv::TransformName(_transform);
        fprintf(fp, "%s\n", tf.c_str());

        //Now column headers
        fprintf(fp, "Accepted\tCluster\tTimestamp\tHeight\tWidth\tSlope\tSlopeD");
        for (int j=0; j<num_samples; ++j)
            fprintf(fp, "\t%d", j);
        fprintf(fp, "\n");

        //Now the data for each record
        for (int i=0, ct=0; i<num_all_records; ++i)
        {
            if ( is_scrubbed.test(i) )
            {
                fprintf(fp, "%d\t%d\t%lld", 0, wv::NULL_CLUSTER,
                        (ULL)wv_vector_int_get(all_timestamps, i));

                for (int j=0; j<num_samples+4; ++j)
                    fprintf(fp, "\t%d", 0);

                fprintf(fp, "\n");
            }
            else
            {
                //The stats
#ifdef USE_GSL
                fprintf(fp, "%d\t%d\t%lld\t%f\t%f\t%f\t%f",
                        (int)_isAccepted.test(ct),
                        wv_vector_int_get(vcluster, ct),
                        (ULL)wv_matrix_get(timestamps.get(), ct, 0),
                        height ? wv_matrix_get(height.get(), ct, 0) : 0.0,
                        width ? wv_matrix_get(width.get(), ct, 0) : 0.0,
                        slope ? wv_matrix_get(slope.get(), ct, 0) : 0.0,
                        slopeD ? wv_matrix_get(slopeD.get(), ct, 0) : 0.0
                        );
#else
                fprintf(fp, "%d\t%d\t%lld\t%f\t%f\t%f\t%f\t%f",
                        (int)_isAccepted.test(ct),
                        wv_vector_int_get(vcluster, ct),
                        (ULL)wv_matrix_get(timestamps.get(), ct, 0),
                        height ? wv_matrix_get(height.get(), ct, 0) : 0.0,
                        width ? wv_matrix_get(width.get(), ct, 0) : 0.0,
                        slope ? wv_matrix_get(slope.get(), ct, 0) : 0.0,
                        slopeD ? wv_matrix_get(slopeD.get(), ct, 0) : 0.0,
                        energy ? wv_matrix_get(energy.get(), ct, 0) : 0.0
                        );
#endif

                //The waveforms
                for (int j=0; j<num_samples; ++j)
                    fprintf(fp, "\t%f", wv_matrix_get(waves.get(), ct, j));

                fprintf(fp, "\n");
                ++ct;
            }
        }
    }
    else
        Log("WARNING: Invalid file name, " + Name() + ", no records loaded.");
}

