#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "../memrep/parameters/sessionpars.h"
#include "../memrep/parameters/cluster/manualpars.h"
#include "../memrep/parameters/cluster/kkparameters.h"
#include "../memrep/parameters/cluster/kmeanspars.h"
#include "../memrep/clustermediator.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent), ui(new Ui::MainWindow),
    _paramGui(nullptr), _progressGui(nullptr),
#ifdef HAVE_EXTRA
    _waveGridGui(nullptr),
#endif
    _batch(nullptr), _channelBatch(nullptr),
    _cmed(ClusterMediator::Instance())//, _parameters(NULL)
{
    qRegisterMetaType<std::string>("std::string");
    qRegisterMetaType<WavesFilePtr>("WavesFilePtr");

    ui->setupUi(this);

#if defined(Q_WS_X11) || defined(Q_OS_LINUX)
    FileBase::SetUsersDir(QApplication::applicationDirPath().toStdString() + "/Users");
#elif defined(Q_WS_MAC)  || defined Q_OS_MACX
    FileBase::SetUsersDir(QApplication::applicationDirPath().toStdString() + "/../../../Users");
#elif defined(Q_WS_WIN)
    FileBase::SetUsersDir(QApplication::applicationDirPath().toStdString() + "/../Users");
#endif

#ifdef QT_DEBUG
    ReceiveProfile("matt");
#if defined(Q_WS_X11) || defined(Q_OS_LINUX)
    FileBase::SetUserDataDir("SampleFiles");
    FileBase::SetUserWSDataDir("Users/matt/wsdata");
#elif defined(Q_WS_MAC)  || defined Q_OS_MACX
    FileBase::SetUserDataDir("../../../SampleFiles");
    FileBase::SetUserWSDataDir("../../../Users/matt/wsdata");
#elif defined(Q_WS_WIN)
    FileBase::SetUserDataDir("SampleFiles");
    FileBase::SetUserWSDataDir("Users/matt/wsdata");
#endif
#else
    setEnabled(false);

    _profileGui = new ProfileGui();
    Qt::WindowFlags pf_flags = Qt::Window
            | Qt::CustomizeWindowHint
            | Qt::WindowTitleHint
            | Qt::WindowMinimizeButtonHint
            | Qt::WindowMaximizeButtonHint
            | Qt::WindowCloseButtonHint;
    _profileGui->setWindowFlags(pf_flags);
    _profileGui->setGeometry(x()+width()/2-75, y()+height()/2, 150, 50);
    connect(_profileGui, SIGNAL(CurrentProfile(const std::string&)), this, SLOT(ReceiveProfile(const std::string&)));
    _profileGui->show();
#endif

    ui->menuBatch->setEnabled(false);

    ui->actionLoad_Classification->setEnabled(false);
    ui->actionAutoClass->setEnabled(false);
    ui->actionKMeans->setEnabled(false);
    ui->actionSPC->setEnabled(false);
    ui->actionKlustaKwik->setEnabled(false);

    _statusBarLabel = new QLabel(this);
    setStatusBar( new QStatusBar(this) );
    statusBar()->addPermanentWidget(_statusBarLabel);

    _waveSpaceGui = new WaveSpaceGui(NULL);
    Qt::WindowFlags rd_flags = Qt::Window
            | Qt::CustomizeWindowHint
            | Qt::WindowTitleHint
            | Qt::WindowMinimizeButtonHint
            | Qt::WindowMaximizeButtonHint;
    _waveSpaceGui->setWindowFlags(rd_flags);
    connect(_waveSpaceGui, SIGNAL(SendParameters(ParamsBase*)), this, SLOT(DisplayParameters(ParamsBase*)));
    connect(this, SIGNAL(StopMovie()), _waveSpaceGui, SLOT(StopMovie()));

    _paramGui = new ParamGui(NULL);
    QMdiSubWindow* pg_win = ui->mdiArea->addSubWindow(_paramGui);
    pg_win->hide();
    pg_win->setAttribute(Qt::WA_DeleteOnClose, false);

    _logGui = new LogGui(ui->mdiArea);
    Qt::WindowFlags log_flags = Qt::Window
            | Qt::CustomizeWindowHint
            | Qt::WindowTitleHint
            | Qt::WindowMinimizeButtonHint
            | Qt::WindowMaximizeButtonHint;
    QMdiSubWindow* subwin = ui->mdiArea->addSubWindow(_logGui, log_flags);
    subwin->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    connect(&SSLog::Instance(), SIGNAL(WriteToGui(std::string)), _logGui, SLOT(LogMessage(std::string)),
            Qt::QueuedConnection);

    _electrodeGui = new ElectrodeGui(ui->mdiArea);
    Qt::WindowFlags elc_flags = Qt::Window
            | Qt::CustomizeWindowHint
            | Qt::WindowTitleHint;
    _electrodeSubWin = new QMdiSubWindow(ui->mdiArea, elc_flags);
    _electrodeSubWin->setWidget(_electrodeGui);
    QRect eg = _electrodeSubWin->geometry();
    eg.setWidth(75);
    _electrodeSubWin->setGeometry(eg);
    _electrodeSubWin->setMaximumHeight(250);
    ui->mdiArea->addSubWindow(_electrodeSubWin, elc_flags);
    connect(_electrodeGui,SIGNAL(SetElectrode(short,WavesFilePtr)),
            this,SLOT(ElectrodeSet(short,WavesFilePtr)));

    _progressGui = &pg();
    _progressGui->setParent(ui->mdiArea);
    Qt::WindowFlags pg_flags = Qt::Window
            | Qt::CustomizeWindowHint
            | Qt::WindowTitleHint
            | Qt::WindowMinimizeButtonHint
            | Qt::WindowMaximizeButtonHint;
    QMdiSubWindow* prog_win = ui->mdiArea->addSubWindow(_progressGui, pg_flags);
    prog_win->hide();

    _aboutGui = new AboutGui();
#ifdef USE_GSL
    _waveSorterProGui = new AboutGui();
#else
    ui->menuHelp->removeAction( ui->actionWaveSorter_Pro );
#endif

#ifdef HAVE_EXTRA
    _waveGridGui = new WaveGridGui();
    connect(_waveGridGui, SIGNAL(CreateLaunchWG(wv::CLUSTER_ID)), this, SLOT(CreateLaunchWG(wv::CLUSTER_ID)));
    connect(_waveGridGui, SIGNAL(SendTabChanged(WavesFilePtr)), this, SLOT(ResetElectrodeGui(WavesFilePtr)));
#endif

    _cmed.SetMainWindow(this);

    QApplication::processEvents();

#ifndef USE_GSL
    CheckLicense();
#endif
}
MainWindow::~MainWindow()
{
    delete ui;
    if (_batch) delete _batch;
    if (_channelBatch) delete _channelBatch;

    QDir temp_files_dir( (FileBase::UserPath() + "/temp").c_str() );
    StringList temp_files = ut::ToStringList( temp_files_dir.entryList() );
    std::for_each(temp_files.begin(), temp_files.end(), [&](const std::string& temp_file)
                                                            {
                                                                temp_files_dir.remove( temp_file.c_str() );
                                                            } );
    QDir pid_files_dir = QDir(FileBase::UserPath().c_str());
    StringList pid_files = ut::ToStringList( pid_files_dir.entryList( QStringList("pid*") ) );
    std::for_each(pid_files.begin(), pid_files.end(), [&](const std::string& pid_file)
                                                            {
                                                                pid_files_dir.remove( pid_file.c_str() );
                                                            } );
    std::string temp_dir;
    if (FileBase::UserName() == (temp_dir = "temp" + std::string( std::to_string(QApplication::applicationPid()) )))
    {
        std::string cmd = "rm -r " + FileBase::UsersDir() + "/" + temp_dir + ";";
        system(cmd.c_str());
    }
}
#ifdef HAVE_EXTRA
void MainWindow::AddToWaveGrid(wv::CLUSTER_ID cid)
{
    short elc = _cmed.GetElectrode(cid);
    GridWaveWidget* gww = new GridWaveWidget(_waveGridGui, elc);
    int num_records = _cmed.GetWaveData(cid)->NumRecords();
    const std::vector<QColor>& color_palette( WaveGui::ColorPalette() );
    size_t palette_sz = color_palette.size();
    std::vector<QColor>* colors = new std::vector<QColor>(num_records);

    const std::vector<Clustering>& clusterings = _cmed.GetClassification(cid)->VClusters();
    const wv_vector_int* vcluster = clusterings.at(0).second;
    std::vector<int> clusters = _cmed.GetClassification(cid)->Clusters(0);
    std::map<int, int> cluster_indexes;
    int ct=0;
    for (auto it : clusters)
        cluster_indexes[it] = ct++;

    for (int i=0; i<num_records; ++i)
    {
        int cidx = cluster_indexes[ wv_vector_int_get(vcluster, i) ];
        (*colors)[i] = color_palette.at( cidx % palette_sz);
    }
    gww->SetWaves(cid, wv::RAW, wv::DEFAULT_TID, colors);
    _waveGridGui->AddWaveWidget(gww, _cmed.GetInputFile(cid));

    if (!_waveGridGui->isVisible()) _waveGridGui->show();
}
#endif
void MainWindow::CreateWaveGui(wv::CLUSTER_ID cid) //slot
{
    WaveGuiPtr wave_gui = WaveGuiPtr( new WaveGui(cid) );

        //When the user clicks the 'accept' button on the WaveGui, the WaveSpaceGui is correspondingly updated.
    connect(wave_gui.get(), SIGNAL(SendClusters(wv::CLUSTER_ID)),
            _waveSpaceGui, SLOT(ReceiveClusters(wv::CLUSTER_ID)), Qt::DirectConnection);

        //When the user completes an expansion or manual clustering in WaveSpaceGui, the _isAccepted data member
        //in WaveGui is updated (but nothing else is changed, and pressing 'Accept' in WaveGui will overwrite this).
    connect(_waveSpaceGui, SIGNAL(UpdateWavesShown()),
            this, SLOT(UpdateWavesShown()), Qt::DirectConnection);

    _cmed.AddWaveGui(cid, wave_gui);
}
void MainWindow::LaunchWaveGui(wv::CLUSTER_ID cid) //slot
{
    WaveGuiPtr wave_gui = _cmed.GetWaveGui(cid);

    connect(wave_gui.get(), SIGNAL(GotFocus(wv::CLUSTER_ID)),
            this, SLOT(WaveGuiFocus(wv::CLUSTER_ID)), Qt::DirectConnection);
    connect(wave_gui.get(), SIGNAL(ResetWaves()),
            _waveSpaceGui, SLOT(on_btnResetClassification_clicked()), Qt::DirectConnection);
    connect(wave_gui.get(), SIGNAL(Closed(wv::CLUSTER_ID)),
            this, SLOT(WaveGuiClosed(wv::CLUSTER_ID)), Qt::DirectConnection);

    ui->mdiArea->addSubWindow(wave_gui.get());
    wave_gui->SetWaves();
    wave_gui->show();

//    w->setGeometry( QRect(20,20,300,200) );
}

void MainWindow::showEvent(QShowEvent*)
{
    int mdi_x = ui->mdiArea->x(),
            mdi_y = ui->mdiArea->y(),
            mdi_height = ui->mdiArea->height();
    _logGui->setGeometry(mdi_x, mdi_y-mdi_height-100, 200, 100);
        //### Not working! >:[
}
void MainWindow::closeEvent(QCloseEvent*)
{
    _waveSpaceGui->close();
#ifdef HAVE_EXTRA
    _waveGridGui->close();
    delete _waveGridGui;
#endif
}

void MainWindow::ReceiveProfile(const std::string& profile)
{
    FileBase::SetUserName(profile);
    setEnabled(true);

    ProfileIn profile_in;
    try
    {
        profile_in.Load();

        QDir temp_dir( (FileBase::UserPath() + "/temp").c_str() );
        StringList temp_files = ut::ToStringList( temp_dir.entryList() );
        std::for_each(temp_files.begin(), temp_files.end(), [&](const std::string& temp_file)
                                                                {
                                                                    temp_dir.remove( temp_file.c_str() );
                                                                } );
    }
    catch (WIOException& e)
    {
        Log("Profile: " + profile + ", " + e.what());
    }
#ifndef QT_DEBUG
    _profileGui->close();
#endif
}

void MainWindow::OnActionLoadWaveformFileTriggered(bool)
{
    std::string dir = QDir(FileBase::UserDataDir().c_str()).exists()
            ? FileBase::UserDataDir() : "";
    std::string full_path = QFileDialog::getOpenFileName(this,
        "Load Waveform File",
        dir.c_str(),
        "standard (*.nev *.plx *.apm *M *.txt);;nev files (*.nev);;plx files (*.plx);;apm files (*.apm)" \
                       ";;Mex files (*M);;ascii files (*.txt);;all files (*)"
                                                         ).toStdString();

    if (full_path.empty()) return; //I.e. user presses cancel

    WavesFilePtr waves_file = LoadDataFile(full_path);
    ui->actionLoad_Classification->setEnabled(true);
    ui->actionAutoClass->setEnabled(true);
    ui->actionKMeans->setEnabled(true);
    ui->actionKlustaKwik->setEnabled(true);
    ui->actionSPC->setEnabled(true);

    ResetElectrodeGui(waves_file);
    QRect rect = _electrodeSubWin->geometry();
    rect.setWidth(120);
    rect.setHeight(30 + 30 * waves_file->NumElectrodes());
    _electrodeSubWin->setGeometry(rect);

    //Put the info specifying these data sessions in ClusterMediator and initialize WaveSpaceGui
    const std::vector<short>& electrodes = waves_file->VElectrodes();
    short electrode0 = electrodes.at(0);
    _cmed.AddInputFile(wv::DEFAULT_CID, waves_file);
    _cmed.SetElectrode(wv::DEFAULT_CID, electrode0);
    _cmed.SetCurrentWaveGui(wv::DEFAULT_CID); //Or else WaveSpaceGui will try to update the WaveGui
        //currently in focus
    _electrodeGui->SetElectrodeButton(electrode0);

    _electrodeCt = 0;
    int num_electrodes = waves_file->NumElectrodes();
    if (num_electrodes>1)
    {
        if (_channelBatch)
            delete _channelBatch;
#ifdef HAVE_EXTRA
        if (num_electrodes>=WaveGridGui::MIN_NUM_CELLS)
            _waveGridGui->Setup(waves_file);
#endif
    }
    waves_file->GenerateTransforms(electrode0);
    PreprocFinished(waves_file, electrode0);

    if (num_electrodes>1)
    {
        waves_file->SetIsBatch(true); // ### Needs its own member/flag
        ui->actionLoad_Waveform_File->setEnabled(false);
        _channelBatch = new ChannelBatch(waves_file, this);
        connect(_channelBatch, SIGNAL(PreprocFinished(WavesFilePtr, short)), this, SLOT(PreprocFinished(WavesFilePtr, short)),
                Qt::QueuedConnection);
        _channelBatch->Run();

        for (auto it : electrodes)
        {
            if (it==electrode0) continue; //This case taken care of above
            waves_file->WData(it)->Unload(wv::RAW);
        }
    }

    _waveSpaceGui->Setup(wv::DEFAULT_CID);

    _waveSpaceGui->show();

    ui->menuBatch->setEnabled(true);
}

void MainWindow::OnActionREXEFileTriggered(bool)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid==wv::DEFAULT_CID)
    {
        Log("WARNING: Must choose classification (click 'Accept') before saving.  No data saved.");
        return;
    }
    std::string dir = QDir(FileBase::UserDataDir().c_str()).exists()
            ? FileBase::UserDataDir() : "";
    std::string old_efile = QFileDialog::getOpenFileName(NULL,
                                              "Choose REX E file",
                                              dir.c_str(),
                                              "REX E-files (*E)").toStdString();
    if (old_efile.empty()) return;
    std::string new_efile = old_efile;
    new_efile.erase( new_efile.length()-1 );
    new_efile += "_S" + std::to_string( _cmed.GetElectrode(wg_cid) ) + "_E";
    short electrode =  _cmed.GetElectrode(wg_cid);
    Log( std::to_string(_cmed.Accepted(wg_cid).count()) );
    RexEFile out(old_efile, new_efile, _cmed.GetInputFile(wg_cid)->WData(electrode),
                 _cmed.Accepted(wg_cid));
    out.Save();

    SaveParameters( _cmed.GetInputFile(wg_cid)->Name() );
    OnActionCluFileTriggered(true); //Otherwise there is really no way to validate the classification
}
void MainWindow::OnActionCluFileTriggered(bool is_prog)
{
    WriteCluFile(is_prog, false);
}
void MainWindow::OnActionTCluFileTriggered(bool is_prog)
{
    WriteCluFile(is_prog, true);
}
void MainWindow::WriteCluFile(bool is_prog, bool with_time)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid==wv::DEFAULT_CID)
    {
        Log("WARNING: Must choose classification (click 'Accept') before saving.  No data saved.");
        return;
    }
    std::string file_name = _cmed.GetInputFile(wg_cid)->Name();
    if (is_prog)
        Log("Saving cluster file, " + file_name);
    short electrode =  _cmed.GetElectrode(wg_cid);
    std::string full_file_name;
    if (with_time)
    {
        full_file_name = FileBase::UserWSDataDir() + "/" + file_name
                + ".tclu." + std::to_string(electrode) + "-" + std::to_string(wg_cid);
        UserTCluFileOut out(full_file_name, wg_cid);
        out.Save();
    }
    else
    {
        full_file_name = FileBase::UserWSDataDir() + "/" + file_name
                + ".clu." + std::to_string(electrode) + "-" + std::to_string(wg_cid);
        UserCluFileOut out(full_file_name, wg_cid);
        out.Save();
    }

    SaveParameters(file_name);
    Log("File " + full_file_name + " saved.");
}

void MainWindow::OnActionSaveAllTriggered(bool)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid==wv::DEFAULT_CID)
    {
        Log("WARNING: Must choose classification (click 'Accept') before saving.  No data saved.");
        return;
    }
    std::string file_name = _cmed.GetInputFile(wg_cid)->Name();
    short electrode =  _cmed.GetElectrode(wg_cid);
    std::string full_file_name = FileBase::UserWSDataDir() + "/" + file_name
            + ".data." + std::to_string(electrode) + "-" + std::to_string(wg_cid);
    UserDataFileOut out(full_file_name, wg_cid);
    out.Save();

    SaveParameters(file_name);
    Log("File " + full_file_name + " saved.");
}
void MainWindow::OnActionSimpleBinaryTriggered(bool)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid==wv::DEFAULT_CID)
    {
        Log("WARNING: Must choose classification (click 'Accept') before saving.  No data saved.");
        return;
    }
    std::string file_name = _cmed.GetInputFile(wg_cid)->Name();
    short electrode =  _cmed.GetElectrode(wg_cid);
    std::string full_file_name = FileBase::UserWSDataDir() + "/" + file_name
            + ".bin." + std::to_string(electrode) + "-" + std::to_string(wg_cid);
    SimpleBinary out(full_file_name, _cmed.GetWaveData(wg_cid), _cmed.Accepted(wg_cid));
    out.Save();

    SaveParameters(file_name);
    OnActionCluFileTriggered(true); //Otherwise there is really no way to validate the classification
    Log("File " + full_file_name + " and corresponding parameter (pars) file saved.");
}
void MainWindow::OnActionSimpleASCIITriggered(bool)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid==wv::DEFAULT_CID)
    {
        Log("WARNING: Must choose classification (click 'Accept') before saving.  No data saved.");
        return;
    }
    std::string file_name = _cmed.GetInputFile(wg_cid)->Name();
    short electrode =  _cmed.GetElectrode(wg_cid);
    std::string full_file_name = FileBase::UserWSDataDir() + "/" + file_name
            + ".ascii." + std::to_string(electrode) + "-" + std::to_string(wg_cid);
    SimpleAscii out(full_file_name, _cmed.GetWaveData(wg_cid), _cmed.Accepted(wg_cid));
    out.Save();

    SaveParameters(file_name);
    OnActionCluFileTriggered(true); //Otherwise there is really no way to validate the classification
    Log("File " + full_file_name + " saved.");
}
void MainWindow::OnActionLoadClassificationTriggered(bool)
{
    MutexLocker lock(_mutex);

    std::string dir = QDir(FileBase::UserDataDir().c_str()).exists()
            ? FileBase::UserDataDir() : "";
    std::string full_path = QFileDialog::getOpenFileName(this,
        "Load Classification",
        dir.c_str()
        ).toStdString();

    if (full_path.empty()) return;

    InputFilePtr input_file = _cmed.GetInputFile(wv::DEFAULT_CID);

    UserCluFileIn in(full_path);
    short electrode = _cmed.GetElectrode(wv::DEFAULT_CID);
    const boost::dynamic_bitset<>& is_scrubbed = input_file->WData(electrode)->Scrubbed();
    in.Load(is_scrubbed);
    ClassificationCPtr cls(in.CData());

    _cmed.IncrementCID();
    ClusterFromFile(input_file, electrode, cls);
}

void MainWindow::OnActionRunBatchTriggered(bool)
{
    //Manual clustering for batch processing not allowed of course
    CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(wv::DEFAULT_CID);
    if (clu_spec->ClusterMethod()==wv::MANUAL)
    {
        Log("You may not run batch processing using the manual clustering method.  Please select"
            " another method");
        return;
    }

    if (ui->actionRunBatch->text() == "Run")
    {
        std::string batch_dir_str = QFileDialog::getExistingDirectory(this,
                                "Directory for batch processing", FileBase::UserDataDir().c_str()
                                                                      ).toStdString();
        if (batch_dir_str.empty()) return;
        QDir batch_dir(batch_dir_str.c_str());
        _batchList.clear();
         for (int i=0; i<batch_dir.entryInfoList(QDir::Files).size(); ++i)
            _batchList.push_back( batch_dir.entryInfoList(QDir::Files).at(i).absoluteFilePath().toStdString() );
        if (_batchList.empty()) return;

        BatchGui* batch_gui = new BatchGui;
        connect(batch_gui, SIGNAL(BatchStart()), this, SLOT(BatchStart()), Qt::QueuedConnection);
        batch_gui->setWindowModality(Qt::ApplicationModal);
        batch_gui->show();
    }
    else
    {
        _batch->Cancel();
        ui->actionRunBatch->setText("Run");
    }
}

void MainWindow::OnActionSessionTriggered(bool)
{
    DisplayParameters( new SessionPars() );
}
void MainWindow::OnActionKlustaKwikTriggered(bool)
{
    DisplayParameters( new KKParameters() );
}
void MainWindow::OnActionAutoClassTriggered(bool)
{
    DisplayParameters( new AutoClassPars() );
}
void MainWindow::OnActionKMeansTriggered(bool)
{
    DisplayParameters( new KMeansPars() );
}
void MainWindow::OnActionSPCTriggered(bool)
{
    DisplayParameters( new SPCParameters() );
}
void MainWindow::OnDisplayParameters(ParamsBase* pars, bool)
{
    std::shared_ptr<ParamsBase> params_ptr(pars);
    if (pars->Name() != "Session")
    {
        //Here, if the clustering method has just been selected from WSG (!prev_params) we set
        //it to the associated parameter type in WSG
        ClusterParamsPtr prev_params = _cmed.GetCluSpecPtr(wv::DEFAULT_CID)->ClusterParameters();
        if (!prev_params)
            _cmed.GetCluSpecPtr(
                        wv::DEFAULT_CID)->SetClusterParameters(
                            std::static_pointer_cast<ClusterParamsBase>(params_ptr) );
    }
    _paramGui->SetParameters(params_ptr);
    if (pars->NumPars() > 0)
        _paramGui->show();
    else
        _paramGui->hide();
}

void MainWindow::on_actionAbout_triggered()
{
    _aboutGui->show();
}
void MainWindow::on_actionWaveSorter_Pro_triggered()
{
    _waveSorterProGui->setWindowTitle("WaveSorter Pro");
    _waveSorterProGui->SetText("WaveSorter Pro is the commercial version of WaveSorter.  It is even faster, with many"
                               " operations executing 2x as fast as in this version or faster, and has many more features,"
                               " such as a Grid View for high-density microarrays (such as the Utah Array(tm))."
                               "\n\nTo find out more information, please send an email \ to info@wavesorter.com.");
    _waveSorterProGui->show();
}

void MainWindow::PreprocFinished(WavesFilePtr waves_file, short elc) //slot
{
    MutexLocker lock(_mutex);

    try
    {
        //Add the original, 'clean' version of the session to _cmed, so that switching between electrodes
        //gets you back to the default
        wv::CLUSTER_ID ecid = -_cmed.IncrementCID();
        _cmed.AddInputFile(ecid, waves_file);
        _cmed.InitSpecs(ecid);
        _cmed.SetElectrode(ecid, elc);
        _cmed.SetElcDefaultCid(waves_file, elc, ecid);
        _waveSpaceGui->InitializeSpecs(ecid);

        WaveData* wave_data = waves_file->WData(elc);
        if (wave_data->HasOnlineClus())
        {
            wv_vector_int* online_vcluster = wave_data->OnlineVCluster();

            ClassificationPtr cls( new Classification(online_vcluster) );
            wv::CLUSTER_ID fcid = ClusterFromFile(waves_file, elc, cls, ecid);
            wave_data->SetOnlineClusterCID(fcid);
            _cmed.GetWaveData(fcid)->Unload();
        }
        std::vector<short> electrodes = waves_file->VElectrodes();
        if (elc != electrodes.at(0)) wave_data->Unload();

        if (++_electrodeCt == (int)electrodes.size())
        {
            size_t num_electrodes = electrodes.size(),
                    num_waves_all = 0;
            Log("\n\n\n");
            for (size_t i=0; i<num_electrodes; ++i)
            {
                int num_waves = waves_file->NumWaves(i);
                Log("Number of waveforms from channel " + std::to_string(i) + ": " + std::to_string(num_waves));
                num_waves_all += num_waves;
            }

            std::string status = "File " + waves_file->Name() + " loaded, "
                    + std::to_string(num_waves_all) + " records across "
                    + std::to_string(num_electrodes)
                    + ((_electrodeCt==1) ? " electrode" : " electrodes");
            Log(status);
            Log("\n\n\n");
            _statusBarLabel->setText( status.c_str() );
            ui->actionLoad_Waveform_File->setEnabled(true);
        }
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

void MainWindow::BatchStart() //slot
{
    CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(wv::DEFAULT_CID);
    WSGSpec& wsg_spec = clu_spec->GetWSGSpec();
    if (clu_spec->ClusterMethod()==wv::UNKNOWN_METHOD)
        clu_spec->SetClusterMethod( wv::KMEANS );
    clu_spec->SetTransform(wsg_spec.GuiTransform(), wsg_spec.GuiTransformId() );

    _batch = new Batch(_batchList);
    connect(_batch, SIGNAL(BatchComplete()), this, SLOT(BatchComplete()), Qt::QueuedConnection);
    _batch->Run();

    ui->actionRunBatch->setText("Cancel");
}
void MainWindow::BatchComplete() //slot
{
    delete _batch;
    _batch = nullptr;
    ui->actionRunBatch->setText("Run");
}
void MainWindow::ClusterFinished(void* pcid)
{
    MutexLocker lock(_mutex);

    try
    {
        wv::CLUSTER_ID cid = *static_cast<wv::CLUSTER_ID*>(pcid);
        delete static_cast<wv::CLUSTER_ID*>(pcid);
        CreateWaveGui(cid); //Batch mode
            //Classification added to ClusterMediator in ClusMethodTOBase::LoadClassification
        _cmed.PurgeClusterMgr(cid);

        //If this clustering arose from batch processing, waves and toggledwaves will now need to be set manually
        CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(cid);
        if (!clu_spec->Waves()) clu_spec->InitWaves(cid);

        LaunchWaveGui(cid);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void MainWindow::CreateLaunchWG(wv::CLUSTER_ID cid)
{
    CreateWaveGui(cid);
    LaunchWaveGui(cid);
}

void MainWindow::UpdateWavesShown() //slot
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if (wg_cid == wv::DEFAULT_CID) throw BadCID("MainWindow::UpdateWavesShown", wg_cid);

    //We check if the file upon which WaveSpaceGui is currently based is the same as that for
    //the WaveGui.  This need not be the case when the update comes as a result of switching between
    //electrodes, and the new electrode is different from the one currently in focus.
    InputFilePtr wg_if = _cmed.GetInputFile(wg_cid),
            wsg_if = _cmed.GetInputFile(wv::DEFAULT_CID);
    short wg_elc = _cmed.GetElectrode(wg_cid),
            wsg_elc = _cmed.GetElectrode(wv::DEFAULT_CID);
    if (wg_if!=wsg_if || wg_elc!=wsg_elc) return;

    const WSGSpec& wsg_spec = _cmed.GetCluSpecPtr(wv::DEFAULT_CID)->GetWSGSpec();
    const boost::dynamic_bitset<>& waves_shown = wsg_spec.WaveShown();
    WaveGuiPtr wave_gui = _cmed.GetWaveGui(wg_cid);
    wave_gui->SetViewable(waves_shown);
}
void MainWindow::ElectrodeSet(short elc, WavesFilePtr waves_file) //slot
{
    if (!_cmed.HasElectrode(elc) || !_cmed.HasDefaultElc(waves_file, elc)
            || elc==_cmed.Electrode(wv::DEFAULT_CID)) return;
    _cmed.SetElectrode(wv::DEFAULT_CID, elc);

    //We use the current wave gui only to get at the *default* cid for this input file/electrode
    wv::CLUSTER_ID cid = _cmed.ElcDefaultCid(waves_file, elc);
    wv::CLUSTER_ID ocid = _cmed.GetWaveData(cid)->OnlineClusterCID();
    if (_cmed.HasWaveGui(ocid))
        _cmed.GetWaveGui(ocid)->raise(); //### Doesn't work!!!!
    _cmed.SetCurrentWaveGui(wv::DEFAULT_CID);
        //Crucial, otherwise a subsequent 'accept cluster' from the previous WaveGui--which probably
        //does *not* have the same electrode as what is now selected--will cause a crash

    _waveSpaceGui->Setup(cid);
}
void MainWindow::ResetElectrodeGui(WavesFilePtr waves_file)
{
    _electrodeGui->SetElectrodes(waves_file);
#ifdef HAVE_EXTRA
    _waveGridGui->SetTab(waves_file);
#endif
}
void MainWindow::WaveGuiFocus(wv::CLUSTER_ID cid) //slot
{
    if (cid==wv::DEFAULT_CID)
        throw BadCID("MainWindow::WaveGuiFocus", cid);
    wv::CLUSTER_ID wg_cid_prev = _cmed.CurrentWaveGui();
    if (cid == wg_cid_prev) return;

    //Conserve memory
    _cmed.GetWaveData(wg_cid_prev)->Unload();

    //Stop the movie if it is playing in WaveSpaceGui
    emit StopMovie();

    //Now reset electrodes in the electrode gui on the basis of the current (possibly new) file
    WavesFilePtr prev_input_file = std::static_pointer_cast<WaveformsInBase>( _cmed.GetInputFile(wg_cid_prev) ),
            input_file = std::static_pointer_cast<WaveformsInBase>( _cmed.GetInputFile(cid) );
    if (prev_input_file != input_file)
        ResetElectrodeGui(input_file);

    //Select the correct electrode
    short elc = _cmed.GetElectrode(cid);
    _electrodeGui->SetElectrodeButton(elc);

    //Set new WaveGui as current
    _cmed.SetCurrentWaveGui(cid);
    WaveGuiPtr wave_gui = _cmed.GetWaveGui(cid);

    //Now get the corresponding WaveData and load it into WaveSpaceGui
    WaveData* wave_data = input_file->WData(elc);
    if (!wave_data->IsLoaded())
    {
        wave_data->Reload();
        wave_gui->SetWaves();
    }

    //Set the parameters in ParamGui to whatever they were for this cluster, if they have been set--otherwise
    //(if clustering was loaded from file)--close the parameter gui
    CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(cid);
    ClusterParamsPtr parameters = clu_spec->ClusterParameters();
    if (parameters) _paramGui->SetParameters(parameters);
    else _paramGui->close();

    //Finally, update WaveSpaceGui
    _waveSpaceGui->Setup(cid);
}
void MainWindow::WaveGuiClosed(wv::CLUSTER_ID cid)
{
    emit StopMovie();
    _cmed.RemoveCID(cid);
    _cmed.SetCurrentWaveGui(wv::DEFAULT_CID);
}

#ifndef USE_GSL
void MainWindow::CheckLicense()
{
    LicenseMgr license_mgr;
    bool b =license_mgr.CheckLicense("license.txt");
    if (!b)
    {
        QMessageBox message_box(QMessageBox::Critical, "License Manager", "Invalid or missing license detected."
                                "WaveSorter cannot launch.  Please contact info@wavesorter.com for assistance.");
        message_box.exec();
        exit(EXIT_SUCCESS);
    }
}
#endif
wv::CLUSTER_ID MainWindow::ClusterFromFile(InputFilePtr input_file, short elc,
                                 ClassificationCPtr cls, wv::CLUSTER_ID cid)
{
    CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(cid)->MintNew(); //The new clustering will be derived from this one
    if (!clu_spec->Waves()) clu_spec->InitWaves(cid);
    wv::CLUSTER_ID fcid = _cmed.LastCID();
    _cmed.AddClusterSpec(fcid, clu_spec);
    _cmed.AddInputFile(fcid, input_file);
    _cmed.SetElectrode(fcid, elc); //Must come *after* AddInputFile
    _cmed.AddClustering(fcid, cls); //classification not deleted when
           //it is destroyed
    CluSpecPtr new_clu_spec = _cmed.GetCluSpecPtr(fcid);
    new_clu_spec->SetClusterMethod(wv::UNKNOWN_METHOD);
    _waveSpaceGui->InitializeSpecs(fcid);
#ifdef HAVE_EXTRA
    if (input_file->Electrodes().size() > WaveGridGui::MIN_NUM_CELLS)
        AddToWaveGrid(fcid);
    else
#endif
        CreateLaunchWG(fcid);
    return fcid;
}
WavesFilePtr MainWindow::LoadDataFile(const std::string& data_file)
{
    WavesFilePtr waves_file = WavesFilePtr(WaveformsInBase::Create(data_file));
    int num_waveforms = waves_file->Load();
    if (num_waveforms==0)
        throw WIOException("MainWindow::LoadDataFile", data_file,
                          "ERROR: File was not opened or read properly, no waveforms loaded.");
    _cmed.AddInputFile(wv::DEFAULT_CID, waves_file);

    return waves_file;
}
void MainWindow::SaveParameters(const std::string& file_name)
{
    wv::CLUSTER_ID wg_cid = _cmed.CurrentWaveGui();
    if ( !_cmed.GetCluSpecPtr(wg_cid)->ClusterParameters() )
    {
        Log("Note, saving an online clustering.  No parameter file saved.");
        return;
    }
    ParFileOut par_file( std::string(FileBase::UserWSDataDir() + "/" + file_name + ".pars." + std::to_string(wg_cid)),
                         file_name, wg_cid );
    par_file.Save();
}


