#include "wavegui.h"
#include "ui_wavegui.h"
#include "../../memrep/clustermediator.h"

int WaveGui::DEFAULT_NUM_SHOWN = 100;
std::vector<QColor> WaveGui::_colorPalette;

WaveGui::WaveGui(wv::CLUSTER_ID cid)
     : QWidget(NULL), ui(new Ui::WaveGui), _cid(cid), _cmed(ClusterMediator::Instance()),
       _numWavesShown(DEFAULT_NUM_SHOWN)
{
    if (cid==wv::DEFAULT_CID) throw BadCID("WaveGui::WaveGui", cid);
        //All WaveGuis must correspond to a specific clustering

    if (_colorPalette.size()==0) MakeColorPalette();
        //It gets made elsewhere (first call of ColorPalette()) if multi-channel file gets loaded

    //Get all the objects needed from ClusterMediator
    ClusterMediator& cmed = ClusterMediator::Instance();
    CluSpecPtr clu_spec = cmed.GetCluSpecPtr(cid);
    InputFilePtr input_file = cmed.GetInputFile(cid);
    short electrode = cmed.GetElectrode(cid);
    wv::TRANSFORM transform = clu_spec->Transform();
    wv::CLUSTER_METHOD cluster_method = clu_spec->ClusterMethod();
    std::string clus_name = cluster_method==wv::UNKNOWN_METHOD
            ? "from file"
            : wv::ClusterName(cluster_method);
    std::string transform_name = wv::TransformName(transform);

    //Set all the WaveGui members accordingly
    WaveData* wave_data = input_file->WData(electrode);
    _isViewable = boost::dynamic_bitset<>(wave_data->NumRecords());
    _haveFirstAccepted = false;

    //Start with GUI properties
    ui->setupUi(this);
    setWindowTitle( ("#" + std::to_string(_cid) + ", " + input_file->Name()).c_str() );
    ui->wgtPlot->SetBackground(Qt::white);
    ui->lblSpecs->setText( ("Using " + clus_name + " on " + transform_name
                            + " transform, electrode " + std::to_string(electrode)).c_str() ) ;
    setAttribute(Qt::WA_DeleteOnClose, true);
    setFocusPolicy(Qt::ClickFocus);
    ui->lsStats->hide();

    connect(ui->sldLevel, SIGNAL(sliderMoved(int)), this, SLOT(LevelSliderMoved(int)));

    Setup();
}

WaveGui::~WaveGui()
{
    delete ui;
}

void WaveGui::SetTemplates()
{
    int level_idx = _cmed.ClusLevel(_cid);
    ClassificationCPtr classification = _cmed.GetClassification(_cid);
    WSGSpec& wsg_spec = _cmed.GetCluSpecPtr(_cid)->GetWSGSpec();

    wsg_spec.MakeTemplates(classification->Clusters(level_idx),
                       classification->VClusters().at(level_idx).second,
                       _cid, level_idx, _colorPalette);

    ui->wgtPlot->SetTemplates(_cid, &_isBoxShown, level_idx);
}
void WaveGui::Setup()
{
    int level_idx = _cmed.ClusLevel(_cid);
    wv::float_t level = _cmed.ClusLevelVal(_cid);
    ClassificationCPtr classification = _cmed.GetClassification(_cid);

    std::string level_str = std::to_string(level);
    level_str.erase(4); //I.e. only the 1st four characters, including decimal pt., are 'significant'
    ui->lblLevel->setText( level_str.c_str() );

    int num_clusterings = classification->VClusters().size();
    if (num_clusterings == 1)
    {
        ui->lblLevel->hide();
        ui->sldLevel->hide();
    }
    else
    {
        ui->lblLevel->show();
        ui->sldLevel->show();
        ui->sldLevel->setMinimum(0);
        ui->sldLevel->setValue(level_idx);
        ui->sldLevel->setMaximum(classification->VClusters().size()-1);
    }

    WaveData* wave_data = _cmed.GetWaveData(_cid);
    int num_records = wave_data->NumRecords();

    /////////Get cluster-based data
    const std::vector<Clustering>& clusterings = classification->VClusters();
    const wv_vector_int* vcluster = clusterings.at(level_idx).second;

    //Check if there are any null cluster values--this can happen when loading a
    //classification saved with a stricter scrub criterion
    bool has_null = false;
    for (int i=0; i<num_records; ++i)
        if ( wv_vector_int_get(vcluster, i)==wv::NULL_CLUSTER )
        {
            has_null = true;
            break;
        }
    std::vector<int> clusters = classification->Clusters(level_idx);
    if (has_null) clusters.push_back(wv::NULL_CLUSTER);
    int num_clusters_selected = clusters.size();

    //Initialize vectors containing cluster statistics
    _percents = std::vector<int>(num_clusters_selected);
    std::vector<int> totals(num_clusters_selected);

    //Set the map from cluster number to ordinal rank--cumbersome but don't see how to
    //get around having this.
    //  Also initialize the record of whether the templates have been constructed or not.
    int ct=0;
    _clusterIndexes.clear();
    for (const auto& it : clusters)
    {
        _clusterIndexes[it] = ct++;
        //Since the selected clusters can be e.g. 2, 3, 5, we have to find the
        //index within the set of selected clusters to which they correspond.
        _haveTemplates[it] = false;
    }

    //Now get the percent of waveforms in the gui, for all selected clusters. Also set up the colors here.
    _colors = std::vector<QColor>(num_records);
    int palette_sz = _colorPalette.size();
    for (int i=0; i<num_records; ++i)
    {
        int clus = wv_vector_int_get(vcluster, i);
        int cidx = _clusterIndexes[clus];
        ++totals[cidx];

        _colors[i] = _colorPalette.at( cidx % palette_sz );
    }
    for (int i=0; i<num_clusters_selected; ++i)
        _percents[i] = (int)( (100.0*totals[i]) / num_records + 0.5 );

    ///////Now set up the checkboxes
    QFont font;
    font.setPointSize(14);
    while (ui->layButtons->count()>1)
    {
        QWidget* cbox = ui->layButtons->takeAt(1)->widget();
        delete cbox;
    }

    //Always at least one cluster, so add in the first checkbox manually
    QPalette palette;
    palette.setColor(QPalette::Foreground, Qt::white);
    ui->cbox1->setPalette(palette);
    ui->cbox1->setFont(font);
    ui->cbox1->setText( ("1: " + std::to_string(_percents[0]) + "%").c_str() );
    ui->cbox1->setStyleSheet("background-color: " + _colorPalette.at(0).name());
    connect(ui->cbox1,SIGNAL(clicked()),this,SLOT(UpdateShown()));

    //Now set up the rest of the checkboxes
    int num_clusters = classification->NumClusters(level_idx);
    for (int i=1; i<num_clusters; ++i)
    {
        QCheckBox* cbox = new QCheckBox( (std::to_string(i+1)).c_str() );
        ui->layButtons->addWidget(cbox);
        QPalette palette;
        palette.setColor(QPalette::Foreground, Qt::white);
        cbox->setPalette(palette);
        cbox->setFont(font);
        cbox->setText( (std::to_string(i+1) + ": " + std::to_string(_percents[i]) + "%").c_str() );
        cbox->setStyleSheet("background-color: " + _colorPalette.at(i%palette_sz).name());
        cbox->setChecked(true);
        connect(cbox,SIGNAL(clicked()),this,SLOT(UpdateShown()));
    }
}
void WaveGui::SetWaves()
{
    WaveData* wave_data = _cmed.GetWaveData(_cid);

    //Now set up the WaveWidget.  This always displays the raw waveforms or their associated templates
        //First the waveforms
    int num_records = wave_data->NumRecords();
    _isViewable = boost::dynamic_bitset<>(num_records);
    int step = std::max(num_records/DEFAULT_NUM_SHOWN, 1);
    for (int i=0; i<num_records; ++i)
        if (i%step==0) _isViewable.set(i, true);
    ui->wgtPlot->SetAxisAbsoluteYRange(wave_data->TMin(wv::RAW),
                                       wave_data->TMax(wv::RAW) );
    ui->wgtPlot->SetWaves(_cid, wv::RAW, wv::DEFAULT_TID, &_colors);

        //Update everything
    UpdateWavePlot();
}
void WaveGui::SetViewable(const boost::dynamic_bitset<>& is_viewable)
{
    _isViewable = is_viewable;
    UpdateShown();
}
const std::vector<QColor>& WaveGui::ColorPalette()
{
    if (_colorPalette.empty()) MakeColorPalette();
    return _colorPalette;
}

void WaveGui::OnBtnAcceptClicked(bool)
{
    setCursor(Qt::WaitCursor);
    _haveFirstAccepted = true;
    CalcClusAccepted();
    emit GotFocus(_cid);
    emit SendClusters(_cid);
        //These two signals are separate because you can get focus to e.g. examine different
        //clusterings without wanting to send them to WSG
    setCursor(Qt::ArrowCursor);
}
void WaveGui::OnBtnAnalyzeClicked(bool)
{
    emit GotFocus(_cid);
    emit ResetWaves(); //So they don't come up as accepted.  Get rid of this when proper colors
        //start getting used.
}
void WaveGui::OnBtnStats(bool is_prog)
{
    if (ui->btnStats->text() == "Show Stats" || is_prog)
    {
        int level = _cmed.ClusLevel(_cid);
        if ( _cmed.Stats(_cid, level).empty() )
        {
            setCursor(Qt::WaitCursor);
            CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(_cid);
            MatCPtr waves = clu_spec->Waves();
            const wv_vector_int* vclusters = _cmed.GetClassification(_cid)->VClusters().at(level).second;
            const std::vector<int>& clusters = _cmed.GetClassification(_cid)->Clusters().at(level);
            int j_begin = clu_spec->SelectedSamplesBegin(),
                    j_end = clu_spec->SelectedSamplesEnd();
            Dunn dunn(waves, vclusters, clusters, j_begin, j_end);
            dunn.Calculate();
            const std::vector<wv::float_t>& dunn_indexes = dunn.Indexes();
            _cmed.SetStats(_cid, level, dunn_indexes);
            setCursor(Qt::ArrowCursor);
        }
        const std::vector<wv::float_t>& dunn_indexes = _cmed.Stats(_cid, level);

        ui->lsStats->clear();
        for (auto it : dunn_indexes)
            ui->lsStats->addItem( std::to_string(it).c_str() );
        ui->lsStats->show();
        ui->btnStats->setText("Hide Stats");
    }
    else
    {
        ui->lsStats->hide();
        ui->btnStats->setText("Show Stats");
    }
}
void WaveGui::OnBtnTemplatesClicked(bool)
{
    if (ui->btnTemplates->text() == "Average")
    {
        int level = _cmed.ClusLevel(_cid);
        if (!_haveTemplates[level])
        {
            SetTemplates();
            _haveTemplates[level] = true;
        }
        ui->wgtPlot->SetMode(WaveWidget::TEMPLATES);
        ui->btnTemplates->setText("Individual");
    }
    else
    {
        ui->wgtPlot->SetMode(WaveWidget::WAVEFORMS);
        ui->btnTemplates->setText("Average");
    }
    ui->wgtPlot->update();
}

void WaveGui::LevelSliderMoved(int value) //slot
{
    _cmed.SetClusLevel(_cid, value);

    wv::float_t level = _cmed.ClusLevelVal(_cid);
    ui->lblLevel->setText( std::to_string(level).c_str() );
    if (ui->btnStats->text()=="Hide Stats") OnBtnStats(true);

    Setup();
    if (ui->wgtPlot->Mode()==WaveWidget::WAVEFORMS) SetWaves();
    else SetTemplates(); //WaveWidget::TEMPLATES
    UpdateWavePlot();
}
void WaveGui::UpdateShown() //slot
{
    CalcShown();
    ui->wgtPlot->update();
}
void WaveGui::UpdateWavePlot() //slot
{
    CalcShown();
    CalcClusAccepted();
    ui->wgtPlot->update();
}
void WaveGui::on_cboxRelative_clicked(bool is_checked)
{
    WaveData* wave_data = _cmed.GetWaveData(_cid);
    if (is_checked)
        ui->wgtPlot->SetAxisLimitMode(WaveWidget::RELATIVE);
    else
    {
        ui->wgtPlot->SetAxisAbsoluteYRange( wave_data->TMin(wv::RAW),
                                            wave_data->TMax(wv::RAW) );
        ui->wgtPlot->SetAxisLimitMode(WaveWidget::ABSOLUTE);
    }
    ui->wgtPlot->update();
}

void WaveGui::closeEvent(QCloseEvent*)
{
    emit Closed(_cid);
}

void WaveGui::CalcShown()
{
    UpdateIsCBoxShown();

    WaveData* wave_data = _cmed.GetWaveData(_cid);
    int level_idx = _cmed.ClusLevel(_cid);
    ClassificationCPtr classification = _cmed.GetClassification(_cid);

    int num_records = wave_data->NumRecords();
    WSGSpec& wsg_spec = _cmed.GetCluSpecPtr(_cid)->GetWSGSpec();
    boost::dynamic_bitset<>& wave_shown = wsg_spec.WaveShown();
    wave_shown = boost::dynamic_bitset<>(num_records);

    const wv_vector_int* clustering = classification->VClusters().at(level_idx).second;
    for (int i=0; i<num_records; ++i)
    {
        int clus = wv_vector_int_get(clustering, i);
        int cidx = _clusterIndexes[clus];

        bool is_shown = _isViewable.test(i) && _isBoxShown[cidx];
        wave_shown.set(i, is_shown);
    }
}
void WaveGui::CalcClusAccepted()
{
    UpdateIsCBoxShown();

    WaveData* wave_data = _cmed.GetWaveData(_cid);
    int level_idx = _cmed.ClusLevel(_cid);
    ClassificationCPtr classification = _cmed.GetClassification(_cid);

    int num_records = wave_data->NumRecords();
    const wv_vector_int* clustering = classification->VClusters().at(level_idx).second;
    wv_vector_int* clus_accepted = wv_vector_int_alloc(num_records);
        //clus_accepted will be owned by ClusterMediator/ResultSpec
    for (int i=0; i<num_records; ++i)
    {
        int clus = wv_vector_int_get(clustering, i);
        int cidx = _clusterIndexes[clus];

        bool is_accepted = _isBoxShown[cidx];
        if (is_accepted)
            wv_vector_int_set(clus_accepted, i, wv::ACCEPT_CLUSTER);
        else if (wv_vector_int_get(clustering, i)==wv::ACCEPT_CLUSTER) //This happens
            //when altering the selection deriving from a manual clustering
            wv_vector_int_set(clus_accepted, i, wv::ZERO_CLUSTER);
        else
            wv_vector_int_set(clus_accepted, i, clus);
    }
    _cmed.SetClusAccepted(_cid, clus_accepted);
}
void WaveGui::MakeColorPalette()
{
    //Set up the set of colors used in plotting
    _colorPalette.push_back(Qt::blue);
    _colorPalette.push_back(Qt::red);
    _colorPalette.push_back( QColor("orange") );
    _colorPalette.push_back(Qt::magenta);
    _colorPalette.push_back(Qt::yellow);
    _colorPalette.push_back(Qt::black);
    _colorPalette.push_back(Qt::cyan);
    _colorPalette.push_back(Qt::gray);
    _colorPalette.push_back( QColor("olive") );
        //Annoyingly this list doesn't have the same ordering as
        //QGlobalColor does.  Don't use green, that's reserved for accepted waveforms in WaveSpaceGui
}
void WaveGui::UpdateIsCBoxShown()
{
    //Identify which of the clusters selected in classification actually get displayed,
    //on the basis of the checkbox selections
    _isBoxShown.clear();
    int num_boxes = ui->layButtons->count();
    for (int i=0; i<num_boxes; ++i)
    {
        QCheckBox* box = static_cast<QCheckBox*>(ui->layButtons->itemAt(i)->widget());
        _isBoxShown[i] = box->isChecked();
    }

    std::vector<int> clusters;
    for (const auto& iter : _clusterIndexes)
    {
        int cidx = iter.second;
        if (_isBoxShown[cidx]) clusters.push_back( iter.first );
    }
    _cmed.AddClusters(_cid, clusters);
}
