#include "histwidget.h"
#include "../../analysis/threads/mgr/histwgtthreadmgr.h"

const int HistWidget::TIMELINE_HEIGHT = 6;

HistWidget::HistWidget(QWidget *parent) :
    QWidget(parent), _histAccept(NULL), _histUnaccept(NULL), _matAccept(NULL), _matUnaccept(NULL),
    _prevMouseX(-1), _prevMouseY(-1), _needPaint(false), _scaleSpec(NULL),
    _selectedDisk(-1), _isDraggingAxis(false), _isDraggingDisk(false),
    _currentWheelDelta(0), _wsgSpec(NULL), _drawlineState(HistWidget::STOPPED),
    _scrubIndex(-1), _scrubIndexFromWaves(false)
{
    setMouseTracking(true);
    _mousePosLabel = new QLabel(this);
    QPalette palette;
    palette.setColor(QPalette::Foreground, QColor("red"));
    _mousePosLabel->setPalette(palette);

    for (int i=0; i<WSGSpec::NUM_DISKS; ++i)
        _diskPixX[i] = _diskPixY[i] = 0;
    _diskColors[0] = QColor(255,255,0,95);
    _diskColors[1] = QColor(0,255,255,95);
    _diskColors[2] = QColor(255,0,255,95);
    _diskColors[3] = QColor(165,42,42,95);
    _diskColors[4] = QColor(240,128,128,95);
}
HistWidget::~HistWidget()
{
    free(_pixelBuf);
}

void HistWidget::InitGraphics(CluSpecPtr clu_spec, ScaleSpec* scale_spec)
{
    _cluSpec = clu_spec;
    _wsgSpec = &_cluSpec->GetWSGSpec();

    _clusterBegin = _wsgSpec->ClusterBegin();
    _clusterEnd = _wsgSpec->ClusterEnd();
    _isAccepted = &static_cast<const WSGSpec*>(_wsgSpec)->Accepted();
    _scaleSpec = scale_spec;
}
void HistWidget::ScaleExpand()
{
    static wv::float_t prev_zoom = _wsgSpec->HistAxisZoom();
    wv::float_t c = _wsgSpec->HistAxisZoom() - prev_zoom + 1.0,
            xmin = _scaleSpec->MinValX(), xmax = _scaleSpec->MaxValX(),
            ymin = _scaleSpec->MinValY(), ymax = _scaleSpec->MaxValY(),
            xloc = _scaleSpec->PixToValX(_currentMouseX), yloc = _scaleSpec->PixToValY(_currentMouseY);
    _scaleSpec->SetValuesX( (xmin - xloc)/c + xloc,
                            (xmax - xloc)/c + xloc);
    _scaleSpec->SetValuesY( (ymin - yloc)/c + yloc,
                            (ymax - yloc)/c + yloc);
    prev_zoom = _wsgSpec->HistAxisZoom();
    emit SendDragEvent();
}
void HistWidget::ScaleShift()
{
    if (_prevMouseX==-1)
    {
        _prevMouseX = _currentMouseX;
        _prevMouseY = _currentMouseY;
        return;
    }
    wv::float_t xmin = _scaleSpec->MinValX(), xmax = _scaleSpec->MaxValX(),
            ymin = _scaleSpec->MinValY(), ymax = _scaleSpec->MaxValY();
    wv::float_t dx = _scaleSpec->PixToValX(_prevMouseX - _currentMouseX) - xmin,
            dy = ymax - _scaleSpec->PixToValY(_currentMouseY - _prevMouseY);
    _scaleSpec->SetValuesX(xmin+dx, xmax+dx);
    _scaleSpec->SetValuesY(ymin+dy, ymax+dy);
    _prevMouseX = _currentMouseX;
    _prevMouseY = _currentMouseY;
    emit SendDragEvent();
}
void HistWidget::StartCenter()
{
    raise(); //Have to give the widget focus
    _drawlineState = DRAWING_CENTER;
    setCursor( Qt::CrossCursor );
}
void HistWidget::StartCircle()
{
    raise(); //Have to give the widget focus
    _drawlineState = DRAWING_CIRCLE;
    setCursor( Qt::CrossCursor );
}
void HistWidget::StartEllipse()
{
    raise(); //Have to give the widget focus
    _drawlineState = DRAWING_ELLIPSE;
    setCursor( Qt::CrossCursor );
}
void HistWidget::StartPolyline()
{
    raise(); //Have to give the widget focus
    _drawlineState = DRAWING_POLYLINE;
    setCursor( Qt::CrossCursor );
}
bool HistWidget::FinishedDrawLine(bool stop)
{
    bool is_ok = true;
    switch (_drawlineState)
    {
        case DRAWING_POLYLINE:
        {
            std::vector<PLSpec*>& polylines = _cluSpec->Polylines();
            is_ok = EndDrawLine(polylines);
            break;
        }
        case DRAWING_CIRCLE:
        {
            std::vector<CSpec*>& circles = _cluSpec->Circles();
            is_ok = EndDrawLine(circles);
            break;
        }
        case DRAWING_ELLIPSE:
            if (!stop) //When the circle is completed but still need to set SD
            {
                std::vector<CSpec*>& circles = _cluSpec->Circles();

                bool is_circ_ok = EndDrawLine(circles);
                if (!is_circ_ok)
                {
                    is_ok = false;
                    break;
                }

                Circle* circle = circles.back()->GetDrawLine();
                BiEllipse* be = new BiEllipse( circle->GetID()+1 );
                be->InitEllipse(_wsgSpec->HistVector(WSGSpec::X),
                               _wsgSpec->HistVector(WSGSpec::Y), circle);
                BiEllipseSpec* be_spec = new BiEllipseSpec(_wsgSpec->Transform(WSGSpec::X),
                                                          _wsgSpec->Transform(WSGSpec::Y),
                                                          _wsgSpec->Sample(WSGSpec::X),
                                                          _wsgSpec->Sample(WSGSpec::Y),
                                                           be);
                delete circle;
                circles.pop_back();
                circles.push_back(be_spec);

                is_ok = EndDrawLine(circles); //The new 'circle' is the ellipse, which is now functionall identical
                    //to a hand-drawn circle
            }
            break;
        default:
            break;
    }
    if (stop) _drawlineState = STOPPED;
    setCursor( Qt::ArrowCursor );
    return is_ok;
}
void HistWidget::ResetDisks(int i)
{
    _diskPixX[i] = _diskPixY[i] = 0;
    _wsgSpec->SetHistDiskSize(i, WSGSpec::DEFAULT_HIST_DISK_SIZE);
    _wsgSpec->ResetDisk(i);
}
void HistWidget::ResetScaleSpecVals()
{
    SetScaleSpecVals(_wsgSpec->Min(WSGSpec::X),
                       _wsgSpec->Max(WSGSpec::X),
                       _wsgSpec->Min(WSGSpec::Y),
                       _wsgSpec->Max(WSGSpec::Y));
}

void HistWidget::SetHistograms(const wv_histogram2d* hist_accept, const wv_histogram2d* hist_unaccept)
{
    _histAccept = hist_accept;
    _histUnaccept = hist_unaccept;
    _needPaint = true;
    update();
}
void HistWidget::SetMatrices(wv_matrix_int* mat_accept, wv_matrix_int* mat_unaccept)
{
    _matAccept = mat_accept;
    _matUnaccept = mat_unaccept;
}
void HistWidget::SetScaleSpecVals(wv::float_t x_min, wv::float_t x_max, wv::float_t y_min, wv::float_t y_max)
{
    _scaleSpec->SetValuesX(x_min, x_max);
    _scaleSpec->SetValuesY(y_min, y_max);
    _scaleSpec->SetValuesPix(0.0, width(), TIMELINE_HEIGHT, HistHeight());
}

void HistWidget::SetIsScrubbing(bool is_scrubbing)
{
    if (is_scrubbing)
        _drawlineState = SCRUBBING;
    else
    {
        _drawlineState = STOPPED;
        _scrubIndex = -1;
        SendScrubIndex(-1);
    }
}

int HistWidget::HistHeight() const
{
    return height() - TIMELINE_HEIGHT;
}

void HistWidget::ReceiveScrubIndex(int scrub_index)
{
    _scrubIndex = scrub_index;
    _scrubIndexFromWaves = true;
    update();
}

// ***Think about using a QGraphicsView for this
void HistWidget::paintEvent(QPaintEvent*)
{
    try
    {
        _scaleSpec->SetValuesPix(0.0, width(), 0.0, HistHeight());

        size_t hist_ht = HistHeight(),
                hist_wd = width();

#ifdef QT_DEBUG
        QTime tic;
        tic.start();
#endif

        int max_val;
        wv::float_t* xbins, * ybins;
        wv_histogram2d* hist;
        size_t num_xbins, num_ybins;
        if ( _wsgSpec->IsScatter() )
        {
            num_xbins = hist_ht;
            num_ybins = hist_wd;
        }
        else
        {
            if (_histAccept==NULL || _histUnaccept==NULL) return;
            num_xbins = wv_histogram2d_nx(_histAccept);
            num_ybins = wv_histogram2d_ny(_histAccept);

            xbins = new wv::float_t[num_xbins+1];
            ybins = new wv::float_t[num_ybins+1];
            wv::float_t bin_lo, bin_hi;
            for (size_t i=0; i<num_xbins; ++i)
            {
                wv_histogram2d_get_xrange(_histAccept, i, &bin_lo, &bin_hi);
                xbins[i] = _scaleSpec->ValToPixX(bin_lo);
            }
            xbins[num_xbins] = _scaleSpec->ValToPixX(bin_hi);
            for (size_t j=0; j<num_ybins; ++j)
            {
                wv_histogram2d_get_yrange(_histAccept, j, &bin_lo, &bin_hi);
                ybins[j] = _scaleSpec->ValToPixY(bin_lo);
            }
            ybins[num_ybins] = _scaleSpec->ValToPixY(bin_hi);

            //The largest bin value, for scaling.
            hist = wv_histogram2d_clone(_histUnaccept);
            wv_histogram2d_add(hist, _histAccept);
            wv::float_t orig_max_val = wv_histogram2d_max_val(hist);
            wv::float_t lum_scale = _wsgSpec->HistLumScale();
            lum_scale =  qBound((wv::float_t)1.0/orig_max_val, lum_scale, (wv::float_t)1.0);
            max_val = lum_scale * orig_max_val + 0.5;
            _wsgSpec->SetHistLumScale(lum_scale);
        }

        QPainter painter(this);
        painter.setRenderHint(QPainter::Antialiasing);
        QBrush brush;
        brush.setStyle(Qt::SolidPattern);

        //Allocate the image buffer, make static for efficiency and because the buffer needs
        //to be valid for the life of the QImage
        uchar* pixels = new (_pixelBuf) uchar[hist_ht*hist_wd*4];
            // *** The best way would probably be to make this a member, allocate to 2048x2048 or something
            //and then just use the QImage constructor with bytes_per_line
            //QImage::Format_RGB32 is 0xffRRGGBB
        bzero(pixels, hist_ht*hist_wd*4);

        if ( _wsgSpec->IsScatter() )
        {
            if (hist_ht!=_matUnaccept->size1 || hist_wd!=_matUnaccept->size2)
                emit RecalculateScatter();
            HistWgtScatterMgr mgr(pixels, this);
            mgr.StartThreads();
            mgr.Wait();
        }
        else
        {
            HistWgtThreadMgr mgr(pixels, this, max_val);
            mgr.StartThreads();
            mgr.Wait();
        }

        QImage image(pixels, hist_wd, hist_ht, QImage::Format_RGB32);
        painter.drawImage(QPoint(0,0), image);

#ifdef QT_DEBUG
//        qDebug() << "HistWidget::paintEvent(), " <<
//                 (_wsgSpec->IsScatter() ? "scatter" : "histogram") << ": " << num_xbins << ", " << num_ybins << ": " << tic.elapsed();
#endif

        //Draw the timeline
            //First the background...
        QRect tline_bgnd(0, 0, width(), TIMELINE_HEIGHT);
        brush.setColor( QColor(95, 95, 95) );
        painter.setBrush(brush);
        painter.fillRect(tline_bgnd, brush);

            //Now the time bar...
        if (_wsgSpec->HistLimitMode()==WSGSpec::TIMESTAMP)
        {
            wv::float_t time_begin = TimeToPix( _wsgSpec->TimeBegin() ),
                    time_end = TimeToPix( _wsgSpec->TimeEnd() ),
                    time_width = time_end - time_begin;
            if (time_width < 1) time_width = 1;
            QRect tline(time_begin, 0, time_width, TIMELINE_HEIGHT);
            brush.setColor( QColor(255, 255, 0) );
            painter.setBrush(brush);
            painter.fillRect(tline, brush);
        }
        // **Done with timeline**

        painter.setRenderHint(QPainter::Antialiasing);

        //Polylines
        int num_polylines = _cluSpec->Polylines().size();
        for (int i=0; i<num_polylines; ++i)
        {
            PLSpec* pl_spec = _cluSpec->Polylines()[i];
            Polyline* polyline = pl_spec->GetDrawLine();

            int polyline_length = polyline->Length();
            if (pl_spec->sample_x != _wsgSpec->Sample(WSGSpec::X)
                    || pl_spec->sample_y != _wsgSpec->Sample(WSGSpec::Y)
                    || pl_spec->transform_x != _wsgSpec->Transform(WSGSpec::X)
                    || pl_spec->transform_y != _wsgSpec->Transform(WSGSpec::Y)
                    || polyline_length < 2) continue;


            //Get and zoom the points appropriately
            const QPointF* poly_points = polyline->Points();
            QPointF* new_pts = new QPointF[polyline_length+2];
            for (int p=0; p<polyline_length; ++p)
                new_pts[p] = _scaleSpec->ValToPix( poly_points[p] );

            //Draw the line segments...
            QPen pen( QColor(255,0,0) );
            pen.setWidth(1);
            painter.setBrush( Qt::NoBrush );
            painter.setPen(pen);
            painter.drawPolyline(new_pts, polyline_length);

            //... and the shading above/below
            //Need to insert the points differently depending upon the up/down rejection direction
            wv::float_t xmin = new_pts[0].x(),
                    xmax = new_pts[polyline_length-1].x();
            QPointF xbase_begin,
                    xbase_end;
            if (polyline->RejectDir() == Polyline::REJECT_UP)
            {
                xbase_begin = QPointF(xmin, 0);
                xbase_end = QPointF(xmax, 0);
            }
            else
            {
                int hist_ht = HistHeight();
                xbase_begin = QPointF(xmin, hist_ht);
                xbase_end = QPointF(xmax, hist_ht);
            }
            new_pts[polyline_length] = xbase_end;
            new_pts[polyline_length+1] = xbase_begin;
                //polyline.length is of course not updated here--these points don't belong to the line

            //Now actually drawing the shading
            brush.setStyle( Qt::BDiagPattern );
            brush.setColor( QColor(255,0,0, 127) );
            painter.setBrush(brush);
            painter.setPen( Qt::NoPen );
            painter.drawPolygon(new_pts, polyline_length+2 );

            //Now *if* the polyline has restricted time extent, draw a corresponding timeline box
            wv::float_t time_begin = TimeToPix( pl_spec->time_begin ),
                    time_end = TimeToPix( pl_spec->time_end );
            if (time_begin!=0 || time_end!=-1)
            {
                QRect tline(time_begin, 0, time_end-time_begin, TIMELINE_HEIGHT-1);
                QBrush brush( QColor(255, 0, 0) );
                painter.setBrush(brush);
                painter.fillRect(tline, brush);
            }
        }

        //Circles and BiEllipses
        int num_circles = _cluSpec->Circles().size();
        for (int i=0; i<num_circles; ++i)
        {
            CSpec* c_spec = _cluSpec->Circles()[i];
            Circle* circle = c_spec->GetDrawLine();

            int circle_length = circle->Length();
            if (c_spec->sample_x != _wsgSpec->Sample(WSGSpec::X)
                    || c_spec->sample_y != _wsgSpec->Sample(WSGSpec::Y)
                    || c_spec->transform_x != _wsgSpec->Transform(WSGSpec::X)
                    || c_spec->transform_y != _wsgSpec->Transform(WSGSpec::Y)
                    || circle_length < 2) continue;


            //Get points and zoom, scale them appropriately
            const QPointF* points = circle->Points();
            QPointF* new_points = new QPointF[circle_length];
            for (int p=0; p<circle_length; ++p)
                new_points[p] = _scaleSpec->ValToPix( points[p] );

            //Draw the circle and shading at the same time.
            QPen pen( QColor(0,255,0) );
            pen.setWidth(1);
            painter.setPen(pen);
            painter.setBrush( Qt::NoBrush );
            if (circle->IsFinished())
                painter.drawPolygon(new_points, circle_length);
            else
                painter.drawPolyline(new_points, circle_length);

            //Now *if* the circle has restricted time extent, draw a corresponding timeline box
            wv::float_t time_begin = TimeToPix( c_spec->time_begin ),
                    time_end = TimeToPix( c_spec->time_end );
            if (time_begin!=0 || time_end!=-1)
            {
                QRect tline(time_begin, 0, time_end-time_begin, TIMELINE_HEIGHT-2);
                QBrush brush( QColor(0, 255, 0) );
                painter.setBrush(brush);
                painter.fillRect(tline, brush);
            }

            delete[] new_points;
        }

        //Centers
        const std::vector<const CenterSpec*>& centers = _cluSpec->Centers();
        for (auto it : centers)
        {
            if (it->sample_x != _wsgSpec->Sample(WSGSpec::X)
                    || it->sample_y != _wsgSpec->Sample(WSGSpec::Y)
                    || it->transform_x != _wsgSpec->Transform(WSGSpec::X)
                    || it->transform_y != _wsgSpec->Transform(WSGSpec::Y))
                continue;

            QColor color(95,95,255);
            brush.setColor(color);
            brush.setStyle( Qt::SolidPattern );
            painter.setBrush(brush);
            QPen pen(color);
            pen.setWidth(1);
            painter.setPen(pen);
            QRectF win = QRect(_scaleSpec->ValToPixX(it->val_x)-7.5,
                               _scaleSpec->ValToPixY(it->val_y)-7.5,
                               15,
                               15);
            painter.drawEllipse(win);
        }

        //Now draw the moving window (bar or disks)
        QColor color(255,255,0,95);
        brush.setColor(color);
        brush.setStyle( Qt::SolidPattern );
        painter.setBrush(brush);
        switch (_wsgSpec->HistCursorMode())
        {
            case WSGSpec::BAR:
                if (_wsgSpec->HistLimitMode()==WSGSpec::TRANSFORM)
                {
                    QRectF window = _window;
                    window.setTop( _window.top() + TIMELINE_HEIGHT );
                    painter.fillRect(window, brush);
                }
                break;
            case WSGSpec::DISK:
            {
                for (int i=WSGSpec::NUM_DISKS-1; i>=0; --i)
                {
                    int ds = _wsgSpec->HistDiskSize(i) * 2; //_diskSize is disk radius in pixels
                    int x = _diskPixX[i] - ds/2,
                            y = _diskPixY[i] - ds/2,
                            wd = ds,
                            ht = ds;

                    //First the shaded part
                    QColor disk_color = _diskColors[i];
                    QRectF win = QRect(x,y,wd,ht);
                    brush.setColor(disk_color);
                    painter.setPen(Qt::NoPen);
                    painter.setBrush(brush);
                    painter.drawEllipse(win);

                    //Now the reference ring
                    QColor pen_color = disk_color;
                    pen_color.setAlpha(160);
                    QPen pen(pen_color);
                    pen.setWidth(2);
                    painter.setPen(pen);
                    painter.setBrush( Qt::NoBrush );
                    QRectF ring = QRect(_diskPixX[i]-WSGSpec::DEFAULT_HIST_DISK_SIZE,
                                                     _diskPixY[i]-WSGSpec::DEFAULT_HIST_DISK_SIZE,
                                                     2*WSGSpec::DEFAULT_HIST_DISK_SIZE,
                                                     2*WSGSpec::DEFAULT_HIST_DISK_SIZE);
                    painter.drawEllipse(ring);
                }
                break;
            }
        }

        //Now a small circle around the point nearest the cursor, if we're scrubbing.
        //Only test within a small range in order to save time.
        if (_drawlineState == SCRUBBING)
        {
            int nearest_pix_x, nearest_pix_y;
            if (_scrubIndexFromWaves)
            {
                wv::float_t nearest_x = _wsgSpec->Value(WSGSpec::X, _scrubIndex),
                        nearest_y = _wsgSpec->Value(WSGSpec::Y, _scrubIndex);
                nearest_pix_x = _scaleSpec->ValToPixX(nearest_x);
                nearest_pix_y = _scaleSpec->ValToPixY(nearest_y);
            }
            else
            {
                //Get the coordinates of the nearest record
                wv::float_t mouse_val_x = _scaleSpec->PixToValX(_currentMouseX),//mouse_pos.x()),
                        mouse_val_y = _scaleSpec->PixToValY(_currentMouseY);//mouse_pos.y());
                _scrubIndex = _wsgSpec->NearestWave(mouse_val_x, mouse_val_y);
                if (_scrubIndex == -1) goto end_scrub;
                nearest_pix_x = _scaleSpec->ValToPixX( _wsgSpec->Value(WSGSpec::X, _scrubIndex) );
                nearest_pix_y = _scaleSpec->ValToPixY( _wsgSpec->Value(WSGSpec::Y, _scrubIndex) );

                //Send the index of the selected wave to ScanWaveWidget
                emit SendScrubIndex(_scrubIndex);
            }

            //Draw an annulus around nearest record, color-coded according to whether
            //a mouse click will accept or reject it
            QColor color = _isAccepted->test(_scrubIndex) ? QColor(0,255,0) : QColor(255,0,0);
            QPen pen(color);
            painter.setPen(pen);
            painter.setBrush( Qt::NoBrush );
            QPoint center(nearest_pix_x, nearest_pix_y);
            painter.drawEllipse( center, 6, 6);
        }
        end_scrub:

        _needPaint = false;

        //Free memory
        if ( !_wsgSpec->IsScatter() )
        {
            wv_histogram2d_free(hist);
            delete[] xbins;
            delete[] ybins;
        }
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

void HistWidget::mousePressEvent(QMouseEvent* event)
{
    try
    {
        switch (_drawlineState)
        {
            case SCRUBBING:
                emit ToggleWaveform(_scrubIndex);
                break;
            case DRAWING_CENTER:
            {
                CenterSpec* center = new CenterSpec(
                                        _wsgSpec->Transform(WSGSpec::X),
                                        _wsgSpec->Transform(WSGSpec::Y),
                                        _wsgSpec->Sample(WSGSpec::X),
                                        _wsgSpec->Sample(WSGSpec::Y),
                                        _scaleSpec->PixToValX(_currentMouseX),
                                        _scaleSpec->PixToValY(_currentMouseY)
                                    );
                _cluSpec->AddCenter(center);
                update();
                break;
            }
            case DRAWING_CIRCLE:
            case DRAWING_ELLIPSE:
            {
                wv::float_t xpos = event->pos().x(),
                    ypos = event->pos().y();  //For the conversion
                Circle* circle = _cluSpec->Circles().back()->GetDrawLine();
                circle->AddPoint( _scaleSpec->PixToVal(QPointF(xpos, ypos)) );
                update();
                break;
            }
            case DRAWING_POLYLINE:
            {
                wv::float_t xpos = event->pos().x(),
                    ypos = event->pos().y();  //For the conversion
                Polyline* polyline = _cluSpec->Polylines().back()->GetDrawLine();
                polyline->AddPoint( _scaleSpec->PixToVal( QPointF(xpos, ypos) ) );
                update();
                break;
            }
            case STOPPED:
            {
                _currentMouseX = event->x();
                _currentMouseY = event->y();
                setCursor(Qt::ClosedHandCursor);
                if (_wsgSpec->HistCursorMode()==WSGSpec::DISK && _selectedDisk!=-1)
                {
                    _isDraggingDisk = true;
                    _wsgSpec->SetHistZoomMode(WSGSpec::DISK_SIZE);
                }
                else if (_wsgSpec->HistZoomMode()==WSGSpec::AXIS_ZOOM)
                {
                    if (_isDraggingAxis)
                        ScaleShift();
                    else
                    {
                        _isDraggingAxis = true;
                        _prevMouseX = _currentMouseX;
                        _prevMouseY = _currentMouseY;
                    }
                    update();
                }
                break;
            }
        }
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void HistWidget::mouseMoveEvent(QMouseEvent* event)
{
    _currentMouseX = event->x();
    _currentMouseY = event->y();
    int xpos = _scaleSpec->PixToValX(_currentMouseX) + 0.5,
            ypos = _scaleSpec->PixToValY(_currentMouseY) + 0.5;
    std::string val_str = "(" + std::to_string(xpos) + ", " + std::to_string(ypos);
    _mousePosLabel->setText(val_str.c_str());
    int labelx = _currentMouseX>width()-125 ? _currentMouseX-100 : _currentMouseX+15,
            labely = _currentMouseY<35 ? _currentMouseY+15 : _currentMouseY-15;
    _mousePosLabel->setGeometry(labelx, labely, 125, 20);

    if (!_isDraggingDisk)
    {
        int disk_in_range = -1;
        for (int i=0; i<WSGSpec::NUM_DISKS; ++i)
            if (abs(_currentMouseX-_diskPixX[i])<WSGSpec::DEFAULT_HIST_DISK_SIZE
                    && abs(_currentMouseY-_diskPixY[i])<WSGSpec::DEFAULT_HIST_DISK_SIZE)
            {
                disk_in_range = i;
                break;
            }

        if (_drawlineState==STOPPED
                && _wsgSpec->HistCursorMode()==WSGSpec::DISK && _selectedDisk==-1)
        {
            if (disk_in_range!=-1)
            {
                setCursor(Qt::OpenHandCursor);
                _wsgSpec->SetHistZoomMode(WSGSpec::DISK_SIZE);
                _selectedDisk = disk_in_range;
            }
            else //I.e. if no disk was found that was sufficiently close to the cursor
            {
                emit ResetZoomMode();
                setCursor(Qt::ArrowCursor);
            }
        }
        else if (disk_in_range==-1)
            _selectedDisk = -1;
    }

    //NB, mouse tracking is enabled so this gets called with every mouse move
    if (event->buttons()==Qt::NoButton && _drawlineState!=SCRUBBING) return;

    // ***These commands are processed *only* if the button is pressed
    switch (_drawlineState)
    {
        case SCRUBBING:
            _scrubIndexFromWaves = false;
            update();
            break;
        case DRAWING_CIRCLE:
        case DRAWING_ELLIPSE:
        case DRAWING_POLYLINE:
            mousePressEvent(event);
            break;
        case DRAWING_CENTER:
        case STOPPED:
            if (_wsgSpec->HistCursorMode()==WSGSpec::DISK && _selectedDisk!=-1)
            {
                _diskPixX[_selectedDisk] = _currentMouseX;
                _diskPixY[_selectedDisk] = _currentMouseY;
                wv::float_t xpos = _scaleSpec->PixToValX(_currentMouseX),
                        ypos = _scaleSpec->PixToValY(_currentMouseY);
                emit DiskMoved(_selectedDisk, xpos, ypos);
            }
            else if (_wsgSpec->HistZoomMode()==WSGSpec::AXIS_ZOOM) //We're dragging the axis in this case
                mousePressEvent(event);
            break;
    }
}
void HistWidget::mouseReleaseEvent(QMouseEvent* event)
{
    switch (_drawlineState)
    {
        case SCRUBBING:
        case DRAWING_CENTER:
        case DRAWING_CIRCLE:
        case DRAWING_ELLIPSE:
        case DRAWING_POLYLINE:
            break;
        case STOPPED:
            if (_selectedDisk==-1)
                setCursor(Qt::ArrowCursor);
            else
            {
                setCursor(Qt::OpenHandCursor);
                _isDraggingDisk = false;
                if (event->button()==Qt::RightButton) //Right-click to get rid of unwanted disks
                {
                    ResetDisks(_selectedDisk);
                    update();
                }
            }
            _prevMouseX = -1;
            _prevMouseY = -1;
            _isDraggingAxis = false;
            break;
    }
}

void HistWidget::mouseDoubleClickEvent(QMouseEvent* event)
{
    switch (_drawlineState)
    {
        case DRAWING_POLYLINE:
        {
            wv::float_t xpos = event->x(),
                    ypos = event->y(),
                    hist_width = width();
            bool is_left = xpos<hist_width/2.0; //Otherwise, to the right

            Polyline* polyline = _cluSpec->Polylines().back()->GetDrawLine();
            if (is_left)
                polyline->AddPoint( _scaleSpec->PixToVal( QPointF(0, ypos) ) );
            else
                polyline->AddPoint( _scaleSpec->PixToVal( QPointF(width(), ypos) ) );

            update();
            break;
        }
        case DRAWING_CENTER:
            mousePressEvent(event);
            break;
        case DRAWING_CIRCLE:
        case DRAWING_ELLIPSE:
        case SCRUBBING:
        case STOPPED:
            if (_wsgSpec->HistZoomMode()==WSGSpec::AXIS_ZOOM)
            {
                ResetScaleSpecVals();
                emit SendDragEvent();
            }
            break;
    }
}
void HistWidget::wheelEvent(QWheelEvent* event)
{
    int num_degrees = event->delta() / 8;
    _currentWheelDelta += num_degrees;
    if ( fabs((wv::float_t)_currentWheelDelta) >= 15.0 && !_needPaint)
        //If the user spins the wheel too quickly, the histogram won't have redrawn before
        //the next event comes in.  _needPaint prevents these events from piling up.
    {
        _currentMouseX = event->x();
        _currentMouseY = event->y();
        if (_wsgSpec->HistZoomMode() == WSGSpec::AXIS_ZOOM)
            ScaleExpand();

        if (_wsgSpec->HistCursorMode()==WSGSpec::DISK && _selectedDisk!=-1)
        {
            wv::float_t xpos = _scaleSpec->PixToValX( _diskPixX[_selectedDisk] ),
                    ypos = _scaleSpec->PixToValY( _diskPixY[_selectedDisk] );
            emit DiskMoved(_selectedDisk, xpos, ypos);
        }

        emit WheelStep( _currentWheelDelta > 0 );
        _currentWheelDelta = 0;
    }

    event->accept();
}

void HistWidget::focusOutEvent(QFocusEvent*)
{
    _drawlineState = STOPPED;
    setCursor(Qt::ArrowCursor);
}

template<typename T>
bool HistWidget::EndDrawLine(std::vector<T*>& dl_specs)
{
    T* dl_spec = dl_specs.back();
    if (_wsgSpec->HistLimitMode()==WSGSpec::TIMESTAMP)
    {
        dl_spec->time_begin = _wsgSpec->TimeBegin();
        dl_spec->time_end = _wsgSpec->TimeEnd();
    }
    else
    {
        dl_spec->time_begin = 0;
        dl_spec->time_end = -1;
    }

    DrawLineBase* draw_line = dl_spec->GetDrawLine();
    bool is_ok;
    if (draw_line->Length() > 2)
    {
        draw_line->SetIsFinished(true);
        is_ok = true;
    }
    else
    {
        delete dl_spec;
        dl_specs.pop_back();
        is_ok = false;
    }
    return is_ok;
}
wv::float_t HistWidget::TimeToPix(wv::float_t tm) const
{
    wv::float_t time_first = _wsgSpec->Timestamp(0),
            time_last = _wsgSpec->Timestamp(-1),
            time_range = time_last - time_first;
    wv::float_t tpix = (wv::float_t)width() * (tm - time_first) / time_range;
    return tpix;
}
