#ifndef HISTWIDGET_H
#define HISTWIDGET_H

#include "../../globals/globals.h"
#include "../../memrep/parameters/cluster/manualpars.h"
#include "../../memrep/drawing/polyline.h"
#include "../../memrep/drawing/biellipse.h"
#include "../../memrep/specs/clusterspec.h"
#include "../../analysis/threads/mgr/histwgtscattermgr.h"
#include "../../memrep/specs/scalespec.h"

/* **NOTE: Due to the complexities presented by polylines--which are created from and must be
    displayed in pixel coordinates, but are used to classify waveforms (defined in terms of time
    and data values), the HistWidget takes info about the polylines and cluster window in pixels,
    and returns it in data value coordinates
*/

class HistWgtThreadMgr;
class HistWidget : public QWidget
{
    Q_OBJECT

    friend class HistWgtTO;
    friend class HistWgtScatterTO;
    friend class HistScatterTO;
    friend class HistScatterMgr;

    public:
        explicit HistWidget(QWidget *parent = 0);
        ~HistWidget();

        enum DRAWLINE_STATE
        {
            SCRUBBING,
            DRAWING_CENTER,
            DRAWING_CIRCLE,
            DRAWING_ELLIPSE,
            DRAWING_POLYLINE,
            STOPPED
        };

        bool FinishedDrawLine(bool stop = true);
        void InitGraphics(CluSpecPtr clu_spec, ScaleSpec* scale_spec);
            //Should be called only int WaveSpaceGui's constructor
        void ResetDisks(int i);
        void ResetScaleSpecVals();
        void ScaleExpand();
        void ScaleShift();
        void StartCenter();
        void StartCircle();
        void StartEllipse();
        void StartPolyline();

        //These are all so that HistWidget can accommodate zooming without crashing.  All of these
        //get set within WaveSpaceGui::UpdateHistogram
        void SetScaleSpecVals(wv::float_t x_min, wv::float_t x_max, wv::float_t y_min, wv::float_t y_max);
        void SetHistograms(const wv_histogram2d* hist_accept, const wv_histogram2d* hist_unaccept);
        void SetMatrices(wv_matrix_int* mat_accept, wv_matrix_int* mat_unaccept);
        void SetWindow(const QRect& window) { _window = window; }
        void SetClusterRect(wv::float_t cluster_begin, wv::float_t cluster_end);
        void SetIsScrubbing(bool is_scrubbing);

        bool NeedPaint() const { return _needPaint; }
        DRAWLINE_STATE DrawlineState() const { return _drawlineState; }
        int HistHeight() const; //The height of the histogram *without* including the time bar at the top.
        int SelectedDisk() const { return _selectedDisk; }
        int DiskPixX(int i) { return _diskPixX[i]; }
        int DiskPixY(int i) { return _diskPixY[i]; }
        const QColor* DiskColors() const { return _diskColors; }

    public slots:
        void ReceiveScrubIndex(int scrub_index);

    signals:
        void DiskMoved(int selected_disk, wv::float_t x, wv::float_t y);
        void ResetZoomMode(); //Since even when cursor is in disk mode, zooming doesn't change the
            //size of the disk unless we're hovering over it
        void RecalculateScatter();
        void WheelStep(bool is_forward); //true for zoom in, false for zoom out
        void SendScrubIndex(int scrub_index);
        void ToggleWaveform(int scrub_index);
        void SendDragEvent(); //A catch-all, for e.g. after an axis drag event

    protected:
        virtual void paintEvent(QPaintEvent*) override; // *** Note, window should probably disappear or become just an
            //outline while polylines are being drawn.
        virtual void mousePressEvent(QMouseEvent* event) override;
        virtual void mouseMoveEvent(QMouseEvent* event) override;
        virtual void mouseReleaseEvent(QMouseEvent*) override;
        virtual void mouseDoubleClickEvent(QMouseEvent* event) override;
        virtual void wheelEvent(QWheelEvent* event) override;

        virtual void focusOutEvent(QFocusEvent*) override;

    private:
        template<typename T>
        bool EndDrawLine(std::vector<T*>& dl_specs);
        wv::float_t TimeToPix(wv::float_t tm) const;

        static const int TIMELINE_HEIGHT;

        const wv_histogram2d* _histAccept, * _histUnaccept;
        wv_matrix_int* _matAccept, * _matUnaccept;
        wv::float_t _clusterBegin, _clusterEnd;
        QRect _window;
        int _currentMouseX, _currentMouseY, _prevMouseX, _prevMouseY;
        bool _needPaint;
        QLabel* _mousePosLabel;

        ScaleSpec* _scaleSpec;
            //The scale spec for the data that are *currently* being shown, changes with every change
            //in transform or moving of the waves slider
        int _selectedDisk; //-1 when no disk is currently grabbed/nearby.  This is !=-1 for closed *and* open hand
        bool _isDraggingAxis, _isDraggingDisk;
#ifndef Q_OS_WIN
        int _diskPixX[WSGSpec::NUM_DISKS], _diskPixY[WSGSpec::NUM_DISKS];
        QColor _diskColors[WSGSpec::NUM_DISKS];
#else
        int _diskPixX[5], _diskPixY[5];
        QColor _diskColors[5];
#endif
        int _currentWheelDelta;

        CluSpecPtr _cluSpec;
        WSGSpec* _wsgSpec;
            //These two must be set ONLY in InitGraphics

        DRAWLINE_STATE _drawlineState;
        const boost::dynamic_bitset<>* _isAccepted; //Need to know for scrubbing
        int _scrubIndex; //The index of the wave currently 'under consideration' for being (un-)scrubbed
        bool _scrubIndexFromWaves;
        uchar _pixelBuf[2048*2048*4]; //For speed, no need to reallocate a block on every redraw
};
#endif // HISTWIDGET_H
