#include "scanwavewidget.h"
#include "../../../memrep/clustermediator.h"

ScanWaveWidget::ScanWaveWidget(QWidget* parent) : WaveWidget(parent)
{
    setMouseTracking(true);
}

void ScanWaveWidget::InitGraphics(CluSpecPtr clu_spec)
{
    _cluSpec = clu_spec;
}
void ScanWaveWidget::SetWaves(wv::CLUSTER_ID cid, wv::TRANSFORM trans, wv::TRANSFORM_ID tid,
                          const std::vector<QColor>* all_colors, bool reset_samples)
{
    if (reset_samples) _sample = 0;
    WaveWidget::SetWaves(cid, trans, tid, all_colors, reset_samples);
}
void ScanWaveWidget::SetScanline(int sample)
{
    _sample = sample - SamplesBegin();
}

void ScanWaveWidget::paintEvent(QPaintEvent* event)
{
    //First the main painting
    WaveWidget::paintEvent(event);

    //Verify it's matrix data
    if (AllWaves()->size2 == 1) return;

    //Find translated sample coordinates, and pixels-per-sample
    int rel_select_samples_begin = _cluSpec->SelectedSamplesBegin() - SamplesBegin(),
            rel_select_samples_end = _cluSpec->SelectedSamplesEnd() - SamplesBegin();
    int current_range = SamplesEnd() - SamplesBegin();
    wv::float_t xpix_per_samp = (wv::float_t)width() / (wv::float_t)(current_range-1);

    //Now draw the scanline
    QPainter painter(this);
    QColor color(255,255,0,95);
    QBrush brush(color);
    painter.setBrush(brush);
    wv::float_t line_pos = (wv::float_t)_sample * xpix_per_samp;
    int line_xbegin = line_pos - xpix_per_samp/2.0 + 0.5,
            line_width = xpix_per_samp + 0.5;
    QRect scanline(line_xbegin, 0 , line_width, height());
    painter.fillRect(scanline,brush);

    //Add in vertical guidelines.
    painter.setBrush( Qt::NoBrush );
    QPen pen( QColor(127,127,127,195) );
    pen.setWidth(1);
    painter.setPen(pen);
    int step_idx = 1; //This is to ensure that the lines aren't too closely spaced
    while (current_range/step_idx > 24) ++step_idx;
    for (int j=0; j<current_range; j+=step_idx)
    {
        int x = (wv::float_t)j * xpix_per_samp + 0.5;
        painter.drawLine(x,0,x,height());
    }

    //Now the samples rectangle, if it has been specified
    if (_cluSpec->SelectedSamplesBegin() != -1)
    {
        //temporary new rectangle
        int rect_xbegin = rel_select_samples_begin*xpix_per_samp + 0.5,
                rect_width = (wv::float_t)(rel_select_samples_end-rel_select_samples_begin)
                * xpix_per_samp + 0.5;
        QRect sample_rect(rect_xbegin-xpix_per_samp/2+4, 3,
                          rect_width+xpix_per_samp-2, height()-6);

        //Draw it
        QPen pen( QColor(127,127,255,195) );
        pen.setWidth(6);
        painter.setBrush( Qt::NoBrush );
        painter.setPen(pen);
        painter.drawRect(sample_rect);
    }
}

void ScanWaveWidget::mousePressEvent(QMouseEvent* event)
{
    switch (SelectMode())
    {
        case GRAB:
            setCursor(Qt::ClosedHandCursor);
            SetMousePos(event->x(), event->y());
            break;
        case SCRUB:
            if (ScrubIndex() != -1)
                emit ToggleWaveform( ScrubIndex() );
            break;
    }
}
void ScanWaveWidget::mouseMoveEvent(QMouseEvent* event)
{
    if (event->buttons()==Qt::NoButton && SelectMode()!=SCRUB) return;

    switch (SelectMode())
    {
        case GRAB:
        {
            int delta_x_px = event->x() - CurrentMouseX(),
                hist_width = width(),
                current_begin = SamplesBegin(),
                current_end = SamplesEnd(),
                samples_range = current_end - current_begin;
            wv::float_t pix_per_sample = (wv::float_t)hist_width / (wv::float_t)samples_range;
            int samples_delta = (wv::float_t)delta_x_px / pix_per_sample;
            if (samples_delta != 0)
            {
                SetSampleRange(current_begin-samples_delta, current_end-samples_delta, true);
                SetMousePos(event->x(), event->y());
            }
                //Note that if the mouse hasn't moved enough to correspond to a shift of at least 1 sample,
                //_currentMouseX is not updated.
            break;
        }
        case SCRUB:
            SetScrubIndex(-1); //In this case, -1 means get the selected wave from mouse position
                //as opposed to receiving it from HistWidget
            SetMousePos(event->x(), event->y());
            break;
    }
}
void ScanWaveWidget::mouseReleaseEvent(QMouseEvent*)
{
    setCursor(Qt::ArrowCursor);
}

void ScanWaveWidget::wheelEvent(QWheelEvent* event)
{
    if (SelectMode() == GRAB)
    {
        int hist_width = width(),
            mouse_xpos = event->x(),
            current_begin = SamplesBegin(),
            current_end = SamplesEnd(),
            num_samples = NumSamples();

        int samples_per_step = num_samples/16;

        int num_degrees = event->delta() / 8;
        wv::float_t num_steps = (wv::float_t)num_degrees / 15.0;

        wv::float_t mouse_xpos_pct = (wv::float_t)mouse_xpos / (wv::float_t)hist_width;
        wv::float_t delta_samples = samples_per_step * num_steps;
        int dsamples_left = mouse_xpos_pct * delta_samples + 0.5,
            dsamples_right = (1-mouse_xpos_pct) * delta_samples + 0.5;

        //Now if either of the deltas have been put out of range, give all of the samples to
        //the other side.
        if (current_begin + dsamples_left < 0) dsamples_right = delta_samples;
        else if (current_end - dsamples_right > num_samples) dsamples_left = delta_samples;

        int new_begin = current_begin + dsamples_left,
            new_end = current_end - dsamples_right;
        if (new_end - new_begin > 2)
            SetSampleRange(new_begin, new_end);

        event->accept();
    }
}
