#include "wavewidget.h"
#include "../../mdi/wavegui.h"
#include "ui_wavegui.h"
#include "../../../memrep/clustermediator.h"
#ifdef HAVE_EXTRA
#include "gridwavewidget.h"
#endif

WaveWidget::WaveWidget(QWidget* parent) : QWidget(parent),
            _allColors(NULL), _axisLimitMode(RELATIVE), _cid(wv::DEFAULT_CID),
            _isTemplateShown(NULL), _mode(WAVEFORMS), _scrubIndex(-1), _selectMode(GRAB),
            _useDashes(false), _waveData(nullptr), _waveTemplates(NULL)
{
    setAutoFillBackground(true);
}
void WaveWidget::ResetSampleRange()
{
    SetSampleRange(0, MSize2());
}

void WaveWidget::SetBackground(QColor background)
{
    QPalette pal = palette();
    pal.setColor(QPalette::Window, background);
    setPalette(pal);
}

void WaveWidget::SetTemplates(wv::CLUSTER_ID cid,
                              const std::map<int, bool>* is_template_shown, int level_idx)
{
    const WSGSpec& wsg_spec = ClusterMediator::Instance().GetCluSpecPtr(cid)->GetWSGSpec();
    if (cid==wv::DEFAULT_CID)
    {
        _waveTemplates = wsg_spec.GetWaveTemplatesWSG();
        _templateColors = wsg_spec.TemplateColorsWSG();
    }
    else
    {
        _waveTemplates = wsg_spec.GetWaveTemplatesWG(level_idx);
        _templateColors = wsg_spec.TemplateColorsWG(level_idx);
    }
    _isTemplateShown = is_template_shown;
}
void WaveWidget::SetWaves(wv::CLUSTER_ID cid, wv::TRANSFORM trans, wv::TRANSFORM_ID tid,
                          const std::vector<QColor>* all_colors, bool reset_samples)
{
    _cid = cid;
    _transform = trans;
    _tid = tid;
    ClusterMediator& cmed = ClusterMediator::Instance();
    _waveData = cmed.GetWaveData(_cid);
    _allColors = all_colors;
    if (reset_samples || SamplesBegin()==-1 || SamplesEnd()==-1)
    {
        SetSamplesBegin(0);
        SetSamplesEnd( MSize2() );
    }
}
void WaveWidget::SetAxisAbsoluteYRange(wv::float_t min, wv::float_t max)
{
    _absoluteYMin = min;
    _absoluteYMax = max;
}

void WaveWidget::paintEvent(QPaintEvent*)
{
    try
    {
        switch(_mode)
        {
            case WAVEFORMS:
                if (CanDraw()) DrawWaves();
                break;
            case TEMPLATES:
                if (_waveTemplates) DrawTemplates();
                break;
        }
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void WaveWidget::ReceiveScrubIndex(int scrub_index)
{
    _scrubIndex = scrub_index;
    update();
}

void WaveWidget::SetSampleRange(int begin, int end, bool is_restrictive)
{
    int num_samples = MSize2();
    if (is_restrictive && (SamplesBegin()<0 || SamplesEnd()>num_samples)) return;
    SetSamplesBegin( qBound(0, begin, num_samples-1) );
    SetSamplesEnd( qBound(SamplesBegin()+1, end, num_samples) );
    update();
    emit RangeChange(SamplesBegin(), SamplesEnd()); //For the slider
}
void WaveWidget::SetMousePos(int x, int y)
{
    _currentMouseX = x;
    _currentMouseY = y;
    update();
}

void WaveWidget::FindScales(wv::float_t ymin, wv::float_t ymax, int sample_range,
                            wv::float_t& ht_scale, wv::float_t& wd_scale) const
{
    wv::float_t range = ymax - ymin;
    int ht = height(),
        wd = width();

    ht_scale = ht/range;
    wd_scale = (wv::float_t)wd/(wv::float_t)(sample_range-1);
}

void WaveWidget::DrawWaves()
{
    const WSGSpec& wsg_spec = ClusterMediator::Instance().GetCluSpecPtr(_cid)->GetWSGSpec();

    //First get the subset of the total waveform set to display.  Need to create
    //a new matrix because scaling should be done only over the waves actually displayed.
    //  For the sake of efficiency make a subvector for colors too.
    int num_waves_shown = NumWavesShown();
    if (num_waves_shown==0) return;
    int num_records = MSize1(),
            num_samples = MSize2();
    if (num_samples==1) return; //For timestamps, etc.
    wv_matrix* waves_shown = wv_matrix_alloc(num_waves_shown,num_samples);
    SetWavesShown(waves_shown);
    std::vector<QColor> colors;
    bool have_scrub_wave = _scrubIndex != -1;
    int scrub_subindex = -1; //Need a corresponding new index for the submatrix of shown waves
    const boost::dynamic_bitset<>& is_accepted = _cid==wv::DEFAULT_CID
                                            ? wsg_spec.Accepted()
                                            : ClusterMediator::Instance().Accepted(_cid);
    boost::dynamic_bitset<> is_accepted_now(num_waves_shown, false);
    const boost::dynamic_bitset<>& wave_shown = WaveShown();
    for (int i=0, ct=0; i<num_records && ct<num_waves_shown; ++i)
    {
        if (!wave_shown.test(i)) continue;
        colors.push_back( _allColors->at(i) );
        if (have_scrub_wave && i==_scrubIndex)
            scrub_subindex = ct;
        is_accepted_now.set(ct, is_accepted.test(i));
        ++ct;
    }

    //Get the submatrix corresponding to the samples actually shown, so that scaling can use the full
    //dynamic range of the data.
    int samples_begin = SamplesBegin(),
            samples_end = SamplesEnd(),
            sample_range = samples_end - samples_begin;
    wv_matrix* submat = wv_matrix_alloc(num_waves_shown, sample_range);
    for (int j=samples_begin, ct=0; j<samples_end; ++j, ++ct)
    {
        wv_vector_const_view col = wv_matrix_const_column(waves_shown, j);
        wv_matrix_set_col(submat, ct, &col.vector);
    }

    //Taking into account the size and location of the plot widget, scale
    //and shift the waveforms appropriately to draw them.
    switch (_axisLimitMode)
    {
        case ABSOLUTE:
            _currentYMin = _absoluteYMin;
            _currentYMax = _absoluteYMax;
            break;
        case RELATIVE:
            _currentYMin = wv_matrix_min(submat);
            _currentYMax = wv_matrix_max(submat);
            break;
    }

    wv::float_t ht_scale, wd_scale;
    FindScales(_currentYMin, _currentYMax, sample_range, ht_scale, wd_scale);

    wv_matrix_add_constant(waves_shown, -_currentYMin);
    wv_matrix_scale(waves_shown, -ht_scale);
    wv_matrix_add_constant(waves_shown, (wv::float_t)height());

    //So basically, when _scrubIndex is set to -1 despite being in SCRUB mode, it means that
    //the scrubbed wave should be determined on the basis of current mouse position
    if (_selectMode==SCRUB && _scrubIndex==-1)
    {
        int nearest_sample = samples_begin,
            min_dist_x = INT_MAX;
        for (int j=0; j<sample_range; ++j)
        {
            int dist = fabs( (wv::float_t)j*wd_scale - _currentMouseX ) + 0.5;
            if (dist<min_dist_x)
            {
                min_dist_x = dist;
                nearest_sample = j + samples_begin;
            }
        }

        int min_dist_y = INT_MAX;
        for (int i=0, ct=0; i<num_records; ++i)
        {
            if (!wave_shown.test(i)) continue;
            int dist = fabs( wv_matrix_get(waves_shown, ct, nearest_sample) - _currentMouseY ) + 0.5;
                //This is the same as if we had just iterated directly over waves_shown, counting to num_waves_shown
            if (dist<min_dist_y)
            {
                min_dist_y = dist;
                scrub_subindex = ct;
                _scrubIndex = i;
            }
            ++ct;
        }

        emit SendScrubIndex(_scrubIndex);
    }

    //Now graph the subselected waves using the subselected colors.
    QPainter painter(this);
    QPen pen;
    pen.setWidth(0);
    QPointF points[256];
    const int DASH_SIZE = 10, SPACE_SIZE = 2;
    std::vector<qreal> dashes, spaces;
    dashes.push_back(DASH_SIZE);
    dashes.push_back(SPACE_SIZE);
    spaces.push_back(SPACE_SIZE);
    spaces.push_back(DASH_SIZE);
    for (int i=0; i<num_waves_shown; ++i)
    {
        if (i == scrub_subindex) continue;
        QColor color( colors.at(i) );
        pen.setWidth(0);
        pen.setColor(color);
        bool is_dashed = _useDashes && is_accepted_now.test(i);
        if (is_dashed)
        {
            pen.setDashPattern( ut::ToQVector(dashes) );
            pen.setDashOffset(0);
        }
        else
            pen.setStyle(Qt::SolidLine);
        painter.setPen(pen);
        int ct = 0;
        for (int j=samples_begin; j<samples_end; ++j)
        {
            QPointF p = QPointF(ct*wd_scale, wv_matrix_get(waves_shown,i,j));
            points[ct] = p;
            ++ct;
        }
        painter.drawPolyline(points, sample_range);

        if (is_dashed)
        {
            pen.setColor(Qt::green);
            pen.setDashPattern( ut::ToQVector(spaces) );
            pen.setDashOffset(DASH_SIZE+SPACE_SIZE);
            painter.setPen(pen);
            painter.drawPolyline(points, sample_range);
        }
    }

    //Draw the wave selected for scrubbing (if there is one) on top:
#ifdef HAVE_EXTRA
    if (_scrubIndex != -1 && !dynamic_cast<GridWaveWidget*>(this))
#else
    if (_scrubIndex != -1)
#endif
    {
        //We scale the selected wave just as the subset of shown waves is scaled
        wv_vector* row = wv_vector_alloc(num_samples);
        wv_matrix_get_row(row, AllWaves().get(), _scrubIndex);
        wv_vector_add_constant(row, -_currentYMin);
        wv_vector_scale(row, -ht_scale);
        wv_vector_add_constant(row, (wv::float_t)height());

        QColor color(255,255,0);
        pen.setColor(color);
        pen.setWidth(4);
        painter.setPen(pen);
        int ct = 0;
        for (int j=samples_begin; j<samples_end; ++j)
        {
            QPointF p = QPointF(ct*wd_scale, wv_vector_get(row,j));
            points[ct] = p;
            ++ct;
        }
        painter.drawPolyline(points, sample_range);
        wv_vector_free(row);
    }

    //Don't forget to free the matrices!
    wv_matrix_free(waves_shown);
    wv_matrix_free(submat);
}
void WaveWidget::DrawTemplates()
{
    const std::map<int, wv_vector*>& templates = _waveTemplates->Templates(),
        & template_sds = _waveTemplates->TemplateSDs();
    int samples_begin = SamplesBegin(),
            samples_end = SamplesEnd(),
            sample_range = samples_end - samples_begin;

    //Find min, max of all template values, get scales
    std::map<int, wv_vector*>::const_iterator it = templates.begin(),
                it_sd = template_sds.begin();
    wv::float_t min = std::numeric_limits<wv::float_t>::max(),
            max = std::numeric_limits<wv::float_t>::min();
    int ct = 0;
    for (auto it : templates) //NB for WaveGui (ClusWaveWidget) this doesn't need to be
        //calculated every draw
    {
        if ( (_isTemplateShown && !_isTemplateShown->at(ct)) || !it.second)
        {
            ++it_sd; ++ct; continue;
        }

        const wv_vector tplt = wv_vector_subvector(it.second, samples_begin, sample_range).vector,
                tplt_sd = wv_vector_subvector(it_sd->second, samples_begin, sample_range).vector;
        wv::float_t sd = wv_stats_mean(tplt_sd.data, tplt_sd.stride, sample_range);
        wv::float_t curr_min = wv_stats_min(tplt.data, tplt.stride, sample_range) - sd;
        wv::float_t curr_max = wv_stats_max(tplt.data, tplt.stride, sample_range) + sd;
        if (curr_min<min) min = curr_min;
        if (curr_max>max) max = curr_max;
        ++it_sd, ++ct;
    }
    wv::float_t ht_scale, wd_scale;
    FindScales(min, max, sample_range, ht_scale, wd_scale);
    int ht = height();

    //Now graph the subselected waves using the subselected colors.
    QPainter painter(this);
    QPen pen;
    QPointF points[256],
            sd_points_hi[256],
            sd_points_lo[256];

    ct = 0;
    for (auto it : templates) //Since std::maps are ordered by key, don't need to explicitly
            //look up cluster gui number.
    {
        if ( (_isTemplateShown && !_isTemplateShown->at(ct)) || !it.second )
        { ++ct; continue; }
        int clus = it.first;

        QColor color( _templateColors.at(ct) );
        pen.setColor(color);

        //Get the data points for the templates and corresponding SD lines
        wv::float_t* wave_template = templates.at(clus)->data,
            * waves_sd = template_sds.at(clus)->data;
        int ct_j = 0;
        for (int j=samples_begin; j<samples_end; ++j)
        {
            wv::float_t yval = (wave_template[j] - min) * -ht_scale + ht;
            QPointF p = QPointF(ct_j*wd_scale, yval);
            points[ct_j] = p;

            wv::float_t sd = waves_sd[j] * ht_scale;
            QPointF p_sd_hi( p.x(), p.y() + sd );
            QPointF p_sd_lo( p.x(), p.y() - sd );
            sd_points_hi[ct_j] = p_sd_hi;
            sd_points_lo[ct_j] = p_sd_lo;

            ++ct_j;
        }

        //Draw the templates
        pen.setWidth(3);
        painter.setPen(pen);
        painter.drawPolyline(points, sample_range);

        //Draw the SD lines
        pen.setWidth(0);
        painter.setPen(pen);
        painter.drawPolyline(sd_points_hi, sample_range);
        painter.drawPolyline(sd_points_lo, sample_range);

        ++ct;
    }
}

MatCPtr WaveWidget::AllWaves() const
{
    if (_waveData)
        return _waveData->Waveforms(_transform, _tid);
    return MatCPtr();
}
void WaveWidget::SetWavesShown(wv_matrix* waves_shown)
{
    int num_records = MSize1();
    const WSGSpec& wsg_spec = ClusterMediator::Instance().GetCluSpecPtr(_cid)->GetWSGSpec();
    const boost::dynamic_bitset<>& wave_shown = wsg_spec.WaveShown();
    for (int i=0, ct=0; i<num_records; ++i)
    {
        if (!wave_shown.test(i)) continue;
        wv_vector_const_view v = wv_matrix_const_row(AllWaves().get(), i);
        wv_matrix_set_row(waves_shown, ct++, &v.vector);
    }
}

int WaveWidget::NumWavesShown() const
{
    return ClusterMediator::Instance()
                        .GetCluSpecPtr(_cid)->GetWSGSpec().WaveShown().count();
}

const boost::dynamic_bitset<>& WaveWidget::WaveShown() const
{
    return ClusterMediator::Instance().GetCluSpecPtr(_cid)->GetWSGSpec().WaveShown();
}
