#include "clustermediator.h"

ClusterMediator* ClusterMediator::_instance = nullptr;

ClusterMediator& ClusterMediator::Instance()
{
    if (_instance) return *_instance;
    return *(_instance = new ClusterMediator);
}

ClusterMediator::ClusterMediator() : _currentWaveGui(wv::DEFAULT_CID), _lastCID(0)
{
}

void ClusterMediator::AddClustering(wv::CLUSTER_ID cid, const ClassificationCPtr cls)
{
    MutexLocker lock(_mutex);
    ResultSpec result_spec = ResultSpec(cls, cls->BestLevel());
    _results.push_back( std::pair<wv::CLUSTER_ID, ResultSpec>(cid, result_spec) );
}
void ClusterMediator::AddClusterMgr(const ClusterMgrPtr cmgr)
{
    MutexLocker lock(_mutex);
    _clusterMgrs.push_back(cmgr);
}
void ClusterMediator::AddClusters(wv::CLUSTER_ID cid, const std::vector<int>& clusters)
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [&](std::pair<wv::CLUSTER_ID,ResultSpec> record)
                                { return record.first == cid; } );
    it->second.AddClusters(clusters);
}
void ClusterMediator::AddClusterSpec(wv::CLUSTER_ID cid, CluSpecPtr clu_spec)
{
    MutexLocker lock(_mutex);
    _cluSpecs[cid] = clu_spec;
}
void ClusterMediator::AddInputFile(wv::CLUSTER_ID cid, InputFilePtr input_file)
{
    MutexLocker lock(_mutex);
    _inputFiles[cid] = input_file;
    _electrodes[cid] = input_file->VElectrodes().at(0);
}
void ClusterMediator::AddWaveGui(wv::CLUSTER_ID cid, WaveGuiPtr wave_gui)
{
    MutexLocker lock(_mutex);
    _waveGuis.push_back( std::pair<wv::CLUSTER_ID,WaveGuiPtr>(cid, wave_gui) );
}
void ClusterMediator::CopyCID(wv::CLUSTER_ID cid)
{
    _mutex.lock();
    if (cid==wv::DEFAULT_CID)
        throw BadCID("ClusterMediator::CopyCID", cid);
    CluSpecPtr new_clus_spec = _cluSpecs[cid]->Clone();
    _mutex.unlock();
    AddClusterSpec( wv::DEFAULT_CID, new_clus_spec );
    AddInputFile( wv::DEFAULT_CID, _inputFiles[cid] );
    SetElectrode( wv::DEFAULT_CID, _electrodes[cid] );

    auto it_new = std::find_if(_results.begin(), _results.end(),
                           [](std::pair<wv::CLUSTER_ID, ResultSpec > it)
                                { return it.first == wv::DEFAULT_CID; });
    auto it_old = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID, ResultSpec > it)
                                { return it.first == cid; });
    it_new->second = it_old->second;
}
wv::CLUSTER_ID ClusterMediator::IncrementCID()
{
    MutexLocker lock(_mutex);
    return ++_lastCID;
}
void ClusterMediator::InitSpecs(wv::CLUSTER_ID cid)
{
    AddClusterSpec(cid, CluSpecPtr( new ClusterSpec ));
    _results.push_back( std::pair<wv::CLUSTER_ID, ResultSpec>(cid, ResultSpec()) );
}
void ClusterMediator::PurgeClusterMgr(wv::CLUSTER_ID cid)
{
    MutexLocker lock(_mutex);
    std::remove_if(_clusterMgrs.begin(), _clusterMgrs.end(),
                   [=](ClusterMgrPtr iter)
        {
            return !iter || (iter->HasCID(cid) && iter->NumThreadsRemaining()==0);
        });
            //Not sure exactly how !iter happens
}
void ClusterMediator::PurgeWaveData(wv::CLUSTER_ID cid)
{
    MutexLocker lock(_mutex);
    for (auto& it : _inputFiles)
    {
        wv::CLUSTER_ID cid_i = it.first;
        if (cid_i == cid) continue;
        short elc = _electrodes.at(cid_i);
        WaveData* wave_data = it.second->WData(elc);
        if (wave_data->IsPreprocFinished() && wave_data->ThreadCount()==0)
            wave_data->Unload();
    }
}
void ClusterMediator::RemoveCID(wv::CLUSTER_ID cid)
{
    MutexLocker lock(_mutex);

    std::remove_if(_results.begin(), _results.end(), [=](std::pair<wv::CLUSTER_ID,ResultSpec> iter)
        { return iter.first == cid; });
    _cluSpecs.erase(cid);
    _electrodes.erase(cid);
    _inputFiles.erase(cid);
    std::remove_if(_waveGuis.begin(), _waveGuis.end(), [=](std::pair<wv::CLUSTER_ID,WaveGuiPtr> iter)
        { return iter.first == cid; });
}

void ClusterMediator::SetClusAccepted(wv::CLUSTER_ID cid, wv_vector_int* clus_accepted)
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::SetClusAccepted", cid);
    it->second.SetClusAccepted(clus_accepted);
}
void ClusterMediator::SetClusLevel(wv::CLUSTER_ID cid, int level)
{
    MutexLocker lock(_mutex);
    if (cid==wv::DEFAULT_CID)
        throw BadCID("ClusterMediator::SetClusLevel", cid);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::SetClusLevel", cid);
    it->second.SetLevel(level);
}
void ClusterMediator::SetElcDefaultCid(InputFilePtr input_file, short electrode, wv::CLUSTER_ID cid)
{
    MutexLocker lock(_mutex);
    if (cid==wv::DEFAULT_CID) throw BadCID("ClusterMediator::SetElcDefaultCid", cid);
    std::pair<InputFilePtr,short> p(input_file, electrode);
    _elecrodeDefaultCids[p] = cid;
}
void ClusterMediator::SetElectrode(wv::CLUSTER_ID cid, short electrode)
{
    MutexLocker lock(_mutex);
    _electrodes[cid] = electrode;
}
void ClusterMediator::SetStats(wv::CLUSTER_ID cid, int level, const std::vector<wv::float_t>& stats)
{
    MutexLocker lock(_mutex);
    if (cid==wv::DEFAULT_CID)
        throw BadCID("ClusterMediator::SetStats", cid);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::SetStats", cid);
    it->second.SetStats(level, stats);
}

const boost::dynamic_bitset<>& ClusterMediator::Accepted(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::Accepted", cid);
    return it->second.Accepted();
}
const wv_vector_int* ClusterMediator::ClusAccepted(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::ClusAccepted", cid);
    return it->second.ClusAccepted();
}
int ClusterMediator::ClusLevel(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::ClusLevel", cid);
    return it->second.Level();
}
wv::float_t ClusterMediator::ClusLevelVal(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    if (cid==wv::DEFAULT_CID) throw BadCID("ClusterMediator::ClusLevelVal", cid);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::ClusLevelVal", cid);
    return it->second.LevelVal();
}
wv::CLUSTER_ID ClusterMediator::ElcDefaultCid(InputFilePtr input_file, short electrode) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_elecrodeDefaultCids.begin(), _elecrodeDefaultCids.end(),
                           [=](std::pair< std::pair<InputFilePtr,short>,wv::CLUSTER_ID> i)
                        {
                                return i.first.first==input_file && i.first.second==electrode;
                        } );
    if (it == _elecrodeDefaultCids.end())
        throw BadCID("ClusterMediator::ElcDefaultCid", electrode);
    return it->second;
}
short ClusterMediator::Electrode(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    return _electrodes.at(cid);
}
ClassificationCPtr ClusterMediator::GetClassification(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::GetClassification", cid);
    return it->second.GetClassification();
}
wv::CLUSTER_ID ClusterMediator::GetClusterId(const WaveGuiPtr wave_gui) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_waveGuis.begin(), _waveGuis.end(),
                           [=](std::pair<wv::CLUSTER_ID,WaveGuiPtr> i) { return i.second==wave_gui; } );
    if (it == _waveGuis.end())
        throw BadWaveGui("ClusterMediator::GetClusterId", wave_gui.get());
    return it->first;
}
const CluSpecPtr ClusterMediator::GetCluSpecPtr(wv::CLUSTER_ID cid) const
{
    //Note that wv::DEFAULT_CID will pick out the cluster specs currently operative in WaveSpaceGui
    MutexLocker lock(_mutex);
    auto it = _cluSpecs.find(cid);
    if (it == _cluSpecs.end())
            throw BadCID("ClusterMediator::GetCluSpecPtr", cid);
    return it->second;
}
short ClusterMediator::GetElectrode(wv::CLUSTER_ID cid) const
{
    //Note that wv::DEFAULT_CID will pick out the electrode currently selected in ElectrodeGui
    MutexLocker lock(_mutex);
    auto it = _electrodes.find(cid);
    if (it == _electrodes.end())
        throw BadCID("ClusterMediator::GetElectrode", cid);
    return it->second;
}
const InputFilePtr ClusterMediator::GetInputFile(wv::CLUSTER_ID cid) const
{
    //Note that wv::DEFAULT_CID will pick out the InputFile in WaveSpaceGui in this case
    MutexLocker lock(_mutex);
    auto it = _inputFiles.find(cid);
    if (it == _inputFiles.end())
        throw BadCID("ClusterMediator::GetInputFile", cid);
    return it->second;
}
WaveData* ClusterMediator::GetWaveData(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    InputFilePtr input_file = _inputFiles.at(cid);
    short elc = _electrodes.at(cid);
    return input_file->WData(elc);
}
const WaveGuiPtr ClusterMediator::GetWaveGui(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    if (cid==wv::DEFAULT_CID) throw BadCID("ClusterMediator::GetWaveGui", cid);
    auto it = std::find_if(_waveGuis.begin(), _waveGuis.end(),
                           [=](std::pair<wv::CLUSTER_ID,WaveGuiPtr> i) { return i.first==cid; } );
    if (it == _waveGuis.end())
        throw BadCID("ClusterMediator::GetWaveGui", cid);
    return it->second;
}
bool ClusterMediator::HasDefaultElc(InputFilePtr input_file, short electrode) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_elecrodeDefaultCids.begin(), _elecrodeDefaultCids.end(),
                           [=](std::pair< std::pair<InputFilePtr,short>,wv::CLUSTER_ID> i)
                        {
                            return i.first.first==input_file && i.first.second==electrode;
                        } );
    return it != _elecrodeDefaultCids.end();
}
bool ClusterMediator::HasElectrode(short elc) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_electrodes.begin(), _electrodes.end(),
                           [=](std::pair<wv::CLUSTER_ID,short> i) { return i.second==elc; } );
    return it != _electrodes.end();
}
bool ClusterMediator::HasElectrodeByCid(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_electrodes.begin(), _electrodes.end(),
                           [=](std::pair<wv::CLUSTER_ID,short> i) { return i.first==cid; } );
    return it != _electrodes.end();
}
bool ClusterMediator::HasResults(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end()) return false;
    return (it->second.Level() != -1);
}
bool ClusterMediator::HasWaveGui(wv::CLUSTER_ID cid) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_waveGuis.begin(), _waveGuis.end(),
                           [=](std::pair<wv::CLUSTER_ID,WaveGuiPtr> i) { return i.first==cid; } );
    return it != _waveGuis.end();
}
wv::CLUSTER_ID ClusterMediator::LastCID() const
{
    MutexLocker lock(_mutex);
    return _lastCID;
}
const std::vector<wv::float_t>& ClusterMediator::Stats(wv::CLUSTER_ID cid, int level) const
{
    MutexLocker lock(_mutex);
    auto it = std::find_if(_results.begin(), _results.end(),
                           [=](std::pair<wv::CLUSTER_ID,ResultSpec> i) { return i.first==cid; } );
    if (it == _results.end())
        throw BadCID("ClusterMediator::Stats", cid);
    return it->second.Stats(level);
}
