#ifndef CLUSTERMEDIATOR_H
#define CLUSTERMEDIATOR_H

#include "../globals/globals.h"
#include "specs/clusterspec.h"
#include "specs/resultspec.h"
#include "../file/input/inputfilebase.h"
#include "../analysis/threads/mgr/clustermgr.h"
#include "../gui/mdi/wavegui.h"

//This class manages the relationships between the Clusterings, WaveGuis, WaveSpaceGui, and
//Input files/WaveData objects.  A singleton.
class MainWindow;
class ClusterMediator
{
    public:
        static ClusterMediator& Instance();

        void AddClustering(wv::CLUSTER_ID cid, const ClassificationCPtr cls);
        void AddClusterMgr(const ClusterMgrPtr cmgr);
        void AddClusters(wv::CLUSTER_ID cid, const std::vector<int>& clusters);
        void AddClusterSpec(wv::CLUSTER_ID cid, CluSpecPtr clu_spec);
        void AddInputFile(wv::CLUSTER_ID cid, InputFilePtr input_file);
            //This also sets cid's electrode to the first one in the set
        void AddWaveGui(wv::CLUSTER_ID cid, WaveGuiPtr wave_gui);
        void CopyCID(wv::CLUSTER_ID cid); //Copies everything in this cid to wv::DEFAULT_CID
        wv::CLUSTER_ID IncrementCID();
        void InitSpecs(wv::CLUSTER_ID cid);
        void PurgeClusterMgr(wv::CLUSTER_ID cid); //Only removes it if all of its threads have completed
        void PurgeWaveData(wv::CLUSTER_ID cid); //cid is spared;
            //all others unloaded if all threads have completed
        void RemoveCID(wv::CLUSTER_ID cid);

        void SetClusAccepted(wv::CLUSTER_ID cid, wv_vector_int* clus_accepted);
        void SetClusLevel(wv::CLUSTER_ID cid, int level);
        void SetCurrentWaveGui(wv::CLUSTER_ID cid) { _currentWaveGui = cid; }
        void SetElcDefaultCid(InputFilePtr input_file, short electrode, wv::CLUSTER_ID cid);
            //### The InputFilePtr instances need to be **deleted** when the last non-default input file
            //is destroyed, otherwise a huge memory leak
        void SetElectrode(wv::CLUSTER_ID cid, short electrode);
        void SetMainWindow(MainWindow* mw) { _mainWindow = mw; }
        void SetStats(wv::CLUSTER_ID cid, int level, const std::vector<wv::float_t>& stats);

        const boost::dynamic_bitset<>& Accepted(wv::CLUSTER_ID cid) const;
        const wv_vector_int* ClusAccepted(wv::CLUSTER_ID cid) const;
        int ClusLevel(wv::CLUSTER_ID cid) const;
        wv::float_t ClusLevelVal(wv::CLUSTER_ID cid) const; //The actual value (e.g., temperature) corresponding to the level index
        wv::CLUSTER_ID CurrentWaveGui() const { return _currentWaveGui; }
        wv::CLUSTER_ID ElcDefaultCid(InputFilePtr input_file, short electrode) const;
        short Electrode(wv::CLUSTER_ID cid) const;
        ClassificationCPtr GetClassification(wv::CLUSTER_ID cid) const;
        wv::CLUSTER_ID GetClusterId(const WaveGuiPtr wave_gui) const;
        const CluSpecPtr GetCluSpecPtr(wv::CLUSTER_ID cid) const;
        short GetElectrode(wv::CLUSTER_ID cid) const;
        const InputFilePtr GetInputFile(wv::CLUSTER_ID cid) const;
        MainWindow* GetMainWindow() { return _mainWindow; }
        WaveData* GetWaveData(wv::CLUSTER_ID cid) const;
        const WaveGuiPtr GetWaveGui(wv::CLUSTER_ID cid) const;
        bool HasDefaultElc(InputFilePtr input_file, short electrode) const;
        bool HasElectrode(short elc) const;
        bool HasElectrodeByCid(wv::CLUSTER_ID cid) const;
        bool HasResults(wv::CLUSTER_ID cid) const;
        bool HasWaveGui(wv::CLUSTER_ID cid) const;
        wv::CLUSTER_ID LastCID() const;
        const std::vector<wv::float_t>& Stats(wv::CLUSTER_ID cid, int level) const;

    private:
        ClusterMediator();
#ifndef Q_OS_WIN
        ClusterMediator(const ClusterMediator&) = delete;
        ClusterMediator& operator=(const ClusterMediator&) = delete;
#else
        ClusterMediator(const ClusterMediator&);
        ClusterMediator& operator=(const ClusterMediator&);
#endif
        static ClusterMediator* _instance;

        MainWindow* _mainWindow; //ClusterTOs need this to send the results back to the Main Window
#ifndef Q_OS_WIN
        VecPairs<wv::CLUSTER_ID, ResultSpec > _results;
#else
        std::vector< std::pair<wv::CLUSTER_ID, ResultSpec> > _results;
#endif
        std::list<ClusterMgrPtr> _clusterMgrs; //Pushed on to the list when they start, erased when the
            //clustering is complete.  Critical or else they get deleted.
        std::map<wv::CLUSTER_ID, CluSpecPtr> _cluSpecs; //wv::DEFAULT_CID picks
            //the current ClusterSpecs on WaveSpaceGui
        wv::CLUSTER_ID _currentWaveGui; //Whichever WaveGui has focus (if there is one)
        std::map< std::pair<InputFilePtr, short>, wv::CLUSTER_ID> _elecrodeDefaultCids;
        std::map<wv::CLUSTER_ID, short> _electrodes;
        std::map<wv::CLUSTER_ID, InputFilePtr> _inputFiles; //wv::DEFAULT_CID picks
            //the current InputFile on WaveSpaceGui
        wv::CLUSTER_ID _lastCID; //CLUSTER_IDs aren't recycled, this number ticks up
            //every time a new clustering is launched
#ifndef Q_OS_WIN
        VecPairs<wv::CLUSTER_ID, WaveGuiPtr> _waveGuis;
#else
        std::vector< std::pair<wv::CLUSTER_ID, WaveGuiPtr> > _waveGuis;
#endif

        mutable std::mutex _mutex;
};

#endif // CLUSTERMEDIATOR_H
