#include "biellipse.h"

const wv::float_t BiEllipse::DEFAULT_ELLIPSE_SD = 2.0;
const wv::float_t BiEllipse::GRAIN = 0.03;

BiEllipse::BiEllipse() : Circle(),
    _rotMatrix(nullptr), _sd(DEFAULT_ELLIPSE_SD), _xSD(0.0), _ySD(0.0)
{
}

BiEllipse::BiEllipse(ID id)
    : Circle(id), _rotMatrix(nullptr), _sd(DEFAULT_ELLIPSE_SD), _xSD(0.0), _ySD(0.0)
{
}

BiEllipse::BiEllipse(const BiEllipse& bi_ellipse) : Circle(bi_ellipse)
{
    Copy(bi_ellipse);
}
BiEllipse& BiEllipse::operator=(const BiEllipse& bi_ellipse)
{
    if (this != &bi_ellipse)
    {
        DrawLineBase::operator=(bi_ellipse);
        Copy(bi_ellipse);
    }
    return *this;
}

BiEllipse::~BiEllipse()
{
    if (_rotMatrix) wv_matrix_free(_rotMatrix);
}

void BiEllipse::InitEllipse(const wv_vector* x, const wv_vector* y, const Circle* circle)
{
    //Get bitset of waveforms included in *circle*,
    int num_records = x->size;
    _inDistribution = boost::dynamic_bitset<>(num_records);
    for (int i=0; i<num_records; ++i)
        if ( circle->Contains( wv_vector_get(x, i), wv_vector_get(y, i) ) )
            _inDistribution.set(i, true);

    //Make the matrix consisting of the data points contained in the circle, in
    //order to get axes of rotation
    int num_circ_pts = _inDistribution.count();
    wv_matrix* circ_pts = wv_matrix_alloc(num_circ_pts, 2);
    for (int i=0, ct=0; i<num_records; ++i)
        if (_inDistribution.test(i))
        {
            wv_matrix_set(circ_pts, ct, 0, wv_vector_get(x, i) );
            wv_matrix_set(circ_pts, ct, 1, wv_vector_get(y, i) );
            ++ct;
        }

    //De-mean the data
    wv_vector_view circ_x = wv_matrix_column(circ_pts, 0),
            circ_y = wv_matrix_column(circ_pts, 1);
    _xCenter = wv_stats_mean(circ_x.vector.data, circ_x.vector.stride, num_circ_pts);
    _yCenter = wv_stats_mean(circ_y.vector.data, circ_y.vector.stride, num_circ_pts);
    wv_vector_add_constant(&circ_x.vector, -_xCenter);
    wv_vector_add_constant(&circ_y.vector, -_yCenter);

    //Rotate these points so that the axis with the most variance is horizontal
    wv_matrix* rotated_pts = wv_matrix_calloc(num_circ_pts, 2);

    if (!_rotMatrix) _rotMatrix = wv_matrix_alloc(2, 2);
#ifdef USE_GSL
    mgsl::PCATransform(circ_pts, rotated_pts, _rotMatrix);
#else
    msl::PCATransform(circ_pts, rotated_pts, false, _rotMatrix);
#endif
    //The eigenvectors are the rotation matrix

    //Create the axis lines, one horizontal and one vertical.
    //onto which each point will be projected
    wv_vector* xline_pt1 = wv_vector_alloc(2),
            * xline_pt2 = wv_vector_alloc(2),
            * yline_pt1 = wv_vector_alloc(2),
            * yline_pt2 = wv_vector_alloc(2);

    wv_vector_set(xline_pt1, 0, 0); wv_vector_set(xline_pt1, 1, 0);
    wv_vector_set(xline_pt2, 0, 1); wv_vector_set(xline_pt2, 1, 0);
    wv_vector_set(yline_pt1, 0, 0); wv_vector_set(yline_pt1, 1, 0);
    wv_vector_set(yline_pt2, 0, 0); wv_vector_set(yline_pt2, 1, 1);

    //Get the projection/distance of each point onto the two lines, and store the results in a vector
    wv_vector* xprojns = wv_vector_alloc(num_circ_pts),
            * yprojns = wv_vector_alloc(num_circ_pts),
            * xdists_to_line = wv_vector_alloc(num_circ_pts),
            * ydists_to_line = wv_vector_alloc(num_circ_pts);
#ifdef USE_GSL
    mgsl::PointsToNLine(xline_pt1, xline_pt2, rotated_pts, xprojns, xdists_to_line);
    mgsl::PointsToNLine(yline_pt1, yline_pt2, rotated_pts, yprojns, ydists_to_line);
#else
    msl::PointsToNLine(xline_pt1,
                      xline_pt2,
                      rotated_pts,
                      xprojns,
                      xdists_to_line);
    msl::PointsToNLine(yline_pt1,
                      yline_pt2,
                      rotated_pts,
                      yprojns,
                      ydists_to_line);
#endif

    //Get the centers and SDs for the two dimensions
    _xSD = wv_stats_sd_m(xprojns->data, xprojns->stride, num_circ_pts, 0);
    _ySD = wv_stats_sd_m(yprojns->data, yprojns->stride, num_circ_pts, 0);
//    qDebug() << _xCenter << _yCenter << _xSD << _ySD << rotation;

    //Free everything
    wv_matrix_free(rotated_pts);
    wv_vector_free(xline_pt1);
    wv_vector_free(xline_pt2);
    wv_vector_free(yline_pt1);
    wv_vector_free(yline_pt2);
    wv_vector_free(xprojns);
    wv_vector_free(yprojns);
    wv_vector_free(xdists_to_line);
    wv_vector_free(ydists_to_line);

    //Update the ellipse
    UpdateBiEllipse();
}

void BiEllipse::SetSD(wv::float_t sd)
{
    _sd = sd;
    UpdateBiEllipse();
}

void BiEllipse::Copy(const BiEllipse& bi_ellipse)
{
    _rotMatrix = wv_matrix_alloc(2, 2);
    wv_matrix_memcpy(_rotMatrix, bi_ellipse._rotMatrix);
}

void BiEllipse::UpdateBiEllipse()
{
    //This is where the circle pts are calculated...
    int i=0;
    wv::float_t M_PI2 = 2.0 * M_PI;
    _length = (int) (M_PI2 / GRAIN);
    wv_vector* vec = wv_vector_alloc(2),
            * rot_vec = wv_vector_alloc(2);
    for (wv::float_t theta=0.0; theta<M_PI2 && i<_length; theta += GRAIN)
    {
        wv::float_t x =  _sd * _xSD * cos( theta ),
                y = _sd * _ySD * sin( theta );
        wv_vector_set(vec, 0, x);
        wv_vector_set(vec, 1, y);
#ifdef USE_GSL
        gsl_blas_dgemv(CblasNoTrans, 1.0, _rotMatrix, vec,
                       0.0, rot_vec);
        _points[i++] = QPointF(wv_vector_get(rot_vec, 0) + _xCenter,
                               wv_vector_get(rot_vec, 1) + _yCenter);
#else
        msl::blas_gemv(CblasRowMajor, CblasNoTrans,
                    _rotMatrix->size1, _rotMatrix->size2,
                    (wv::float_t)1.0,
                    _rotMatrix->data, _rotMatrix->tda,
                    vec->data, vec->stride,
                    (wv::float_t)0.0,
                    rot_vec->data, rot_vec->stride);
        _points[i++] = QPointF(wv_vector_get(rot_vec, 0) + _xCenter,
                               wv_vector_get(rot_vec, 1) + _yCenter);
#endif
    }
    wv_vector_free(vec);
    wv_vector_free(rot_vec);
}
