#include "resultspec.h"

ResultSpec::ResultSpec() : _level(-1), _vAcceptedClus(nullptr)
{
}
ResultSpec::ResultSpec(ClassificationCPtr cls, int level) : _cls(cls), _level(level), _vAcceptedClus(nullptr)
{
    //Initialized _accepted; by default, all clusters for each level are selected.
    const std::vector< std::vector<int> >& clusterings = _cls->Clusters();
    int num_clusterings = clusterings.size();
    int num_records = _cls->VClusters().at(0).second->size;
    for (int i=0; i<num_clusterings; ++i)
    {
        std::vector<int> clusters = clusterings.at(i);
        int num_clusters = clusters.size();
        std::unordered_set<int> selected_clusters;
        for (int j=0; j<num_clusters; ++j)
            selected_clusters.insert( clusters.at(j) );
        _acceptedClus[i] = selected_clusters;

        _isAccepted[i] = boost::dynamic_bitset<>(num_records);
        _isAccepted[i].set();

        _stats[i] = std::vector<wv::float_t>();
    }
    _vAcceptedClus = wv_vector_int_alloc(num_records);
    wv_vector_int_set_all(_vAcceptedClus, wv::DEFAULT_CLUSTER);
}
ResultSpec::ResultSpec(const ResultSpec& result_spec)
{
    Copy(result_spec);
}
ResultSpec& ResultSpec::operator=(const ResultSpec& result_spec)
{
    if (this!=&result_spec) Copy(result_spec);
    return *this;
}
void ResultSpec::Copy(const ResultSpec& result_spec)
{
    _acceptedClus = result_spec._acceptedClus;
    _cls = result_spec._cls;
    _isAccepted = result_spec._isAccepted;
    _level = result_spec._level;
    _stats = result_spec._stats;

    if (_cls)
    {
        int num_records = _cls->VClusters().at(0).second->size;
        if (result_spec._vAcceptedClus)
        {
            _vAcceptedClus = wv_vector_int_alloc(num_records);
            wv_vector_int_memcpy(_vAcceptedClus, result_spec._vAcceptedClus);
                //So there can be no stored pointer to _vAcceptedClus!
        }
        else
            _vAcceptedClus = nullptr;
    }
    else
        _vAcceptedClus = nullptr;
}

ResultSpec::~ResultSpec()
{
    if (_vAcceptedClus) wv_vector_int_free(_vAcceptedClus);
}

void ResultSpec::AddClusters(const std::vector<int>& clusters)
{
    _acceptedClus[_level].empty();
    for (const auto& clus : clusters)
        _acceptedClus[_level].insert(clus);
}

void ResultSpec::SetClusAccepted(wv_vector_int* clus_accepted)
{
    if (_vAcceptedClus) wv_vector_int_free(_vAcceptedClus);
    if ((_vAcceptedClus = clus_accepted)) CalcAccepted();
}
void ResultSpec::SetLevel(int level)
{
    if (level<0 || level>=(int)_acceptedClus.size())
        throw BadValue("ResultSpec::SetLevel", std::to_string(level));
    _level = level;
}
void ResultSpec::SetStats(int level, const std::vector<wv::float_t>& stats)
{
    if (level<0 || level>=(int)_acceptedClus.size())
        throw BadValue("ResultSpec::SetStats", std::to_string(level));
    _stats[level] = stats;
}

const wv_vector_int* ResultSpec::ClusAccepted() const
{
    return _vAcceptedClus;
}
wv::float_t ResultSpec::LevelVal() const
{
    return _cls->VClusters().at(_level).first;
}
const std::vector<wv::float_t>& ResultSpec::Stats(int level) const
{
    if (level<0 || level>=(int)_acceptedClus.size())
        throw BadValue("ResultSpec::Stat", std::to_string(level));
    return _stats.at(level);
}

void ResultSpec::CalcAccepted()
{
    boost::dynamic_bitset<>& is_accepted = _isAccepted[_level];
    int num_records = _vAcceptedClus->size;
    for (int i=0; i<num_records; ++i)
        is_accepted.set(i, wv_vector_int_get(_vAcceptedClus, i)==wv::ACCEPT_CLUSTER );
}
