#include "wavespaceguispec.h"
#include "../clustermediator.h"

const wv::float_t WaveSpaceGuiSpec::DEFAULT_LUM_SCALE = 1.0;
const int WaveSpaceGuiSpec::DEFAULT_HIST_DISK_SIZE = 20;
#ifdef Q_OS_WIN
const int WaveSpaceGuiSpec::NUM_DISKS = 5;
#else
const int WaveSpaceGuiSpec::NUM_DISKS;
#endif
const wv::float_t WaveSpaceGuiSpec::ZOOM_INCREMENT = 1.075;

WaveSpaceGuiSpec::WaveSpaceGuiSpec()
    : _activeAxis(WSGSpec::Y), _clusterBegin(DOUBLE_MAX), _clusterEnd(DOUBLE_MAX),
       _currentTime(-1.0), _currentValue(0.0),
      _histAxisZoom(1.0), _histLimitMode(TRANSFORM), _histLumScale(DEFAULT_LUM_SCALE),
      _histPrevLumScale(DEFAULT_LUM_SCALE), _histZoomMode(LUMINANCE),
      _isScatter(false), _movieState(STOPPED), _samplesBegin(-1), _samplesEnd(-1), _sampleX(0), _sampleY(0),
      _tidGui(wv::DEFAULT_TID), _tidX(wv::DEFAULT_TID), _tidY(wv::DEFAULT_TID),
      _transformWaves(wv::UNKNOWN_TRANSFORM), _transformX(wv::UNKNOWN_TRANSFORM), _transformY(wv::UNKNOWN_TRANSFORM),
      _userTCIndex(-1), _userTidIndex(-1),
      _viewX(new wv_vector_const_view()), _viewY(new wv_vector_const_view()), _waveTemplatesWSG(nullptr)
{
    for (int i=0; i<WSGSpec::NUM_DISKS; ++i)
        _histDiskSize[i] = DEFAULT_HIST_DISK_SIZE;
}

WaveSpaceGuiSpec::WaveSpaceGuiSpec(const WaveSpaceGuiSpec& wsg_spec)
{
    Copy(wsg_spec);
}
WaveSpaceGuiSpec& WaveSpaceGuiSpec::operator=(const WaveSpaceGuiSpec& wsg_spec)
{
    if (this != &wsg_spec) Copy(wsg_spec);
    return *this;
}
void WaveSpaceGuiSpec::Copy(const WaveSpaceGuiSpec& wsg_spec)
{
    _activeAxis =           wsg_spec._activeAxis;
    _clusterBegin =         wsg_spec._clusterBegin;
    _clusterEnd =           wsg_spec._clusterEnd;
    _currentTime =          wsg_spec._currentTime;
    _currentValue =         wsg_spec._currentValue;
    _histAxisZoom =         wsg_spec._histAxisZoom;
    _histCursorMode =       wsg_spec._histCursorMode;
    _histLimitMode =        wsg_spec._histLimitMode;
    _histLumScale =         wsg_spec._histLumScale;
    _histPrevLumScale =     wsg_spec._histPrevLumScale;
    _histZoomMode =         wsg_spec._histZoomMode;
    _isAccepted =           wsg_spec._isAccepted;
    _isScatter =            wsg_spec._isScatter;
    _isWaveShown =          wsg_spec._isWaveShown;
    _movieState =           wsg_spec._movieState;
    _sampleX =              wsg_spec._sampleX;
    _sampleY =              wsg_spec._sampleY;
    _samplesBegin =         wsg_spec._samplesBegin;
    _samplesEnd =           wsg_spec._samplesEnd;
    _templateColorsWG =     wsg_spec._templateColorsWG;
    _templateColorsWSG =    wsg_spec._templateColorsWSG;
    _tidGui =               wsg_spec._tidGui;
    _tidWaves =             wsg_spec._tidWaves;
    _tidX =                 wsg_spec._tidX;
    _tidY =                 wsg_spec._tidY;
    _timeBegin =            wsg_spec._timeBegin;
    _timeEnd =              wsg_spec._timeEnd;
    _transformGui =         wsg_spec._transformGui;
    _transformWaves =       wsg_spec._transformWaves;
    _transformX =           wsg_spec._transformX;
    _transformY =           wsg_spec._transformY;
    _userTCIndex =          wsg_spec._userTCIndex;
    _userTidIndex =         wsg_spec._userTidIndex;
    _viewX =                new wv_vector_const_view( *wsg_spec._viewX );
    _viewY =                new wv_vector_const_view( *wsg_spec._viewY );
    _waveData =             wsg_spec._waveData;
    _waveTemplatesWSG =        wsg_spec._waveTemplatesWSG
                                ? new WaveTemplate( *wsg_spec._waveTemplatesWSG )
                                : nullptr;
    _window =               wsg_spec._window;
    _windowBegin =          wsg_spec._windowBegin;
    _windowEnd =            wsg_spec._windowEnd;
    _xMin =                 wsg_spec._xMin;
    _xMax =                 wsg_spec._xMax;
    _yMin =                 wsg_spec._yMin;
    _yMax =                 wsg_spec._yMax;

    for (int i=0; i<NUM_DISKS; ++i)
    {
        _diskCursors[i] =   wsg_spec._diskCursors[i];
        _histDiskSize[i] =  wsg_spec._histDiskSize[i];
    }

    for (auto it : wsg_spec._waveTemplatesWG)
    {
        int clus = it.first;
        WaveTemplate* wave_template = it.second;
        _waveTemplatesWG[clus] = wave_template ? new WaveTemplate(*wave_template)
                                               : nullptr;
    }
}

WaveSpaceGuiSpec::~WaveSpaceGuiSpec()
{
    for (auto it : _waveTemplatesWG) if (it.second) delete it.second;
    if (_waveTemplatesWSG) delete _waveTemplatesWSG;
    delete _viewX;
    delete _viewY;
}

void WaveSpaceGuiSpec::InitWaveShown()
{
    int num_records = _waveData->NumRecords();
    _isWaveShownDefault = boost::dynamic_bitset<>(num_records);
    int step = std::max( num_records/WaveGui::DEFAULT_NUM_SHOWN, 1);
    for (int i=0; i<num_records; i+=step)
        _isWaveShownDefault.set(i, true);
    _isWaveShown = _isWaveShownDefault;
}
void WaveSpaceGuiSpec::MakeTemplates(const std::vector<int>& clusters, const wv_vector_int* clus_accepted,
                                     wv::CLUSTER_ID cid,
                                   int level_idx, const std::vector<QColor>& color_palette)
{
    bool is_for_wsg = level_idx == -1;
        //If this is for WaveSpaceGui templates, otherwise it's for a WaveGui
    int num_clusters = clusters.size();

    if (is_for_wsg)
    {
        //Make templates
        MatCPtr mat_waves = _waveData->Waveforms(_transformWaves, _tidWaves);
        _waveTemplatesWSG = new WaveTemplate(mat_waves, _window,
                                             clusters, clus_accepted);

        _templateColorsWSG = std::vector<QColor>(num_clusters);
        _templateColorsWSG[0] = Qt::green;
        //wv::ACCEPT_CLUSTERS has the lowest enum value. so we can count on this
        for (int i=1; i<num_clusters; ++i)
            _templateColorsWSG[i] = Qt::white;
    }
    else //It's for the WaveGui
    {
        //Only make the templates for the WaveGui once.
        if ( _waveTemplatesWG.find(level_idx)!=_waveTemplatesWG.end() && _waveTemplatesWG[level_idx] )
            return;

        //Check palette size
        int palette_sz = color_palette.size();
        if (palette_sz==0)
            throw BadWaveGui("WaveSpaceGuiSpec::MakeTemplates", NULL);

        //Set the templates
        if (_window.size()==0) _window = boost::dynamic_bitset<>(clus_accepted->size);
            //This can happen on the first call from a WaveGui, if the clustering is loaded from file
        boost::dynamic_bitset<> window(_window.size());
        window.set(); //For WaveGui templates, all waveforms are used
        MatCPtr waves = ClusterMediator::Instance().GetWaveData(cid)->Waveforms(wv::RAW);
            //Always use raw waveforms for WaveGui templates
        _waveTemplatesWG[level_idx]
                = new WaveTemplate(waves, window, clusters, clus_accepted);

        //Set the colors
        _templateColorsWG[level_idx] = std::vector<QColor>(num_clusters);
        for (int i=0; i<num_clusters; ++i)
            _templateColorsWG[level_idx][i] = color_palette.at( i%palette_sz );
    }
}
void WaveSpaceGuiSpec::ResetDisk(int i)
{
    _diskCursors[i] = boost::dynamic_bitset<>( Timestamps()->size1 );
}
void WaveSpaceGuiSpec::ResetView(AXIS axis) const
{
    switch (axis)
    {
        case X:
            if (MatX())
            {
                delete _viewX;
                _viewX = new wv_vector_const_view( wv_matrix_const_column(MatX().get(), _sampleX) );
            }
            break;
        case Y:
            if (MatY())
            {
                delete _viewY;
                _viewY = new wv_vector_const_view( wv_matrix_const_column(MatY().get(), _sampleY) );
            }
            break;
    }
}
void WaveSpaceGuiSpec::ResetViews() const
{
    ResetView(WSGSpec::X);
    ResetView(WSGSpec::Y);
}

//All in native coordinates, no scaling/zooming here.
void WaveSpaceGuiSpec::SetDiskCursor(int selected_disk, wv::float_t x, wv::float_t y, wv::float_t xdist, wv::float_t ydist)
{
    const wv_vector* xvec = HistVector(WSGSpec::X),
            * yvec = HistVector(WSGSpec::Y);
    int num_records = xvec->size;
    wv_vector* xvec2 = wv_vector_alloc(num_records),
            * yvec2 = wv_vector_alloc(num_records);
    wv::float_t range_x = Range(WSGSpec::X),
            range_y = Range(WSGSpec::Y);
    wv::float_t dist = xdist*xdist/(range_x*range_x) + ydist*ydist/(range_y*range_y);

    wv_vector_memcpy(xvec2, xvec);
    wv_vector_memcpy(yvec2, yvec);

    wv_vector_add_constant(xvec2, -x);
    wv_vector_add_constant(yvec2, -y);
    wv_vector_scale(xvec2, 1.0/range_x);
    wv_vector_scale(yvec2, 1.0/range_y);

#ifdef QT_DEBUG
    QTime tic;
    tic.start();
#endif

    wv::float_t xlo =  -xdist/range_x,
            xhi = xdist/range_x,
            ylo = -ydist/range_y,
            yhi = ydist/range_y;
    _diskCursors[selected_disk] = boost::dynamic_bitset<>(num_records);
    boost::dynamic_bitset<>& disk_cursor = _diskCursors[selected_disk];
    for (int i=0; i<num_records; ++i)
    {
        wv::float_t xval = wv_vector_get(xvec2, i),
                yval = wv_vector_get(yvec2, i);
        if (xval>=xlo && xval<=xhi && yval>=ylo && yval<=yhi)
            //Only check records which are sufficiently close to (x, y)--saves time
        {
            wv::float_t di = xval*xval + yval*yval;
            if (di<=dist) disk_cursor.set(i, true);
        }
    }

#ifdef QT_DEBUG
    qDebug() << " WaveSpaceGuiSpec::SetDiskCursor(): " << num_records << ", " << tic.elapsed();
#endif

    wv_vector_free(xvec2);
    wv_vector_free(yvec2);
}
void WaveSpaceGuiSpec::SetGuiTransform(wv::TRANSFORM transform)
{
    _transformGui = transform;
    if (transform==wv::USER)
        _tidGui = _waveData->UserTid(_userTidIndex);
    else
        _tidGui = wv::DEFAULT_TID;
}

void WaveSpaceGuiSpec::SetRange(WSGSpec::AXIS axis)
{
    wv::float_t* min(nullptr), * max(nullptr);
    switch (axis)
    {
        case X:
            min = &_xMin;
            max = &_xMax;
            break;
        case Y:
            min = &_yMin;
            max = &_yMax;
            break;
    }

    const wv_vector* vec = HistVector(axis);
    wv::float_t min_raw, max_raw;
    int num_records = vec->size;
    wv_stats_minmax(&min_raw, &max_raw, vec->data, vec->stride, num_records);
    wv::float_t mean = wv_stats_mean(vec->data, vec->stride, num_records),
            sd = wv_stats_sd_m(vec->data, vec->stride, num_records, mean),
            sd_limit;
    try
    {
        sd_limit = boost::lexical_cast<wv::float_t>( session_profile.Par("Session", "Scrub threshold (SDs)") );
    }
    catch(...)
    {
        Log("WARNING: Invalid value for Scrub threshold set, using 3.0");
        sd_limit = 3;
    }

    *min = qMax(min_raw, mean - sd_limit * sd);
    *max = qMin(max_raw, mean + sd_limit * sd);
}
void WaveSpaceGuiSpec::SetUserTidIndex(int user_tid_index)
{
    _userTidIndex = user_tid_index;
    if (_transformGui==wv::USER) _tidGui = user_tid_index;
}

void WaveSpaceGuiSpec::SetSample(AXIS axis, int sample)
{
    switch (axis)
    {
        case X:
            _sampleX = sample;
            break;
        case Y:
            _sampleY = sample;
            break;
    }
    ResetView(axis);
    if (_transformX!=wv::UNKNOWN_TRANSFORM && _transformY!=wv::UNKNOWN_TRANSFORM)
        SetRange(axis);
}
void WaveSpaceGuiSpec::SetTransform(AXIS axis, wv::TRANSFORM transform, int sample)
{
    wv::TRANSFORM_ID tid = (transform==wv::USER) ? UserTid() : wv::DEFAULT_TID;
    switch (axis)
    {
        case X:
            _transformX = transform;
            _tidGui = _tidX = tid;
            SetSample(X, sample);
            break;
        case Y:
            _transformY = transform;
            _tidGui = _tidY = tid;
            SetSample(Y, sample);
            break;
    }
}
void WaveSpaceGuiSpec::SetWavesTransform(wv::TRANSFORM waves_transform, wv::TRANSFORM_ID tid)
{
    ClusterMediator& cmed = ClusterMediator::Instance();
    _transformWaves = waves_transform;
    if (waves_transform==wv::USER)
    {
        if (tid==wv::DEFAULT_TID)
            _tidWaves = _tidGui;
        else
            _tidWaves = cmed.GetCluSpecPtr(tid)->TransformId();
    }
    else
        _tidWaves = wv::DEFAULT_TID;
}
void WaveSpaceGuiSpec::SetWindowLimits(int window_size)
{
    //So, the general idea here is that histogram creation essentially sorts all of the records by
    //their x and y data values.  Therefore, to get window size/window limits, we do the same sort
    //and find the records #'s which define the begin/end of the sort

    const wv_vector* value_vector;
    wv::float_t current_value;
    switch (_histLimitMode)
    {
        case TRANSFORM:
            value_vector = HistVector(WSGSpec::X);
            current_value = _currentValue;
            break;
        case TIMESTAMP:
            wv_vector_const_view view = wv_matrix_const_column(Timestamps().get(), 0);
            value_vector = &view.vector;
            current_value = _currentTime;
            break;
    }

    int num_records = value_vector->size;
    static wv_vector* sorting_vec = wv_vector_alloc(num_records);
    if (sorting_vec->size != (size_t)num_records)
    {
        wv_vector_free(sorting_vec);
        sorting_vec = wv_vector_alloc(num_records);
    }
        //This can happen at least in theory, since sorting_vec is static

    //Doing this right--size the interval correctly in either direction so as to
    //get window_size records in it.
    wv_vector_memcpy(sorting_vec, value_vector);
#ifdef USE_GSL
    gsl_sort_vector(sorting_vec);
#else
    std::sort(sorting_vec->data, sorting_vec->data+num_records); //Measured std::sort to be >3x faster than qsort, gcc and clang
#endif
    Index idx(sorting_vec);

    int win_sz2 = window_size/2,
        i = idx(current_value),
            idx_begin = i-win_sz2 + 0.5,
            idx_end = i+win_sz2 + 0.5;
    idx_begin = qBound(0, idx_begin, num_records-1);
    idx_end = qBound(idx_begin+1, idx_end, num_records);
    switch (_histLimitMode)
    {
        case TRANSFORM:
            _windowBegin = wv_vector_get(sorting_vec, idx_begin);
            _windowEnd = wv_vector_get(sorting_vec, idx_end-1);
            break;
        case TIMESTAMP:
            _timeBegin = wv_vector_get(sorting_vec, idx_begin);
            _timeEnd = wv_vector_get(sorting_vec, idx_end-1);
            break;
    }

    //Now create the window, this should be done here while we have idx_begin and idx_end
    _window = boost::dynamic_bitset<>(num_records);
    ut::FillBitset(_window, true, idx_begin, idx_end);
}

WaveSpaceGuiSpec::AXIS WaveSpaceGuiSpec::Opposite(AXIS axis) const
{
    AXIS opp;
    switch (axis)
    {
        case X:
            opp = Y;
            break;
        case Y:
            opp = X;
            break;
    }
    return opp;
}
const wv_vector* WaveSpaceGuiSpec::HistVector(AXIS axis) const
{
    const wv_vector* vec;
    switch (axis)
    {
        case X:
            ResetView(X);
            vec = &_viewX->vector;
            break;
        case Y:
            ResetView(Y);
            vec = &_viewY->vector;
            break;
    }
        //So note that this requires that _viewX, _viewY are always properly
        //updated, done through SetSample
    return vec;
}
wv_vector_const_view WaveSpaceGuiSpec::HistView(AXIS axis) const
{
//    const wv_vector* vec;
    switch (axis)
    {
        case X:
            ResetView(X);
            return *_viewX;
        case Y:
            ResetView(Y);
            return *_viewY;
    }
    return *_viewX; //Just to avoid 'control reached' warnings
}
wv::TRANSFORM WaveSpaceGuiSpec::Transform(AXIS axis) const
{
    wv::TRANSFORM transform;
    switch (axis)
    {
        case X:
            transform = _transformX;
            break;
        case Y:
            transform = _transformY;
            break;
    }
    return transform;
}
wv::TRANSFORM_ID WaveSpaceGuiSpec::TransformId(AXIS axis) const
{
    wv::TRANSFORM_ID tid;
    switch (axis)
    {
        case X:
            tid = _tidX;
            break;
        case Y:
            tid = _tidY;
            break;
    }
    return tid;
}
wv::float_t WaveSpaceGuiSpec::Timestamp(int i) const
{
    wv::float_t tm;
    if (i==-1)
    {
        int last = Timestamps()->size1 - 1;
        tm = wv_matrix_get(Timestamps().get(), last, 0);
    }
    else
        tm = wv_matrix_get(Timestamps().get(), i, 0);
    return tm;
}
wv::float_t WaveSpaceGuiSpec::Value(AXIS axis, int record) const
{
    return wv_vector_get( HistVector(axis), record);
}
int WaveSpaceGuiSpec::Sample(AXIS axis) const
{
    int sample;
    switch (axis)
    {
        case X:
            sample = _sampleX;
            break;
        case Y:
            sample = _sampleY;
            break;
    }
    return sample;
}
wv::float_t WaveSpaceGuiSpec::Min(AXIS axis) const
{
    wv::float_t min;
    switch (axis)
    {
        case X:
            min = _xMin;
            break;
        case Y:
            min = _yMin;
            break;
    }
    return min;
}
wv::float_t WaveSpaceGuiSpec::Max(AXIS axis) const
{
    wv::float_t max;
    switch (axis)
    {
        case X:
            max = _xMax;
            break;
        case Y:
            max = _yMax;
            break;
    }
    return max;
}
wv::float_t WaveSpaceGuiSpec::Range(AXIS axis) const
{
    return Max(axis) - Min(axis);
}
int WaveSpaceGuiSpec::NearestWave(wv::float_t x, wv::float_t y) const
{
//    int num_records = _histX->vector.size;
    int num_records = Timestamps()->size1;
    wv::float_t range_x = Range(WSGSpec::X),
            range_y = Range(WSGSpec::Y),
            min_x = Min(WSGSpec::X),
            min_y = Min(WSGSpec::Y);
    wv::float_t xpct = (x - min_x) / range_x,
            ypct = (y - min_y) / range_y;
    wv::float_t xlo = xpct - 0.05,
            xhi = xpct + 0.05,
            ylo = ypct - 0.05,
            yhi = ypct + 0.05;

    //Get the x, y vectors and scale them so that distance comparisons are normalized
    const wv_vector* xvals = HistVector(WSGSpec::X),
            * yvals = HistVector(WSGSpec::Y);
//    const wv_vector* xvals = &_histX->vector,
//            * yvals = &_histY->vector;
    wv_vector* xscaled = wv_vector_alloc(num_records),
            * yscaled = wv_vector_alloc(num_records);
    wv_vector_memcpy(xscaled, xvals);
    wv_vector_memcpy(yscaled, yvals);
    wv_vector_add_constant(xscaled, -min_x);
    wv_vector_add_constant(yscaled, -min_y);
    wv_vector_scale(xscaled, 1.0/range_x);
    wv_vector_scale(yscaled, 1.0/range_y);

    //Find the closest record using simple euclidean distance.
#ifdef USE_GSL
    wv::float_t min_dist = GSL_DBL_MAX;
#else
    wv::float_t min_dist = std::numeric_limits<wv::float_t>::max();
#endif
    int min_dist_idx = -1;
    for (int i=0; i<num_records; ++i)
    {
        wv::float_t xval = wv_vector_get(xscaled, i),
                yval = wv_vector_get(yscaled, i);
        if (xval>=xlo && xval<=xhi && yval>=ylo && yval<=yhi)
            //Only check records which are sufficiently close to (x, y)--saves time
        {
#ifdef USE_GSL
            wv::float_t dist = gsl_hypot(xpct-xval, ypct-yval);
#else
            wv::float_t dist = std::hypot(xpct-xval, ypct-yval);
#endif
            if (dist<min_dist)
            {
                min_dist = dist;
                min_dist_idx = i;
            }
        }
    }
    wv_vector_free(xscaled);
    wv_vector_free(yscaled);
    return min_dist_idx;
}

MatCPtr WaveSpaceGuiSpec::MatGui() const
{
    if (_waveData)
        return _waveData->Waveforms(_transformGui, _tidGui);
    return MatCPtr();
}
MatCPtr WaveSpaceGuiSpec::MatWaves() const
{
    if (_waveData)
        return _waveData->Waveforms(_transformWaves, _tidWaves);
    return MatCPtr();
}
MatCPtr WaveSpaceGuiSpec::MatX() const
{
    if (_waveData)
        return _waveData->Waveforms(_transformX, _tidX);
    return MatCPtr();
}
MatCPtr WaveSpaceGuiSpec::MatY() const
{
    if (_waveData)
        return _waveData->Waveforms(_transformY, _tidY);
    return MatCPtr();
}
