#ifndef WAVESPACEGUISPEC_H
#define WAVESPACEGUISPEC_H

#include "../../globals/globals.h"
#include "../../memrep/drawing/polyline.h"
#include "../../memrep/drawing/circle.h"
#include "../../memrep/parameters/cluster/clusterparamsbase.h"
#include "../../memrep/wavedata.h"
#include "../../memrep/wavetemplate.h"

//This should really hold literally *every piece* of information that should be stored by cid
//and recovered every time the corresponding WaveGui regains focus.
// ***Only*** setter/getter, add/remove methods in this and all specs classes
class WaveSpaceGuiSpec
{
    friend class WaveSpaceGui;
    friend class ClusterSpec;
    friend class WaveGui;
    friend class WaveWidget;
    friend class GridWaveWidget;

    public:
        WaveSpaceGuiSpec();
        ~WaveSpaceGuiSpec();

        static const wv::float_t DEFAULT_LUM_SCALE;
        static const int DEFAULT_HIST_DISK_SIZE;
#ifdef Q_OS_WIN
        static const int NUM_DISKS;
#else
        static const int NUM_DISKS = 5;
#endif
        static const wv::float_t ZOOM_INCREMENT;

        enum AXIS //General purpose, to minimize code duplication
        {
            X,
            Y
        };
        enum HIST_CURSOR_MODE
        {
            BAR,
            DISK
        };
        enum HIST_LIMIT_MODE //Are we windowing by transform value or by time?
        {
            TRANSFORM,
            TIMESTAMP
        };
        enum HIST_ZOOM_MODE //More accurately, scroll wheel mode
        {
            AXIS_ZOOM,
            RESOLUTION,
            LUMINANCE,
            DISK_SIZE
        };
        enum MOVIE_STATE
        {
            STOPPED,
            PLAYING
        };

        struct PLData
        {
            wv::TRANSFORM transform_x, transform_y;
            int sample_x, sample_y;
            Polyline polyline;
            boost::dynamic_bitset<> is_ok;
        };

        void InitWaveShown();
        void MakeTemplates(const std::vector<int>& clusters, const wv_vector_int* clus_accepted,
                           wv::CLUSTER_ID cid, int level_idx = -1, const std::vector<QColor>& color_palette = std::vector<QColor>());
            //NB: There are *two* sets of templates associated with a given cid, one for the clusterings in the WaveGui
            //and one for the templates in the WaveSpaceGui.  Each call of this function makes *one* of them, so it
            //must be called separately from WaveSpaceGui and WaveGui.
        void ResetDisk(int i);
        void ResetView(AXIS axis) const;
        void ResetViews() const;

        void SetActiveAxis(AXIS active_axis) { _activeAxis = active_axis; }
        void SetClusBegin(wv::float_t begin) { _clusterBegin = begin; }
        void SetClusEnd(wv::float_t end) { _clusterEnd = end; }
        void SetCurrentTime(int current_time) { _currentTime = current_time; }
        void SetCurrentValue(int current_value) { _currentValue = current_value; }
        void SetDiskCursor(int selected_disk, wv::float_t x, wv::float_t y, wv::float_t xdist, wv::float_t ydist);
        void SetGuiTransform(wv::TRANSFORM transform);
        void SetGuiTransformId(wv::TRANSFORM_ID tid) { _tidGui = tid; }
        void SetHistAxisZoom(wv::float_t zoom) { _histAxisZoom = zoom; }
        void SetHistCursorMode(HIST_CURSOR_MODE mode) { _histCursorMode = mode; }
        void SetHistDiskSize(int i, int size) { _histDiskSize[i] = size; }
        void SetHistLimitMode(HIST_LIMIT_MODE hist_limit_mode) { _histLimitMode = hist_limit_mode; }
        void SetHistLumScale(wv::float_t scale) { _histLumScale = scale; }
        void SetHistPrevLumScale(wv::float_t scale) { _histPrevLumScale = scale; }
        void SetHistZoomMode(HIST_ZOOM_MODE mode) { _histZoomMode = mode; }
        void SetIsScatter(bool is_scatter) { _isScatter = is_scatter; }
        void SetMovieState(MOVIE_STATE state) { _movieState = state; }
        void SetSamplesBegin(int samples_begin) { _samplesBegin = samples_begin; }
        void SetSamplesEnd(int samples_end) { _samplesEnd = samples_end; }
        void SetSample(AXIS axis, int sample);
        void SetTimeLimits(int window_size); //Sets begin, end of the histogram time interval, via an appropriate
            //call to SetWindowLimits
        void SetTransform(AXIS axis, wv::TRANSFORM transform, int sample = 0);
            //I.e., by default don't change the sample number
            //Parameterize on cid since when clustering is loaded from file, it need not pass through WaveGuiSpace
        void SetWavesTransform(wv::TRANSFORM waves_transform, wv::TRANSFORM_ID tid = wv::DEFAULT_TID);
        void SetWindowLimits(int window_size);
            //Sets begin, end of the histogram window

        const boost::dynamic_bitset<>& Accepted() const { return _isAccepted; }
        AXIS ActiveAxis() const { return _activeAxis; }
        wv::TRANSFORM ActiveTransform() const { return Transform( _activeAxis ); }
        wv::TRANSFORM_ID ActiveTransformId() const { return TransformId( _activeAxis ); }
        wv::float_t ClusterBegin() const { return _clusterBegin; }
        wv::float_t ClusterEnd() const { return _clusterEnd; }
        wv::float_t CurrentTime() const { return _currentTime; }
        wv::float_t CurrentValue() const { return _currentValue; }
        const boost::dynamic_bitset<>* DiskCursors() const { return _diskCursors; }
        const WaveTemplate* GetWaveTemplatesWG(int level_idx) const { return _waveTemplatesWG.at(level_idx); }
        const WaveTemplate* GetWaveTemplatesWSG() const { return _waveTemplatesWSG; }
        wv::TRANSFORM GuiTransform() const { return _transformGui; }
        wv::TRANSFORM_ID GuiTransformId() const { return _tidGui; }
        wv::float_t HistAxisZoom() const { return _histAxisZoom; }
        HIST_CURSOR_MODE HistCursorMode() const { return _histCursorMode; }
        int* HistDiskSize() { return _histDiskSize; }
        int HistDiskSize(int i) const { return _histDiskSize[i]; }
        HIST_LIMIT_MODE HistLimitMode() const { return _histLimitMode; }
        wv::float_t HistLumScale() const { return _histLumScale; }
        wv::float_t HistPrevLumScale() const { return _histPrevLumScale; }
        const wv_vector* HistVector(AXIS axis) const;
        wv_vector_const_view HistView(AXIS axis) const;
        HIST_ZOOM_MODE HistZoomMode() const { return _histZoomMode; }
        AXIS InactiveAxis() const { return Opposite(_activeAxis); }
        wv::TRANSFORM InactiveTransform() const { return Transform( InactiveAxis() ); }
        bool IsAccepted(int i) const { return _isAccepted.test(i); }
        bool IsScatter() const { return _isScatter; }
        wv::float_t Max(AXIS axis) const;
        wv::float_t Min(AXIS axis) const;
        MOVIE_STATE MovieState() const { return _movieState; }
        int NearestWave(wv::float_t x, wv::float_t y) const;
        AXIS Opposite(AXIS axis) const;
        wv::float_t Range(AXIS axis) const;
        int Sample(AXIS axis) const;
        int SamplesBegin() const { return _samplesBegin; }
        int SamplesEnd() const { return _samplesEnd; }
        const std::vector<QColor>& TemplateColorsWG(int level_idx) const { return _templateColorsWG.at(level_idx); }
        const std::vector<QColor>& TemplateColorsWSG() const { return _templateColorsWSG; }
        wv::float_t TimeBegin() const { return _timeBegin; }
        wv::float_t TimeEnd() const { return _timeEnd; }
        wv::float_t Timestamp(int i) const; //Set i==-1 to get the last timestamp
        MatCPtr Timestamps() const { return _waveData->Waveforms(wv::TIMESTAMP); }
        wv::TRANSFORM Transform(AXIS axis) const;
        wv::TRANSFORM_ID TransformId(AXIS axis) const;
        int UserTCIndex() const { return _userTCIndex; }
        wv::TRANSFORM_ID UserTid() const { return _waveData->UserTid(_userTidIndex); }
        int UserTidIndex() const { return _userTidIndex; }
        wv::float_t Value(AXIS axis, int record) const;
        const boost::dynamic_bitset<>& WaveShown() const { return _isWaveShown; }
        wv::TRANSFORM WavesTransform() const { return _transformWaves; }
        wv::TRANSFORM_ID WavesTransformId() const { return _tidWaves; }
        const boost::dynamic_bitset<>& Window() const { return _window; }
        wv::float_t WindowBegin() const { return _windowBegin; }
        wv::float_t WindowEnd() const { return _windowEnd; }

    private:
        WaveSpaceGuiSpec(const WaveSpaceGuiSpec&);
        WaveSpaceGuiSpec& operator=(const WaveSpaceGuiSpec&);
        void Copy(const WaveSpaceGuiSpec& wsg_spec);

        class Index
        {
            public:
                Index(const wv_vector* vec) : _vec(vec) //vec must be **sorted** in ascending order, and have
                    // **stride==1**
                {
                }
                int operator()(wv::float_t val) //Returns the array index for the given value
                {
                    int i = 0;
                    int num_records = _vec->size;
                    while (i<num_records && wv_vector_get(_vec, i++)<val) ;
                    return i;
                }
            private:
                const wv_vector* _vec;
        };

        void SetIsAccepted(int i, bool b) { _isAccepted[i] = b; }
        void SetRange(AXIS axis);
        void SetUserTCIndex(int user_tc_index) { _userTCIndex = user_tc_index; }
        void SetUserTidIndex(int user_tid_index);
        void SetWaveData(WaveData* wave_data) { _waveData = wave_data; }

        boost::dynamic_bitset<>& Accepted() { return _isAccepted; }
        MatCPtr MatGui() const;
        MatCPtr MatWaves() const;
        MatCPtr MatX() const;
        MatCPtr MatY() const;
        const boost::dynamic_bitset<>& DefaultWaveShown() { return _isWaveShownDefault; }
        boost::dynamic_bitset<>& WaveShown() { return _isWaveShown; }

        AXIS _activeAxis; //This is the axis whose values are updated when the Waveform slider is moved
        wv::float_t _clusterBegin, _clusterEnd; //The boundaries, in *x value coordinates* of the region used in the most recent clustering analysis
            //except for Manual clustering.  Only valid for the specific sample used in the clustering.
        wv::float_t _currentTime, //Selected by the time slider
                _currentValue; //Selected by whichever HistWidget slider is active
        boost::dynamic_bitset<> _diskCursors[5];//NUM_DISKS]; //What's inside the selection disk (if in disk mode).  We don't set window limits of any
                    //kind here, it's just calculated directly
        wv::float_t _histAxisZoom;
        HIST_CURSOR_MODE _histCursorMode;
        int _histDiskSize[5];//[NUM_DISKS];
        HIST_LIMIT_MODE _histLimitMode;
        wv::float_t _histLumScale, _histPrevLumScale;
        HIST_ZOOM_MODE _histZoomMode;
        boost::dynamic_bitset<> _isAccepted;
        bool _isScatter;
        boost::dynamic_bitset<> _isWaveShown, _isWaveShownDefault;
        MOVIE_STATE _movieState;
        int _samplesBegin, _samplesEnd, //In the ScanWaveWidget, the range of samples shown
            _sampleX, _sampleY; //The sample # whence the transform vector was pulled for the given
            //axis
        std::vector<QColor> _templateColorsWSG; //Green for accepted, white otherwise
        std::map<int, std::vector<QColor> > _templateColorsWG;
        wv::TRANSFORM_ID _tidGui, _tidWaves, _tidX, _tidY;
            //_tidGui is wv::DEFAULT_TID until a user transform has been created in the gui, then
            //it has that transform's TID--REGARDLESS of whether that transform is actually being
            //used in the current histogram or not.  _tidWaves is the tid of the transform in the Waves
            //window.  _tidX and _tidY are one of exactly two things,
            //wv::DEFAULT_TID or _tidGui
        wv::float_t _timeBegin, _timeEnd; //Time interval boundaries.  *Inclusive*, default is entire window
        wv::TRANSFORM _transformGui, //The transform currently selected, need not be in use
            _transformWaves, //Which values are shown in WaveWidget
            _transformX, _transformY;
        int _userTCIndex, _userTidIndex; //Which component of the transform is shown in the gui,
            //and the value of the TID spin box
        mutable const wv_vector_const_view* _viewX, * _viewY;
            //It turns out, because of GSL's inflexibility with wv_vector_const_view, declaring these
            //on the heap is the only way to manage const vectors without endless copying
        std::map<int, WaveTemplate*> _waveTemplatesWG; //Correspond to the clusterings, defined over all waveforms (entire
            //session) start out as NULL, filled in as they're requested so that they only need to be calculated once.
        WaveData* _waveData;
        WaveTemplate* _waveTemplatesWSG; //For the WSG, green/white, defined wrt the current selection
        boost::dynamic_bitset<> _window; //All the waves in the selection window (bar), shown or unshown
        wv::float_t _windowBegin, _windowEnd; //Histogram window boundaries by value
        wv::float_t _xMin, _xMax, _yMin, _yMax;
};
    //The idea is that the information here, plus the original waveform records
    //and other parameters, let the user recreate the analysis exactly, up to the
    //point of pressing 'Cluster'.  This spec covers everything determined by the
    //user input to this gui.
    //  These are stored in std::map<wv::CLUSTER_ID, WaveSpaceGui::Spec> MainWindow::_waveSpaceGuiSpecs


typedef WaveSpaceGuiSpec::HIST_ZOOM_MODE HIST_ZOOM_MODE;
typedef WaveSpaceGuiSpec::HIST_CURSOR_MODE HIST_CURSOR_MODE;

typedef WaveSpaceGuiSpec WSGSpec;

#endif // WAVESPACEGUISPEC_H
