#include "wavedata.h"
#include "../file/input/wfmfilein.h"
#include "../file/output/wfmfileout.h"
#include "../file/wfmfileappend.h"
#include "../file/input/waves/waveformsinbase.h"

#define BOOST_FILESYSTEM_VERSION 3

std::vector<Signature> WaveData::_utSignatures;

const int WaveData::NUM_BUFFER_ROWS = 2*4096;
const int WaveData::NUM_TBUFFER_ROWS = WaveData::NUM_BUFFER_ROWS * 64; //Note, 64 is an *estimate* of the number
    //of samples there will be.

WaveData::WaveData(const WaveformsInBase* waves_file, short electrode, wv::float_t us_per_samp)
    : _allTimestamps(nullptr), _electrode(electrode), _fpTimestamps(nullptr), _fpWaves(nullptr), _fullPath(waves_file->Path()),
      _hasOnlineClus(false), _isLoaded(true), _isPreprocFinished(false),
      _loadTimeBuffer(nullptr), _loadWavesBuffer(nullptr), _loadTimeCt(0), _loadWavesCt(0),
      _numRecords(-1),
      _onlineVCluster(nullptr), _onlineClusterCID(wv::NO_CLUSTER_ID), _threadCount(0), _usPerSamp(us_per_samp),
      _wavesFile(waves_file)
{
    StringVec elts = ut::Split<StringVec>(_fullPath, "/");
    _fileName = elts.back();
    elts.pop_back();
    _filePath = boost::algorithm::join(elts, "/");
//    boost::filesystem::path bpath = boost::filesystem::path(full_path); //elts.begin(), elts.end());
//    _fileName = bpath.filename().c_str();
//    _filePath = bpath.parent_path().c_str();

    _loadTimeBuffer = wv_matrix_alloc(NUM_TBUFFER_ROWS, 1);
    _loadWavesBuffer = wv_matrix_alloc(NUM_BUFFER_ROWS, waves_file->NumSamples());

    _transforms.push_back(wv::HEIGHT);
    _transforms.push_back(wv::WIDTH);
    _transforms.push_back(wv::SLOPE);
    _transforms.push_back(wv::SLOPE_D);
    _transforms.push_back(wv::ENERGY);
    _transforms.push_back(wv::TIMESTAMP);
    _transforms.push_back(wv::USER);
    _transforms.push_back(wv::RAW);
    _transforms.push_back(wv::RAW_D);
#ifdef USE_GSL
    _transforms.push_back(wv::SPLINE);
    _transforms.push_back(wv::SPLINE_D);
#endif
    _transforms.push_back(wv::PCA);
    _transforms.push_back(wv::PCA_D);
#ifdef USE_GSL
    _transforms.push_back(wv::FOURIER);
    _transforms.push_back(wv::FOURIER_D);
#endif
    _transforms.push_back(wv::WAVELET);
    _transforms.push_back(wv::WAVELET_D);

    _isCalculated[wv::HEIGHT] = false;
    _isCalculated[wv::WIDTH] = false;
    _isCalculated[wv::SLOPE] = false;
    _isCalculated[wv::SLOPE_D] = false;
    _isCalculated[wv::ENERGY] = false;
    _isCalculated[wv::TIMESTAMP] = false;
    _isCalculated[wv::USER] = false;
    _isCalculated[wv::RAW] = false;
    _isCalculated[wv::RAW_D] = false;
    _isCalculated[wv::PCA] = true;
    _isCalculated[wv::PCA_D] = true;
#ifdef USE_GSL
    _isCalculated[wv::SPLINE] = true;
    _isCalculated[wv::SPLINE_D] = true;
    _isCalculated[wv::FOURIER] = true;
    _isCalculated[wv::FOURIER_D] = true;
#endif
    _isCalculated[wv::WAVELET] = true;
    _isCalculated[wv::WAVELET_D] = true;
        //Note that rawD is *not* counted among transforms that need to be calculated.

    for (auto it : _transforms) _have[it] = false;
    for (auto it : _transforms) _maxima[it] = std::numeric_limits<wv::float_t>::min();
    for (auto it : _transforms) _minima[it] = std::numeric_limits<wv::float_t>::max();
    for (auto it : _transforms) _threads[it] = nullptr;
    for (auto it : _transforms) _transformTOs[it] = nullptr;
    for (auto it : _transforms) _want[it] = false;
    for (auto it : _transforms) _waves[it] = MatPtr(nullptr);

    _want[wv::RAW] = _want[wv::TIMESTAMP] = _want[wv::HEIGHT]
            = _want[wv::WIDTH] = _want[wv::SLOPE] = _want[wv::ENERGY] = true;
}
WaveData::~WaveData()
{
    TerminateProcesses();
    FreeAll();
    if (_onlineVCluster) wv_vector_int_free(_onlineVCluster);
    if (_allTimestamps) wv_vector_int_free(_allTimestamps);
}

void WaveData::AddToUser(wv::TRANSFORM_ID tid, wv_vector_const_view view, wv::TRANSFORM transform,
                         int sample, bool update_signatures)
{
    UserTransformSpec& ut_spec = UTSpec(tid);
    if (ut_spec.NumSpecs()>0 && !ut_spec.Waves())
        ut_spec.SetWaves( LoadTransform(wv::USER, tid) ); //Needs to be reloaded before we add anything
        //to it
    ut_spec.AddVector(view, transform, sample);
    if (update_signatures) UpdateUTSignatures();
    WriteUserTransform(tid);
}
bool WaveData::CheckUT(const Signature& signature) const
{
    for (const auto& it : signature)
    {
        wv::TRANSFORM transform = it.first;
        if (!_have.at(transform)) return false;
    }
    return true;
}
void WaveData::AllocateMemory(int num_records)
{
    if (_hasOnlineClus) _onlineVCluster = wv_vector_int_alloc(num_records);
}
wv::TRANSFORM_ID WaveData::CreateUserTransform(wv::TRANSFORM_ID old_id)
{
    UTS user_spec;
    _userTransforms.push_back(user_spec);
    wv::TRANSFORM_ID tid = user_spec.ID();
    if (old_id!=wv::DEFAULT_TID) //In this case, we are essentially copying the previous transform
    {
        const UTS& old_spec = UTSpec(old_id);
        int num_utrans = old_spec.NumSpecs();
        for (int j=0; j<num_utrans; ++j)
        {
            wv::TRANSFORM transform = old_spec.UserSpec(j).first;
            int sample = old_spec.UserSpec(j).second;
            wv_vector_const_view view = wv_matrix_const_column(old_spec.Waves().get(), j);
            AddToUser(tid, view, transform, sample);
        }
    }
    return tid;
}
wv::TRANSFORM_ID WaveData::CreateUserTransform(const UTS& other_uspec)
{
    UTS user_spec;
    _userTransforms.push_back(user_spec);
    wv::TRANSFORM_ID tid = user_spec.ID();
    int num_utrans = other_uspec.NumSpecs();
    for (int j=0; j<num_utrans; ++j)
    {
        wv::TRANSFORM transform = other_uspec.UserSpec(j).first;
        size_t sample = other_uspec.UserSpec(j).second;
        MatCPtr waves = Waveforms(transform);
        wv_vector_const_view view = wv_matrix_const_column(waves.get(), sample);
            //This is different from the other version of CreateUserTransform.  We build the new user
            //transform out of the waves matrices, we don't copy in an old user waves matrix column by column.
        AddToUser(tid, view, transform, sample);
    }
    return tid;
}
void WaveData::DeleteFPs()
{
    AppendToFile(wv::RAW, _loadWavesBuffer, _loadWavesCt);
    delete _fpWaves;
    _fpWaves = nullptr;

    AppendToFile(wv::TIMESTAMP, _loadTimeBuffer, _loadTimeCt);
    delete _fpTimestamps;
    _fpTimestamps = nullptr;

    wv_matrix_free(_loadWavesBuffer);
    wv_matrix_free(_loadTimeBuffer);
    _loadWavesBuffer = _loadTimeBuffer = nullptr;
}
void WaveData::DisplayUseCounts() const
{
    qDebug() << "Transform use counts for channel " << Electrode();
    for (auto it : _transforms)
        qDebug() << "\t" << wv::TransformName(it).c_str() << ", " << _waves.at(it).use_count();
}
void WaveData::Finished(void* data)
{
    _mutex.lock();

    auto pair = static_cast< std::pair<wv::TRANSFORM, MatPtr>* >(data);
    wv::TRANSFORM transform = pair->first;
    if (transform==wv::UNKNOWN_TRANSFORM)
    {
        _mutex.unlock();
        throw BadTransform("WaveData::Finished", wv::TransformName(transform));
    }
    _waves[transform] = pair->second;
    pair->second.reset();

    QThread* thread = _threads[transform];
    if (!_waves[transform])
    {
        Log("WARNING: " + wv::TransformName(transform)
            + " analysis failed, null pointer returned.  " \
            "This is most likely caused by insufficient memory.");
        thread->quit();
        _threads[transform] = nullptr;
        --_threadCount;
        _mutex.unlock();
        return;
    }

    TransformTOBase* transform_to = _transformTOs[transform];
    disconnect(transform_to, SIGNAL(Finished(void*)), this, SLOT(Finished(void*)));
        //Unless there is this explicit disconnect, this slot will get spuriously called twice for
        //a single transform when a new clustering has been accepted in between the time it started and
        //and finished.
    Log( wv::TransformName(transform) + " analysis finished after "
         + std::to_string(transform_to->TimeElapsed()/1000.0) + " seconds.");
    _have[transform] = true;
    emit EnableTransform(transform, _electrode);
    thread->quit();
    _threads[transform] = nullptr;
    --_threadCount;
    _minima[transform] = wv_matrix_min( _waves[transform].get() );
    _maxima[transform] = wv_matrix_max( _waves[transform].get() );

    delete transform_to; //This frees the transform
    _transformTOs[transform] = nullptr;
    if (!_isLoaded)
    {
        Unload(transform, false);
        wv::TRANSFORM raw_tf = IsDerivative(transform) ? wv::RAW_D : wv::RAW;
//        qDebug() << _electrode << ", " << this << ", " << wv::TransformName(transform).c_str() << ", " << _waves[raw_tf].use_count();
        Unload(raw_tf, false);
            //This is necessary since when WaveData::Unload is called, wv::RAW, wv::RAW_D will still
            //have use_count>1 and so won't be reset
    }

    _mutex.unlock();
    SetUTsFromSignatures(); //See if the construction of this transform enables any User Transforms
    //now to be built.
}
void WaveData::FreeAll()
{
    MutexLocker lock(_mutex);
#ifdef QT_DEBUG
        int i = 0;
#endif
    for (auto& it : _waves)
    {
        int use_count = it.second.use_count();
#ifdef QT_DEBUG
        qDebug() << Electrode() << ", " << wv::TransformName( _transforms.at(i) ).c_str() << ", "
                 << use_count << ", " << it.second.get();
        ++i;
#endif
        if (use_count==1)
            it.second.reset();
    }
    for (auto& it : _userTransforms)
        if (it.Waves().use_count()==1)
            it.Waves().reset();
}
MatPtr& WaveData::Lookup(wv::TRANSFORM transform, wv::TRANSFORM_ID tid)
{
    if (transform==wv::UNKNOWN_TRANSFORM)
        throw BadTransform("WaveData::Lookup", wv::TransformName(transform));
    if (transform==wv::USER) return UserWaves(tid);
    return _waves[transform];
}
int WaveData::Preprocess()
{
    MutexLocker lock(_mutex);
    _waves[wv::RAW] = LoadTransform(wv::RAW);
    _waves[wv::TIMESTAMP] = LoadTransform(wv::TIMESTAMP);

    int num_records_scrubbed = ScrubArtifacts();
    Log(std::to_string(num_records_scrubbed) + " waveforms scrubbed from " + _fileName);
    RemoveDC();

    WriteToDisk(wv::RAW, _waves[wv::RAW]);
    _have[wv::RAW] = true;

    SetWantedTransforms();
    if (_want[wv::RAW_D]) MakeDiff(); //The derivative of the raw matrix, after 3-pt cumulative
        //square wave smoothing

#ifndef USE_GSL
    MakeEnergy();
#endif

    for (auto it : _transforms)
        if (_want[it] && _isCalculated[it]) IsBatch() ? CalcForBatch(it) : Calculate(it);

    WriteFeatureVectors(); //Timestamp, height, width, etc., everything that's 1-dimensional.
        // *** Mean, std, slope-D, --lots of interesting 1-dim. statistics, create drop-down list?

    _isPreprocFinished = true;

    return _waves[wv::RAW]->size1; //This now has the number of scrubbed records subtracted.
}
void WaveData::ReadUTSignatures(FILE* fp)
{
    _utSignatures.clear();
    int num_uts;
    fscanf(fp, "User Transforms: %d\n", &num_uts);
    for (int i=0; i<num_uts; ++i)
    {
        int num_specs, ct;
        fscanf(fp, "UT %d: %d\n", &ct, &num_specs);
        Signature signature;
        for (int j=0; j<num_specs; ++j)
        {
            char tname[64];
            int sample;
            fscanf(fp, "\t%s\t%d\n", tname, &sample);
            wv::TRANSFORM transform = wv::TransformEnumVal(tname);
            std::pair<wv::TRANSFORM, int> spec(transform, sample);
            signature.push_back(spec);
        }
        _utSignatures.push_back(signature);
    }
}
void WaveData::Reload()
{
    for (size_t i=0; i<_userTransforms.size(); ++i)
    {
        int tid = _userTransforms.at(i).ID();
        _userTransforms[i].SetWaves( LoadTransform(wv::USER, tid) );
    }

    for (auto it : _transforms)
        if (it==wv::RAW_D && _want[it])
        {
            if (_have[it])
                _waves[it] = LoadTransform(it);
            else
                MakeDiff();
        }
        else if (_isCalculated[it] && _want[it])
        {
            if (_have[it])
                _waves[it] = LoadTransform(it);
            else
                IsBatch() ? CalcForBatch(it) : Calculate(it);
        }
    _isLoaded = true;
}
void WaveData::RemoveFromUser(wv::TRANSFORM_ID tid, int col)
{
    auto uiter = std::find_if(_userTransforms.begin(), _userTransforms.end(),
                              [=](const UTS& iter)
                                {
                                    return iter.ID() == tid;
                                }
                            );
    if (uiter==_userTransforms.end())
        throw BadTID("WaveData::RemoveFromUser", tid);
    uiter->RemoveVector(col);
    if (uiter->Waves()) WriteUserTransform(tid);
    UpdateUTSignatures();
}
void WaveData::RemoveUserTransform(int tid)
{
    auto new_end = std::remove_if(_userTransforms.begin(), _userTransforms.end(), [=](UTS& iter)
                                    {
                                        return iter.ID() == tid;
                                    }
                       );
    if (new_end==_userTransforms.end()) return;
    _userTransforms.erase(new_end);
    UpdateUTSignatures();
}
void WaveData::SetupWavefiles(int num_waves, int num_samples)
{
    //Make a dummy matrix so as to supply the appropriate dimensions to WriteToDisk, WfmFileOut
    wv_matrix* waves = wv_matrix_alloc(num_waves, num_samples);
    MatCPtr waves_cptr = MatCPtr(waves, wv_matrix_free);
    WriteToDisk(wv::RAW, waves_cptr, true);

    //Make a dummy matrix so as to supply the appropriate dimensions to WriteToDisk, WfmFileOut
    wv_matrix* timestamps = wv_matrix_alloc(num_waves, 1);
    MatCPtr timestamps_cptr = MatCPtr(timestamps, wv_matrix_free);
    WriteToDisk(wv::TIMESTAMP, timestamps_cptr, true);

    //Create and save the file objects, we keep these pointers open throughout loading
    _fpWaves = new WfmFileAppend(
                FileBase::MakeTempFileName(FileBase::UserPath(), _fileName, wv::RAW, 0, "wfm", Electrode() ) );
    _fpTimestamps = new WfmFileAppend(
                FileBase::MakeTempFileName(FileBase::UserPath(), _fileName, wv::TIMESTAMP, 0, "wfm", Electrode() ) );
}
void WaveData::SetUTsFromSignatures()
{
    bool emit_enabled = false;
    int num_signatures = _utSignatures.size();
    for (int i=0; i<num_signatures; ++i)
    {
        Signature sig = _utSignatures.at(i);

        //If the transform is already loaded, don't create it again
        auto it_sig = std::find_if(_userTransforms.cbegin(), _userTransforms.cend(),
                                   [=](const UTS& uts)
        {
            return sig==uts.UserSpecs();
        });
        if (it_sig!=_userTransforms.cend()) continue;

        wv::TRANSFORM_ID tid = (_userTransforms.size()<=i)
            ? CreateUserTransform()
            : _userTransforms.at(i).ID();
            //If we are now able to fill a previously empty transform, do it; otherwise
            //create a new one.

        if (!CheckUT(sig)) continue;
            //When the underlying transforms haven't all been built yet.  The new transform
            //remains empty.

        for (const auto& it_us : sig)
        {
            wv::TRANSFORM transform = it_us.first;
            MatCPtr waves = Waveforms(transform, tid);
            int sample = it_us.second;
            try
            {
                if (sample>=(int)waves->size2) throw;
                wv_vector_const_view view = wv_matrix_const_column(waves.get(), sample);
                AddToUser(tid, view, transform, sample);
            }
            catch (...)
            {
                Log("Warning: Transform " + wv::TransformName(transform) + ", sample " + std::to_string(sample)
                    + " not available, not adding.");
            }
        }
        emit_enabled = true;
    }

    _have[wv::USER] = true;
    if (emit_enabled) emit EnableTransform(wv::USER, _electrode);

}
void WaveData::TerminateProcesses()
{
    for (auto it : _transforms)
        if ( _isCalculated[it] && !_have[it] && _want[it])
        {
            disconnect(_transformTOs[it],SIGNAL(Finished(void*)),this,SLOT(Finished(void*)));
            if (_threads[it]) _threads[it]->quit();
        }
}
void WaveData::Unload()
{
    FreeAll();
    _isLoaded = false;
}
void WaveData::Unload(wv::TRANSFORM transform, bool lock)
{
    if (lock) _mutex.lock();
//    if (Electrode()==1) DisplayUseCounts();
//    qDebug() << "Unload" << wv::TransformName(transform).c_str() << ", " << Electrode() << ", use_count: " << _waves[wv::RAW].use_count();
    if (_waves[transform].use_count()==1)
        _waves[transform].reset();
    if (lock) _mutex.unlock();
}
void WaveData::WriteTimestamp(uint timestamp)
{
    wv_matrix_set(_loadTimeBuffer, _loadTimeCt++, 0, (wv::float_t)timestamp);
    if (_loadTimeCt%NUM_TBUFFER_ROWS == 0)
    {
        AppendToFile(wv::TIMESTAMP, _loadTimeBuffer);
        _loadTimeCt = 0;
    }
}
void WaveData::WriteUserTransform(wv::TRANSFORM_ID tid) const
{
    MatPtr waves = UserWaves(tid);

    FileBase::WfmHeader header =
    {
        (uchar)0,
        (int)waves->size1,
        (int)waves->size2,
        sizeof(wv::float_t),
        0,
        1.0,
        0.0
    };
    std::string full_path = FileBase::UserPath();
    std::string temp_file = FileBase::MakeTempFileName(full_path, _fileName,
                                              wv::USER, tid, "wfm", Electrode());
    WfmFileOut(temp_file, waves, header).Save();
}
void WaveData::WriteUTSignatures(FILE* fp)
{
    fprintf(fp, "User Transforms: %lu\n", _utSignatures.size());
    int ct = 0;
    for (const auto& it : _utSignatures)
    {
        fprintf(fp, "UT %d: %lu\n", ct++, it.size());
        for (const auto& itsp : it)
        {
            std::string tname = wv::TransformName(itsp.first);
            fprintf(fp, "\t%s\t%d\n", tname.c_str(), itsp.second);
        }
    }
}
void WaveData::WriteWaveform(wv_vector_view wave_view)
{
//    static int i=0;
//    if (i++ % 1000 == 0) qDebug() << "waveform " << i;
    wv_matrix_set_row(_loadWavesBuffer, _loadWavesCt++, &wave_view.vector);
    if (_loadWavesCt%NUM_BUFFER_ROWS == 0)
    {
        AppendToFile(wv::RAW, _loadWavesBuffer);
        _loadWavesCt = 0;
    }
}
void WaveData::WriteWaveform(short wave[], int num_samples)
{
//    static int i=0;
//    if (i++ % 1000 == 0) qDebug() << "waveform " << i;
    for (int j=0; j<num_samples; ++j)
        wv_matrix_set(_loadWavesBuffer, _loadWavesCt, j, (wv::float_t)wave[j]);
    ++_loadWavesCt;
    if (_loadWavesCt%NUM_BUFFER_ROWS == 0)
    {
        AppendToFile(wv::RAW, _loadWavesBuffer);
        _loadWavesCt = 0;
    }
}

void WaveData::SetOnlineClus(int i, int clus)
{
    wv_vector_int_set(_onlineVCluster, i, clus);
}
void WaveData::SetTimestamp(int i, long long int timestamp)
{
    wv_matrix_set(_waves[wv::TIMESTAMP].get(), i, 0, (wv::float_t)timestamp);
}
void WaveData::SetWaveform(int i, short wave[])
{
    int num_samples = _waves[wv::RAW]->size2;
    for (int j=0; j<num_samples; ++j)
        wv_matrix_set(_waves[wv::RAW].get(), i, j, (wv::float_t)wave[j]);
}
void WaveData::SetWaveform(int i, const wv_vector_view wave)
{
    wv_matrix_set_row(_waves[wv::RAW].get(), i, &wave.vector);
}


bool WaveData::HaveTransform(wv::TRANSFORM transform, wv::TRANSFORM_ID tid) const
{
    if (transform==wv::UNKNOWN_TRANSFORM)
        throw BadTransform("WaveData::HaveTransform", wv::TransformName(transform));
    if (transform==wv::USER)
    {
        if (tid==wv::DEFAULT_TID) return _have.at(wv::USER);
        for (auto it : _userTransforms)
            if (it.ID() == tid) return true;
        return false;
    }
    if (transform!=wv::RAW_D && !_isCalculated.at(transform)) return true;
    return _have.at(transform);

}
bool WaveData::IsBatch() const
{
    return _wavesFile->IsBatch();
}
bool WaveData::IsDerivative(wv::TRANSFORM transform) const
{
    std::string name = wv::TransformName(transform);
    return ut::EndsWith(name, "_d");
}
int WaveData::NumSamples(wv::TRANSFORM transform, wv::TRANSFORM_ID tid)
{
    return Waveforms(transform, tid)->size2; //NB, There will probably be transforms of different
        //sizes (e.g. splines) so this number is genuinely transform-specific
}
MatCPtr WaveData::Timestamps()
{
    MutexLocker lock(_mutex);
    if (!_waves[wv::TIMESTAMP]) return LoadTransform(wv::TIMESTAMP);
    return std::static_pointer_cast<const wv_matrix>(
                _waves.at(wv::TIMESTAMP));
}
wv::float_t WaveData::TMin(wv::TRANSFORM transform) const
{
    if (transform==wv::UNKNOWN_TRANSFORM)
        throw BadTransform("WaveData::TMin", wv::TransformName(transform));
    return _minima.at(transform);

}
wv::float_t WaveData::TMax(wv::TRANSFORM transform) const
{
    if (transform==wv::UNKNOWN_TRANSFORM)
        throw BadTransform("WaveData::TMax", wv::TransformName(transform));
    return _maxima.at(transform);
}
std::string WaveData::TransformLabel(wv::TRANSFORM transform, int sample, wv::TRANSFORM_ID tid) const
{
    std::string label_text;
    if (transform==wv::USER)
    {
        if (tid==wv::DEFAULT_TID)
            throw BadTID("WaveData::TransformLabel", tid);
        std::string tr = wv::TransformName( UserTransform(tid, sample) );
        int user_sample = UserSample(tid, sample);
        label_text = "User: " + tr + ", sample " + std::to_string(user_sample);
    }
    else
        label_text = wv::TransformName(transform) + ", sample " + std::to_string(sample);
    return label_text;
}
int WaveData::UserSample(wv::TRANSFORM_ID id, int sample) const
{
    int tr_sample = -1;
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == id)
        {
            tr_sample = _userTransforms.at(i).UserSpec(sample).second;
            break;
        }
    return tr_sample;
}
int WaveData::UserSize(wv::TRANSFORM_ID id) const
{
    int size = 0;
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == id)
        {
            size = _userTransforms.at(i).NumSpecs();
            break;
        }
    return size;
}
wv::TRANSFORM WaveData::UserTransform(wv::TRANSFORM_ID id, int sample) const
{
    wv::TRANSFORM transform = wv::UNKNOWN_TRANSFORM;
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == id)
        {
            transform = _userTransforms.at(i).UserSpec(sample).first;
            break;
        }
    return transform;
}
UTS& WaveData::UTSpec(wv::TRANSFORM_ID tid)
{
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == tid)
            return _userTransforms[i];
    throw BadTID("WaveData::UTSpec", tid);
}
wv::TRANSFORM_ID WaveData::UserTid(size_t index) const
{
    if (index>=_userTransforms.size())
        throw BadUserTransform("WaveData::UserTid", wv::NO_TRANSFORM_ID);
    return _userTransforms.at(index).ID();
}
UTS& WaveData::UTSpecAt(int sample)
{
    if (_userTransforms.empty() || sample<0 || sample>=(int)_userTransforms.size())
        throw BadUserTransform("WaveData::UTSpecAt", sample);
    return _userTransforms[sample];
}
MatCPtr WaveData::Waveforms(wv::TRANSFORM transform, wv::TRANSFORM_ID tid)
{
    MutexLocker lock(_mutex);
    MatPtr& mat_ptr = Lookup(transform, tid);
    if (transform==wv::USER && tid==wv::DEFAULT_TID)
        throw BadUserTransform("WaveData::Waveforms", tid);
    else if (mat_ptr.use_count()==0)
//    {
//          QTime tic;
//          tic.start();
        mat_ptr = LoadTransform(transform);
//        qDebug() << "Wfm file load time: " << tic.elapsed();

//    }
    return std::static_pointer_cast<const wv_matrix>(mat_ptr);
}


void WaveData::SetWantedTransforms()
{
    int num_samples = NumSamples();

    if (num_samples > 3)
    {
        _want[wv::RAW_D] = (session_profile.Par("Session", "Make Raw-D") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Raw-D") == "yes");
        _want[wv::PCA] = (session_profile.Par("Session", "Make PCA") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make PCA") == "yes");
        _want[wv::PCA_D] = (session_profile.Par("Session", "Make PCA-D") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make PCA-D") == "yes");
#ifdef USE_GSL
        _want[wv::SPLINE] = (session_profile.Par("Session", "Make Spline") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Spline") == "yes");
        _want[wv::SPLINE_D] = (session_profile.Par("Session", "Make Spline-D") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Spline-D") == "yes");
        _want[wv::FOURIER] = (session_profile.Par("Session", "Make Fourier") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Fourier") == "yes");
        _want[wv::FOURIER_D] = (session_profile.Par("Session", "Make Fourier-D") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Fourier-D") == "yes");
#endif
        _want[wv::WAVELET] = (session_profile.Par("Session", "Make Wavelets") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Wavelets") == "yes");
        _want[wv::WAVELET_D] = (session_profile.Par("Session", "Make Wavelets-D") == Profile::BAD_PARAMETER
                || session_profile.Par("Session", "Make Wavelets-D") == "yes");

        _want[wv::RAW_D] |= _want[wv::PCA_D]
#ifdef USE_GSL
                || _want[wv::SPLINE_D] || _want[wv::FOURIER_D]
#endif
                || _want[wv::WAVELET_D] || _want[wv::SLOPE_D];
            //Must make _rawD if we derive any of the transforms that depend on it!
    }
    else
        _want[wv::RAW_D] = _want[wv::PCA] = _want[wv::PCA_D]
#ifdef USE_GSL
                = _want[wv::SPLINE] =  _want[wv::SPLINE_D] = _want[wv::FOURIER] = _want[wv::FOURIER_D]
#endif
                = _want[wv::WAVELET] = _want[wv::WAVELET_D] = false;

#ifdef USE_GSL
    if (num_samples<12) _want[wv::SPLINE_D] = _want[wv::SPLINE] = false;  //Possibly could relax this to 6 (->8) and have an option for fewer breakpoints.
#endif
}
MatPtr WaveData::LoadTransform(wv::TRANSFORM transform, wv::TRANSFORM_ID tid) const
{
//    qDebug() << _electrode << ", Loading " << wv::TransformName(transform).c_str();
    std::string wfm_path = FileBase::MakeTempFileName(FileBase::UserPath(), _fileName, transform, tid, "wfm", _electrode);
    WfmFileIn temp_file(wfm_path);
    temp_file.Load();
    return MatPtr( temp_file.GData(), wv_matrix_free );
}
int WaveData::ScrubArtifacts()
{
    int num_records = _waves[wv::RAW]->size1,
        num_samples = _waves[wv::RAW]->size2;
    wv_matrix* wave_height = wv_matrix_alloc(num_records, 1),
               * wave_width = wv_matrix_alloc(num_records, 1),
               * wave_slope = wv_matrix_alloc(num_records, 1);
    wv::float_t sd_limit = boost::lexical_cast<wv::float_t>(
                session_profile.Par("Session", "Scrub threshold (SDs)").c_str() );
    if (sd_limit<=0.0)
    {
        Log("WARNING: Invalid Scrub Threshold, please reset.  SD Limit = 3.00 being used");
        sd_limit = 4.0;
    }

    //Get alignment preference
    ALIGNMENT alignment;
    std::string alignment_str = session_profile.Par("Session", "Align waves");
    if (alignment_str == "peak") alignment = PEAK;
    else if (alignment_str == "trough") alignment = TROUGH;
    else alignment = NO_ALIGNMENT;

    //Get height and width and max index for aligning, if used.
    bool do_align = alignment != NO_ALIGNMENT;
    wv_vector_int* align_indexes = nullptr;
    if (do_align) align_indexes = wv_vector_int_alloc(num_records);
    _minima[wv::RAW] = _maxima[wv::RAW] = wv_matrix_get(_waves[wv::RAW].get(), 0, 0);
    for (int i=0; i<num_records; ++i)
    {
        wv_vector_view row = wv_matrix_row(_waves[wv::RAW].get(), i);
        size_t imin, imax;
        wv_vector_minmax_index(&row.vector, &imin, &imax);
        wv::float_t min = wv_vector_get(&row.vector, imin),
                max = wv_vector_get(&row.vector, imax);
        if (min<_minima[wv::RAW]) _minima[wv::RAW] = min;
        if (max<_maxima[wv::RAW]) _maxima[wv::RAW] = max;
            //Could use gsl_stats_quantile_from_sorted_data on vectors of mins, maxes to
            //get a better range but the user will probably be zooming in anyway

        //height and width
        wv_matrix_set(wave_height, i, 0, max-min);
        wv::float_t width = abs((int)(imax-imin)) * _usPerSamp;
        wv_matrix_set(wave_width, i, 0, width);
        if (imin==imax)
            wv_matrix_set(wave_slope, i, 0, 0.0);
        else
            wv_matrix_set(wave_slope, i, 0, (wv::float_t)(max-min) / width);

        //Alignment indices
        if (do_align)
        {
            wv_vector_minmax_index(&row.vector, &imin, &imax);
            if (alignment==PEAK) wv_vector_int_set(align_indexes, i, imax);
            else wv_vector_int_set(align_indexes, i, imin);
        }
    }

    //Now calculate alignment, if used
    if (do_align)
    {
        AlignWaveforms(align_indexes);
        wv_vector_int_free(align_indexes);
    }

    //Get all 4 thresholds, upper and lower bounds for minima and maxima
    wv_vector_view height_view = wv_matrix_column(wave_height, 0);
    wv::float_t height_mean = wv_stats_mean(height_view.vector.data, height_view.vector.stride, num_records);
    wv::float_t height_sd = wv_stats_sd_m(height_view.vector.data, height_view.vector.stride,
                                       num_records, height_mean);
    wv::float_t height_ub = height_mean + sd_limit * height_sd,
            height_lb = height_mean - sd_limit * height_sd;

    //Now create the bit array of waves which fall outside the limits
    _isScrubbed = boost::dynamic_bitset<>(num_records);
    int ct=0;
    for (int i=0; i<num_records; ++i)
    {
        wv::float_t height = wv_matrix_get(wave_height, i, 0);
        if (height>height_ub || height<height_lb)
        {
            _isScrubbed.set(i, true);
            ++ct;
        }
    }

    //Create the new data matrix
        //Allocate
    int num_records_scrubbed = ct,
        num_new_records = num_records - num_records_scrubbed;
    wv_matrix* new_data = wv_matrix_alloc(num_new_records, num_samples),
            * new_timestamps = wv_matrix_alloc(num_new_records, 1),
            * new_wave_height = wv_matrix_alloc(num_new_records, 1),
            * new_wave_width = wv_matrix_alloc(num_new_records, 1),
            * new_wave_slope = wv_matrix_alloc(num_new_records, 1);
    wv_vector_int* new_online_clus(nullptr);
    if (_hasOnlineClus) new_online_clus = wv_vector_int_alloc(num_new_records);

        //Write in the non-scrubbed data points and construct the all-timestamps vector
    ct = 0;
    _allTimestamps = wv_vector_int_alloc(num_records);
    wv_matrix* raw_mat = _waves[wv::RAW].get(),
            * time_mat = _waves[wv::TIMESTAMP].get();
    for (int i=0; i<num_records; ++i)
    {
        wv::float_t tm = wv_matrix_get(time_mat, i, 0);
        wv_vector_int_set(_allTimestamps, i, (int)tm);

        if (!_isScrubbed.test(i))
        {
            //waveforms
            wv_vector_view row = wv_matrix_row(raw_mat, i);
            wv_matrix_set_row(new_data, ct, &row.vector);

            //timestamps
            wv_matrix_set(new_timestamps, ct, 0, tm);

            //metrics
            wv::float_t ht = wv_matrix_get(wave_height, i, 0),
                    wd = wv_matrix_get(wave_width, i, 0),
                    slope = wv_matrix_get(wave_slope, i, 0);
            wv_matrix_set(new_wave_height, ct, 0, ht);
            wv_matrix_set(new_wave_width, ct, 0, wd);
            wv_matrix_set(new_wave_slope, ct, 0, slope);

            //Online cluster if there is one
            if (_hasOnlineClus)
            {
                int clus = wv_vector_int_get(_onlineVCluster, i);
                wv_vector_int_set(new_online_clus, ct, clus);
            }

            ++ct;
        }
    }
    _numRecords = num_new_records;

    //Replace metrics
    _waves[wv::HEIGHT] = MatPtr( new_wave_height, wv_matrix_free );
    _waves[wv::WIDTH] = MatPtr( new_wave_width, wv_matrix_free );
    _waves[wv::SLOPE] = MatPtr( new_wave_slope, wv_matrix_free );

    //Enable metrics
    _have[wv::HEIGHT] = true;
    _have[wv::WIDTH] = true;
    _have[wv::SLOPE] = true;
    emit EnableTransform(wv::HEIGHT, _electrode);
    emit EnableTransform(wv::WIDTH, _electrode);
    emit EnableTransform(wv::SLOPE, _electrode);

    //Free resources and replace old waveforms.
    _waves[wv::RAW] = MatPtr( new_data, wv_matrix_free );
    _waves[wv::TIMESTAMP] = MatPtr( new_timestamps, wv_matrix_free );
    if (_hasOnlineClus)
    {
        wv_vector_int_free(_onlineVCluster);
        _onlineVCluster = new_online_clus;
    }

    wv_matrix_free(wave_height);
    wv_matrix_free(wave_width);
    wv_matrix_free(wave_slope);

    //Finally report the outcome
    return num_records_scrubbed;
}
void WaveData::AlignWaveforms(const wv_vector_int* align_indexes)
{
    //Align relative to the modal index value for waveform peak/trough.  Get the mode by histogramming the
    //indexes, then taking the bin with the largest value.
    int num_records = _waves[wv::RAW]->size1,
            num_samples = _waves[wv::RAW]->size2;

        //Finding the modal sample # for peak/trough
    wv_histogram* hist = wv_histogram_alloc(num_samples);
    wv_histogram_set_ranges_uniform(hist, -0.5, (wv::float_t)num_samples - 0.5);
    for (int i=0; i<num_records; ++i)
    {
        wv::float_t index = (wv::float_t)wv_vector_int_get(align_indexes, i);
        wv_histogram_increment(hist, index);
    }
    int align_sample = wv_histogram_max_bin(hist);

        //Get offsets
    wv_vector_int* offsets = wv_vector_int_alloc(num_records);
    wv_vector_int_memcpy(offsets, align_indexes);
    wv_vector_int_add_constant(offsets, -align_sample);

        //Now go through all the waveforms and do a circular shift as required.
        //But watch out for spurious side effects depending on the amount of shift,
        //with the first few samples!  I.e. when the last samples become the first
        //producing a bogus classification.
    AlignmentMgr::AlignmentSpec alignment_spec(_waves[wv::RAW].get(), offsets);
    AlignmentMgr mgr(alignment_spec);
    mgr.StartThreads();
    mgr.Wait();

    wv_histogram_free(hist);
    wv_vector_int_free(offsets);
}
void WaveData::AppendToFile(wv::TRANSFORM transform, const wv_matrix* mat, int num_rows) const
{
    WfmFileAppend* out = transform==wv::RAW ? _fpWaves : _fpTimestamps;
    out->Append(mat, num_rows);
}
void WaveData::CalcForBatch(wv::TRANSFORM transform)
{
//    qDebug() << Electrode() << ", " << wv::TransformName(transform).c_str() << ", CalcForBatch";
    TransformTOBase* transform_to = TransformTOBase::Create(
                transform,
                IsDerivative(transform) ? _waves[wv::RAW_D] : _waves[wv::RAW],
#ifdef USE_GSL
                _usPerSamp,
#endif
                _fullPath, _electrode);
    transform_to->StartTimer();
    transform_to->Start();

    _threads[transform] = nullptr;
    _transformTOs[transform] = transform_to;

    _waves[transform] = transform_to->Waveforms();
    if (!_waves[transform])
    {
        Log("WARNING: " + wv::TransformName(transform)
            + " analysis failed, null pointer returned.  " \
            "This is most likely caused by insufficient memory.");
        return;
    }

    Log( wv::TransformName(transform) + " analysis finished in "
         + std::to_string(transform_to->TimeElapsed()/1000.0) + " seconds.");
    _have[transform] = true;
    _minima[transform] = wv_matrix_min(_waves[transform].get());
    _maxima[transform] = wv_matrix_max(_waves[transform].get());

    delete transform_to; //This frees the transform shared_ptr
    _transformTOs[transform] = nullptr;

    //In either batch mode, we know we don't want to keep these transforms around
    Unload(transform, false); //It's locked from Preprocess
}
    //Annoyingly, signals/slots don't work from a worker thread, e.g. in batch.  So for batch processing,
    //run all transformation calculations serially. --But it's better that way anyway
void WaveData::Calculate(wv::TRANSFORM transform)
{
//    qDebug() << Electrode() << ", " << wv::TransformName(transform).c_str() << ", Calculate";
    TransformTOBase* transform_to = TransformTOBase::Create(
                transform,
                IsDerivative(transform) ? _waves[wv::RAW_D] : _waves[wv::RAW],
#ifdef USE_GSL
                _usPerSamp,
#endif
                _fullPath, _electrode);
    QThread* thread = new QThread;
    connect(this, SIGNAL(Start()), transform_to, SLOT(Start()), Qt::QueuedConnection);
    connect(transform_to, SIGNAL(Finished(void*)), this, SLOT(Finished(void*)));
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
    transform_to->moveToThread(thread);
    thread->start();
    emit Start();
    disconnect(this, SIGNAL(Start()), transform_to, SLOT(Start()));
    transform_to->StartTimer();
    Log("Starting " + wv::TransformName(transform) + " transform.");

    if (transform!=wv::RAW && transform!=wv::RAW_D) ++_threadCount;
        //wv::RAW and wv::RAW_D are not calculated in separate threads
    _threads[transform] = thread;
    _transformTOs[transform] = transform_to;
}
void WaveData::RemoveDC()
{
    //Best way: Calculate the mean of each record, then take mean of means, then constant subtract
    int num_records = _waves[wv::RAW]->size1,
            num_samples = _waves[wv::RAW]->size2;
    wv_vector* means = wv_vector_alloc(num_records);

    //Get the mean of each sample
    for (int i=0; i<num_records; ++i)
    {
        wv_vector_view row = wv_matrix_row(_waves[wv::RAW].get(), i);
        wv::float_t m = wv_stats_mean(row.vector.data, row.vector.stride, num_samples);
        wv_vector_set(means, i, m);
    }

    //Get the grand mean and subtract from the matrix
    wv::float_t gm = wv_stats_mean(means->data, means->stride, num_records);
    wv_matrix_add_constant(_waves[wv::RAW].get(), -gm);

    _minima[wv::RAW] -= gm;
    _maxima[wv::RAW] -= gm;

    //Free resources
    wv_vector_free(means);
}
void WaveData::MakeDiff()
{
    //Allocate
    int num_records = _waves[wv::RAW]->size1,
        num_samples = _waves[wv::RAW]->size2,
        num_samples_d = num_samples - 1;
    wv_matrix* raw_d = wv_matrix_alloc(num_records, num_samples_d),
        * slope_d = wv_matrix_alloc(num_records, 1);

    bool do_smooth = (num_samples>24);
        //Obviously need enough samples for this to be a good idea

    //row-wise differentiation
    DiffMgr::DiffSpec diff_spec(_waves[wv::RAW].get(), raw_d, slope_d);
    diff_spec.do_smooth = do_smooth;
    diff_spec.us_per_samp = _usPerSamp;
    DiffMgr mgr(diff_spec);
    mgr.StartThreads();
    mgr.Wait();

    _waves[wv::RAW_D] = MatPtr( raw_d, wv_matrix_free);
    _waves[wv::SLOPE_D] = MatPtr( slope_d, wv_matrix_free);

    emit EnableTransform(wv::SLOPE_D, _electrode);
    _have[wv::SLOPE_D] = true;

    _have[wv::RAW_D] = true;
    _minima[wv::RAW_D] = wv_matrix_min(_waves[wv::RAW_D].get());
    _maxima[wv::RAW_D] = wv_matrix_max(_waves[wv::RAW_D].get());

    WriteToDisk(wv::RAW_D, _waves[wv::RAW_D]);
}
#ifndef USE_GSL
void WaveData::MakeEnergy()
{
    const int num_rows = NumRecords();
    wv_matrix* energy = wv_matrix_alloc(num_rows, 1);
    for (int i=0; i<num_rows; ++i)
    {
        wv_vector_const_view cv = wv_matrix_const_row(_waves[wv::RAW].get(), i); // ###
        wv_vector v;
        v.size = cv.vector.size;
        v.stride = cv.vector.stride;
        v.data = cv.vector.data;
        wv_matrix_set(energy, i, 0, msl::Energy(&v));
    }
    _waves[wv::ENERGY] = MatPtr( energy, wv_matrix_free );
    _have[wv::ENERGY] = true;
    WriteToDisk(wv::ENERGY, _waves[wv::ENERGY]);
}
#endif
void WaveData::UpdateUTSignatures() const
{
    MutexLocker lock(_mutex);
    _utSignatures.clear();
    for (const auto& it : _userTransforms)
    {
        Signature signature;
        for (const auto& its : it.UserSpecs())
            signature.push_back(its);
            //So each transform, integer pair gets taken from UserSpecs and pushed directly
            //into the signatures.
        _utSignatures.push_back(signature);
    }
}
MatPtr& WaveData::UserWaves(wv::TRANSFORM_ID tid)
{
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == tid)
        {
            MatPtr& mat_ptr = _userTransforms[i].Waves();
            if (mat_ptr.use_count() == 0) mat_ptr = LoadTransform(wv::USER, tid);
            return mat_ptr;
        }
    throw BadTID("WaveData::UserWaves", tid);
}
const MatPtr& WaveData::UserWaves(wv::TRANSFORM_ID tid) const
{
    int num_userts = _userTransforms.size();
    for (int i=0; i<num_userts; ++i)
        if (_userTransforms.at(i).ID() == tid)
        {
            const MatPtr& mat_ptr = _userTransforms.at(i).Waves();
            if (mat_ptr.use_count() == 0)
                throw BadUserTransform("WaveData::UserWaves", tid);
            return mat_ptr;
        }
    throw BadTID("WaveData::UserWaves", tid);
}
void WaveData::WriteToDisk(wv::TRANSFORM transform, MatCPtr mat, bool header_only) const
{
    //To lazy to write a switch...

    FileBase::WfmHeader header =
    {
        (uchar)0,
        (int)mat->size1,
        (int)mat->size2,
        sizeof(wv::float_t),
        0,
        1.0,
        0.0
    };
    std::string full_path = FileBase::UserPath();

    //The raw data
    std::string temp_file = FileBase::MakeTempFileName(full_path, _fileName, transform, 0, "wfm", Electrode());
    WfmFileOut out(temp_file, mat, header, header_only);
    out.Save();
//    QApplication::processEvents();
}
void WaveData::WriteFeatureVectors()
{
    //Initialize common parameters
    FileBase::WfmHeader header =
    {
        (uchar)0,
        (int)_waves.at(wv::RAW)->size1,
        1,
        sizeof(wv::float_t),
        0,
        1.0,
        0.0
    };
    std::string full_path = FileBase::UserPath();

    //Go through the feature vectors
    std::string temp_file;
    temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::TIMESTAMP, 0, "wfm", Electrode());
    WfmFileOut(temp_file, _waves.at(wv::TIMESTAMP), header).Save();
    temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::HEIGHT, 0, "wfm", Electrode());
    WfmFileOut(temp_file, _waves.at(wv::HEIGHT), header).Save();
    temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::WIDTH, 0, "wfm", Electrode());
    WfmFileOut(temp_file, _waves.at(wv::WIDTH), header).Save();
    temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::SLOPE, 0, "wfm", Electrode());
    WfmFileOut(temp_file, _waves.at(wv::SLOPE), header).Save();
    if (_have[wv::SLOPE_D])
    {
        temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::SLOPE_D, 0, "wfm", Electrode());
        WfmFileOut(temp_file, _waves.at(wv::SLOPE_D), header).Save();
    }
#ifndef USE_GSL
    temp_file = FileBase::MakeTempFileName(full_path, _fileName, wv::ENERGY, 0, "wfm", Electrode());
    WfmFileOut(temp_file, _waves.at(wv::ENERGY), header).Save();
#endif
}
