#ifndef WAVEDATA_H
#define WAVEDATA_H

#include "../globals/globals.h"
#include "parameters/profile.h"
#include "../memrep/specs/usertransformspec.h"
#include "../analysis/transforms/pca_to.h"
#ifdef USE_GSL
#include "../analysis/transforms/splineto.h"
#include "../analysis/transforms/fourierto.h"
#endif
#include "../analysis/transforms/haarto.h"
#include "../analysis/threads/mgr/alignmentmgr.h"
#include "../analysis/threads/mgr/diffmgr.h"

//This class is the representation of the waveforms, plus additional necessary information,
//suitable for any analysis.

class WfmFileAppend;
class WaveformsInBase; //Compiler won't let me forward-declare WavesFilePtr
class WaveData : public QObject//Eventually turn this into a hierarchy to account for idiosyncratic
        //information known by MEX vs. other spike acquisition software
{
    //NO FRIENDS--can't risk any accces to transform pointers

    Q_OBJECT

    public:
        static std::vector<Signature> _utSignatures;

        WaveData(const WaveformsInBase* waves_file, short electrode, wv::float_t us_per_samp = 1.0);
            //Raw pointer is dangerous!!
#ifndef Q_OS_WIN
        WaveData(const WaveData&) = delete;
        WaveData& operator=(const WaveData&) = delete;
#endif
        ~WaveData();

        void AllocateMemory(int num_records);
        void AddToUser(wv::TRANSFORM_ID tid, wv_vector_const_view view,
                       wv::TRANSFORM transform, int sample, bool update_signatures = false);
        void AddUserTransform(UTS ut_spec) { _userTransforms.push_back(ut_spec); }
        bool CheckUT(const Signature& signature) const; //Whether the transforms out of which this
            //User Transform is composed have been constructed yet
        wv::TRANSFORM_ID CreateUserTransform(wv::TRANSFORM_ID old_id = wv::DEFAULT_TID);
        wv::TRANSFORM_ID CreateUserTransform(const UTS& other_uspec); //Creates a UTS from *another* WaveData object;
            //used during batch processing
        void DecThreadCount() { --_threadCount; }
        void DeleteFPs();
        void DisplayUseCounts() const;
        void IncThreadCount() { ++_threadCount; }
        int Preprocess(); //Scrubbing artifacts et al.
        static void ReadUTSignatures(FILE* fp);
        void Reload(); //For batch processing.  Read in all data files (where they exist), generate otherwise.
        void RemoveFromUser(wv::TRANSFORM_ID tid, int col);
        void RemoveUserTransform(wv::TRANSFORM_ID id);
            //These manage the user-defined transform.  Min and max are included in the first
            //one to make scaling more efficient.
        void SetOnlineClusterCID(wv::CLUSTER_ID cid) { _onlineClusterCID = cid; }
        void SetupWavefiles(int num_waves, int num_samples);
        void SetUTsFromSignatures(); //Only call if WaveData currently has 0 associated UTs
        void TerminateProcesses(); //When a new input file is loaded, any processes not yet finished
            //here need to be terminated, or else they would enable the WaveSpaceGui radiobutton
        void Unload();
            // *All* memory is freed.  Call this every time a clustering gets switched out
        void Unload(wv::TRANSFORM transform, bool lock = true); //For when the raw data is loaded for a WaveData instance that is otherwise
            //not active.  E.g., GridWaveWidgets
        void WriteTimestamp(uint timestamp);
        void WriteUserTransform(wv::TRANSFORM_ID tid) const;
        static void WriteUTSignatures(FILE* fp);
        void WriteWaveform(wv_vector_view wave_view);
        void WriteWaveform(short wave[], int num_samples);

        void SetHasOnlineClus(bool b) { _hasOnlineClus = b; }
        void SetOnlineClus(int i, int clus);
        void SetTimestamp(int i, long long int timestamp);
        void SetWaveform(int i, short wave[]); //Overload as needed
        void SetWaveform(int i, const wv_vector_view wave);

        const wv_vector_int* AllTimestamps() const { return _allTimestamps; }
        short Electrode() const { return _electrode; }
        std::string FullPath() const { return _fullPath; }
        std::string FileName() const { return _fileName; }
        std::string FilePath() const { return _filePath; }
        bool HasOnlineClus() const { return _hasOnlineClus; }
        bool HaveTransform(wv::TRANSFORM transform, wv::TRANSFORM_ID tid = wv::DEFAULT_TID) const;
        bool IsBatch() const;
        bool IsLoaded() const { return _isLoaded; }
        bool IsDerivative(wv::TRANSFORM transform) const;
        bool IsPreprocFinished() const { return _isPreprocFinished; }
        int NumAllRecords() const { return _isScrubbed.size(); } //Includes scrubbed waveforms
        int NumRecords() const { return _numRecords; }
            //Because sometimes you need this within a const object.  Naturally, can only be used when
            //it's known that _raw actually exists.
        int NumRecords() { return _numRecords; }
        int NumSamples() { return _waves[wv::RAW]->size2; }
        int NumSamples(wv::TRANSFORM transform, wv::TRANSFORM_ID tid);
            //Not all transforms need to have the same size, particularly since we're doing dimension reduction.
        int NumUserTransforms() const { return _userTransforms.size(); }
        wv_vector_int* OnlineVCluster() { return _onlineVCluster; }
        wv::CLUSTER_ID OnlineClusterCID() const { return _onlineClusterCID; }
        const boost::dynamic_bitset<>& Scrubbed() const { return _isScrubbed; }
        int ThreadCount() { QApplication::processEvents(); return _threadCount; }
        MatCPtr Timestamps(); //Need this separate from Waveforms b/c it's const
        wv::float_t TMin(wv::TRANSFORM transform) const; //T for transform
        wv::float_t TMax(wv::TRANSFORM transform) const;
        std::string TransformLabel(wv::TRANSFORM transform, int sample,
                                   wv::TRANSFORM_ID tid = wv::DEFAULT_TID) const;
        int UserSample(wv::TRANSFORM_ID tid, int sample) const;
            //sample here refers to the selected sample of the *user* transform, which *indexes* a
            //particular transform/sample pair.
        int UserSize(wv::TRANSFORM_ID tid) const;
        UTS& UTSpec(wv::TRANSFORM_ID tid);
        UTS& UTSpecAt(int sample);
        wv::TRANSFORM_ID UserTid(size_t index) const; //tid of the ith user transform
        wv::TRANSFORM UserTransform(wv::TRANSFORM_ID id, int sample) const;
        const std::vector<UTS>& UserTransforms() const { return _userTransforms; }
        MatCPtr Waveforms(wv::TRANSFORM transform, wv::TRANSFORM_ID id = wv::DEFAULT_TID);

    signals:
        void Start() const;

        void EnableTransform(wv::TRANSFORM transform, short electrode);
            //This is for the WaveSpaceGui, but only update if data for the current electrode is shown!

    public slots:
        void Finished(void* data);
        //returns a pointer to the matrix.  WaveData is responsible for
        //its deletion.

    private:
        static const int NUM_BUFFER_ROWS;
        static const int NUM_TBUFFER_ROWS;
        enum ALIGNMENT
        {
            NO_ALIGNMENT,
            PEAK,
            TROUGH
        };

#ifdef Q_OS_WIN
        WaveData(const WaveData&);
        WaveData& operator=(const WaveData&);
#endif

        void SetWantedTransforms();

        void AlignWaveforms(const wv_vector_int* align_indexes);
        void AppendToFile(wv::TRANSFORM transform, const wv_matrix* mat, int num_rows = -1) const;
        void CalcForBatch(wv::TRANSFORM transform);
        void Calculate(wv::TRANSFORM transform);
        void FreeAll();
        MatPtr LoadTransform(wv::TRANSFORM transform, wv::TRANSFORM_ID tid = wv::DEFAULT_TID) const;
        MatPtr& Lookup(wv::TRANSFORM transform, wv::TRANSFORM_ID tid = wv::DEFAULT_TID);
        void MakeDiff(); //Create _rawD, the differentiated raw data
#ifndef USE_GSL
        void MakeEnergy();
#endif
        void RemoveDC(); //Subtract the grand mean from the raw waveforms matrix
        int ScrubArtifacts();
            //Eliminate waveforms whose peaks or valleys fall >SD_LIMIT std devs from the mean
            // **** Also raw wave metrics are created here
        void UpdateUTSignatures() const; //Every time utspecs is changed, all signatures are
            //simply rewritten.
        MatPtr& UserWaves(wv::TRANSFORM_ID tid); //A reference because we don't want to increase
            //the use count internally
        const MatPtr& UserWaves(wv::TRANSFORM_ID tid) const;
            //A reference because we don't want to increase
            //the use count internally
        void WriteFeatureVectors(); //Writes _timestamps, _waveHeight, etc.
        void WriteToDisk(wv::TRANSFORM transform, MatCPtr mat, bool header_only = false) const;
            //Writes wfm files for _raw, _rawD

        wv_vector_int* _allTimestamps;
        const short _electrode; //Which electrode the data comes from
        std::string _fileName, _filePath;
        WfmFileAppend* _fpTimestamps, * _fpWaves; //The files for the initial write-in of raw waveforms and timestamps
        std::string _fullPath;
        bool _hasOnlineClus;
        std::map<wv::TRANSFORM, volatile bool> _have;
            //volatile of course because these are computed in worker threads
        std::map<wv::TRANSFORM, bool> _isCalculated; //Have to specify this individually
        bool _isLoaded, //Transforms that are being calculated will be unloaded when they are completed
            _isPreprocFinished;
        boost::dynamic_bitset<> _isScrubbed; //Have to keep track of which waveforms got scrubbed, for rewriting to
            //the simple binary and other storage formats
        wv_matrix* _loadTimeBuffer, * _loadWavesBuffer;
        int _loadTimeCt, _loadWavesCt; //These 4 are critical for efficient file loading.  We separate
            //counters for time and waves so that it doesn't matter which gets called first in the input
            //file.
        std::map<wv::TRANSFORM, wv::float_t> _maxima;
        std::map<wv::TRANSFORM, wv::float_t> _minima;
        mutable std::mutex _mutex;
        int _numRecords; //Its own member to avoid mutex waiting
        wv_vector_int* _onlineVCluster; //NULL if there isn't one
        wv::CLUSTER_ID _onlineClusterCID;
        volatile int _threadCount;
        std::map<wv::TRANSFORM, QThread*> _threads;
        std::vector<wv::TRANSFORM> _transforms; //Holds all the transforms, so they can be iterated
            //over
        std::map<wv::TRANSFORM, TransformTOBase*> _transformTOs;
        const wv::float_t _usPerSamp;
        std::vector<UTS> _userTransforms; //When they are *empty*, that means they haven't been loaded yet
        std::map<wv::TRANSFORM, bool> _want;
        std::map<wv::TRANSFORM, MatPtr> _waves;
        const WaveformsInBase* _wavesFile;
};

#endif // WAVEDATA_H
