#include "wavetemplate.h"
#include "clustermediator.h"

WaveTemplate::WaveTemplate(MatCPtr waves, const boost::dynamic_bitset<>& window,
                           const std::vector<int>& selected, const wv_vector_int* vcluster)
    : _vCluster(vcluster)
{
    int num_records = waves->size1,
        num_samples = waves->size2;

    //Clear out old templates
    DeleteTemplates();

    //###Multi-thread this
    //Now calculate all the information
    wv_vector* diff = wv_vector_alloc(num_samples); //Difference between template, wave
        //These are used to temporarily hold row vectors
    for (auto clus : selected)
    {
        if (clus == wv::NULL_CLUSTER) continue;

        //Allocate a matrix of the appropriate size to hold all the waveforms in this cluster
        int numc = ClusterCount(clus, window);
        if (numc==0)  //Will happen when the window doesn't contain any of a given cluster
        {
            _templates[clus] = nullptr;
            _templateSDs[clus] = nullptr;
            continue;
        }
        wv_matrix* clus_mat = wv_matrix_alloc(numc, num_samples);

        //Allocate data members
        _templates[clus] = wv_vector_alloc(num_samples);
        _templateSDs[clus] = wv_vector_alloc(num_samples);

        //Now fill it with those waveforms
        int ct = 0;
        for (int i=0; i<num_records; ++i)
            if ( window.test(i) && wv_vector_int_get(_vCluster, i) == clus )
            {
                wv_vector_const_view row = wv_matrix_const_row(waves.get(), i);
                wv_matrix_set_row(clus_mat, ct++, &row.vector);
            }

        //Now calculate columnwise average and sd, put values into the corresponding template data members
        for (int j=0; j<num_samples; ++j)
        {
            wv_vector_view col = wv_matrix_column(clus_mat, j);
            wv::float_t mean = wv_stats_mean(col.vector.data, col.vector.stride, numc);
            wv::float_t sd = wv_stats_sd(col.vector.data, col.vector.stride, numc);

            wv_vector_set(_templates[clus], j, mean);
            wv_vector_set(_templateSDs[clus], j, sd);
        }

        //Free the resources, which were temporary
        wv_matrix_free(clus_mat);
    }
    wv_vector_free(diff);
}
WaveTemplate::WaveTemplate(const WaveTemplate& wave_template)
{
    Copy(wave_template);
}
WaveTemplate& WaveTemplate::operator=(const WaveTemplate& wave_template)
{
    if (this!=&wave_template) Copy(wave_template);
    return *this;
}
WaveTemplate::~WaveTemplate()
{
    DeleteTemplates();
}

int WaveTemplate::ClusterCount(int c, const boost::dynamic_bitset<>& window)
{
    int num_records = _vCluster->size;
    int ct = 0;
    for (int i=0; i<num_records; ++i)
        if ( window.test(i) && wv_vector_int_get(_vCluster, i) == c )
            ++ct;
    return ct;
}

void WaveTemplate::Copy(const WaveTemplate& wave_template)
{
    DeleteTemplates();

    auto wave_templates_end = wave_template._templates.end();
    for (auto it = wave_template._templates.begin(), it_sd = wave_template._templateSDs.begin();
         it!= wave_templates_end;
         ++it, ++it_sd)
    {
        int key = it->first;
        if (it->second == nullptr)
        {
            _templates[key] = nullptr;
            _templateSDs[key] = nullptr;
            continue;
        }
        wv_vector* temp = wv_vector_alloc( it->second->size ),
                * temp_sd = wv_vector_alloc( it->second->size );

        wv_vector_memcpy(temp, it->second);
        _templates[key] = temp;

        wv_vector_memcpy(temp_sd, it_sd->second);
        _templateSDs[key] = temp_sd;
    }
}

void WaveTemplate::DeleteTemplates()
{
    for (auto it : _templates) if (it.second) wv_vector_free(it.second);
    for (auto it : _templateSDs) if (it.second) wv_vector_free(it.second);
    _templates.clear();
    _templateSDs.clear();
}
